                      ;********************************************************************************************************************************
                      ; This  is  the  ROM-add-on  that  acts  depending  upon  the  drive  number,  via  DW4  or  via  DOS  as  usual
                      ; Verified 43 commands of the DosPlus49b, all of them worked flawlessly with drive / VDK of any size (180-360-360-720)
                      ; I have left BOOT apart because it seems that only drive1 can be booted
                      ; To BOOT an OS-9 or NitrOS-9 disk it should be mounted on DriveWire slot 0
                      ;    and it has to be a special boot disk to communicate with the DW server via parallel adapter or Becker Port
                      ; NOTE: The commands that modify directory sectors (T20), don't update backup T16 (by these DOS design)
                      ;********************************************************************************************************************************
                      ; From $fc to $ff will contain a slot number for units 1-2-3-4
                      ; if they are greater than 4, then they must be pointing to a reserved DW4 slot, else they are actual floppy drives (1 to 4)
                      ; last modification: slot 0 is also considered to be Drivewire (for booting OS-9 or NitrOS-9)
                      ; numbers from 1 to 4 are ONLY accepted at the UNIT with the SAME number and just for physical drives
                      ;********************************************************************************************************************************
                      ; This extender is intended for DPlus50
                      ; DW42-0.25.10 - 2020-08-07 by Pere Serrat (pser1)
                      ;********************************************************************************************************************************
                            IF drgbin > 0                                                     ; to create Patcher - Extender binaries
DFF7                           org   $e000-9                                          ; to create the BIN header
DFF7  55                                      fcb     #$55                                                    ; header values for a Dragon binary file. magic 1st byte
DFF8  02                                      fcb     #$02                                                    ; file type
DFF9  6000                                    fdb     #$6000                                          ; load
DFFB  1B67                                    fdb     #absEnd-$e000                           ; length
DFFD  6002                                    fdb     #$6002                                          ; exec
DFFF  AA                                      fcb     #$aa                                                    ; magic end byte
                                      ELSE
                                              org     $e000                                                   ; real program beginning
                               put   $e000                                                    ; allocated at $e000
                            ENDIF
                                                                                                                      ; ROM routines
009F                  getNChr equ     $9f                                                     ; get next char from input buffer
00A5                  redLChr equ     $a5                                                     ; get last read char from input buffer
8000                  basini  equ   $8000                                                     ; basic rom beginning
8344                  syserr   equ   $8344                                                    ; system error
8371                  okprmpt  equ   $8371                                                    ; BASIC OK prompt, command loop
83ED                  basvect2 equ   $83ed                                                    ; initialize BASIC
841F                  basvect1 equ   $841f                                                    ; reset BASIC vectors, reset stack etc
849F                  runbasic equ   $849f                                                    ; run_basic, enable IRQs etc
84ED                  romcmd  equ     $84ed                           ; rom entry to dispatch a tokenized command
89AA                  ckComma equ     $89aa                                                   ; check for comma routine
89B4                  synerr  equ   $89b4                             ; rom entry to show syntax error
8B8D                  fcerr   equ   $8b8d                             ; entry to send message error
8E51                  asc2Num equ     $8e51                                                   ; convert ASCII to number (1 byte)
8E83                  get16bN equ   $8e83                                                     ; get a 16 bit number into x
90A5                  sendCR  equ     $90a5                                                   ; send a CR to screen
90E5                  outstr   equ   $90e5                                                    ; print string at X+1 to DEVNUM
957A                  outnum  equ   $957a                                     ; print number in d to devnum
9A89                  backspc  equ   $9a89                                                    ; print a backspace to DEVNUM
B3B4                  resetRt equ     $b3b4                                                   ; reset routine
B44F                  wrmStrt equ     $b44f                                                   ; warmStart routine
B54A                  outchr   equ   $b54a                                                    ; print char A to DEVNUM
B7AA                  getfnam  equ   $b7aa                                                    ; read file name and length into $1d1
B84B                  ioerror  equ   $b84b                                                    ; IO_error
BCAB                  outRegA equ     $bcab                                                   ; output reg A to screen as number
C000                  romini  equ   $c000                                                     ; beginning of roms area 
                                                                                                                      ; variables
0012                  dnum    equ   $12                                                       ; assigned drive number by dw4 server
0076                  chksum  equ   $76                                                       ; checksum of the sector to be sent to dw4 (2 bytes)
0076                  hdwFlo  equ     $76                                                     ; slot number where floppies are found (only for tstHDW)
0077                  hdwSdc  equ     $77                                                     ; slot number where SDC is found (only for tstHDW)
00E7                  retry           equ     $e7                                                     ; to catch 1st acess to DW4 drives error
00F8                  numUni  equ     $f8                                                     ; unit number for new Commands (link, mount, bank)
00F9                  trknum  equ     $f9                                                     ; tracknumber in process for DSKINIT 4th patch
00FA                  geodat  equ     $fa                                                     ; numTracks - numSect x track (2 bytes)
00FB                  secxtrk equ     $fb                                                     ; numSect x track (low byte of previous word)
00FC                  drvtab  equ     $fc                                                     ; system table for linked numbers to the 4 units (fc-fd-fe-ff)
0114                  floSDC  equ     $114                                                    ; bit 0 for unit 1, bit 1 for unit 2. If 0 use Floppy, if 1 use SDC
                                                                                                                              ; bit 2 is 1 if Floppies present else 0
                                                                                                                              ; bit 3 is 1 if SDC present else 0
                                                                                                                              ; bit 5-4 show the slot where floppies are connected (0-3)
                                                                                                                              ; bit 7-6 show the slot where SDC is connected (0-3)
0115                  idxHD           equ     $115                                                    ; index to a HD array (maximum 90 discs of 720+1 sectors DW4 VDK type) (values 1-90)
01D1                  nambuf  equ   $1d1                                                      ; fixed by BASIC, read filename function which stores in $1d1
                                                                                                                      ; constants
0050                  dimask  equ   $50                                                       ; disable IRQ and FIRQ
00AF                  eimask  equ   $af                                                       ; enable interrupts
                                                                                                                      ; program init
E000  4549            mark            fcb     $45,$49                                         ; "EI" mark to signal to the DOS that this extension wants to be called to initialize itself
E002  7EE58F                                  jmp     initext                                         ; initialize extension
E005  161B0A                                  lbra    reptch                                          ; direct jump to DOSPlus patcher
                      ;********************************************************************************************************************************
                      ; Entry point for DOS command dispatcher. 1st patch
                      ; Data received:
                      ;       A = track number
                      ;       B = command code  (2=READ - 3=Unknown - 4=WRITE - 7=VERIFY(dummy read)
                      ;       X = address of system variable "number of sectors x track" for this drive 
                      ;       Y = LSN
                      ;       U = sector counter (if not directory track) 
                      ;********************************************************************************************************************************
E008  3402            entry    pshs  a                                                                ; save register 'A' (code overwritten. It's the first instruction in the DOS)
E00A  3401                                    pshs  cc                                                                ; save flags
E00C  8D0F                                    bsr     phyvir                                          ; detect physical or virtual drive
E00E  241E                                    bcc     dw4                                                     ; if greater than 4 is a DriveWire slot number
E010  4D                                      tsta                                                                    ; is linked value equal 0? - demanat per KenH
E011  271B                                    beq     dw4                                                     ; yes, so DriveWire slot 0
                      ;********************************************************************************************************************************
                      ; The DOS part respects the original code
                      ;********************************************************************************************************************************
E013  3501            todos           puls    cc                                                              ; restore flags 
E015  170C80                                  lbsr    flOrSd                                          ; control access to floppy or SDC
E018  8602                                    lda   #$02                                      ; get value #2 (number of retries). It's the second instruction in the DPlus49b
E01A  7EC10A                                  jmp     $c10a                                                   ; back to next sentence of DOS original code
                      ;********************************************************************************************************************************
                      ; Detects if drive is physical or virtual (code in A)
                      ; returns Carry set if DW4 slot number detected (now any value greater than 4)
                      ; must use extended addressing because for DSKINIT, DP arrives here with $FF and if not virtual should return like that
                      ;********************************************************************************************************************************
E01D  3410            phyvir  pshs    x                                                               ; save working register
E01F  8E00FB                                  ldx     #drvtab-1                                       ; point to 1 byte before the slots table
E022  B600EB                                  lda     >$00eb                                          ; get drive number
E025  A686                                    lda     a,x                                                     ; get slot number for that drive
E027  B70012                                  sta     >dnum                                                   ; put it into DW4 variable
E02A  8105                                    cmpa    #5                                                              ; is a DW4 number? - if not, Carry will be set
E02C  3590                                    puls    x,pc                                                    ; restore register and return
                      ;********************************************************************************************************************************
                      ; The dw4 part distinguishes between READ, VERIFY and WRITE opcodes
                      ;********************************************************************************************************************************
E02E  3501            dw4     puls    cc                                                              ; get saved flags
                                                                                                                      ; to avoid problems created by SOUND and PLAY over DW4
E030  3402                                    pshs    a                                                               ; save register
E032  8602                                    lda     #2                                                              ; get value to mark output
E034  B7FF20                                  sta     BBOUT                                                   ; to DW4
E037  4F                                      clra                                                                    ; set counter
E038  4A              dw4L01  deca                                                                    ; decrement counter
E039  26FD                                    bne     dw4L01                                          ; not zero, loopback
E03B  3502                                    puls    a                                                               ; restore register
E03D  3474                                    pshs  b,x,y,u                                           ; save rest of registers (A was already onto stack) before recalculatimg Y
E03F  1A50                                    orcc  #dimask                                           ; disable interrupts to work with dw4                                                   
E041  8D4A                                    bsr     calcy                                                   ; do not relay on Y. Recalculate it from $eb-ec-ed
E043  8D68                                    bsr     ctrlHD                                          ; control index in HD if needed
E045  3021                                    leax  1,y                               ; X points now to the rigth dw4 sector (skipping the header) - [DW needs registers X-Y inverted!)
E047  109EEE                                  ldy   <$ee                                              ; Y points to RAM buffer address
E04A  C102                    cmpb  #2                                                                ; is this a read command?
E04C  270F                                    beq   reddw4                                            ; yes, read sector from dw4
E04E  C107                    cmpb  #7                                                                ; is this a verify command?
E050  2606                                    bne   wrtdw4                                            ; no, write sector to dw4
E052  108EFCFF                                ldy     #$fcff                                          ; initial byte of 256 area to be overwritten while verifying
E056  2005                                    bra     reddw4                                          ; go read that sector
E058  BDE2C7          wrtdw4  jsr   dowrit                                            ; write a sector from buffer to dw4
E05B  2010                                    bra   verify                                            ; verify operation result
                      ;********************************************************************************************************************************
                      ; This part asks dw4 a sector to be read and dw4 log reflects it accordingly. It is as simple as possible.
                      ;********************************************************************************************************************************
E05D  BDFA0F          reddw4  jsr   DoRead                                            ; call dw4 to read one sector
E060  250B                                    bcs     verify                                          ; if operation error, give message
E062  2609                                    bne     verify                                          ; if incompleted, give message
E064  E6E4                                    ldb     ,s                                                              ; get pushed B (opcode / command)
E066  C102                                    cmpb    #2                                                              ; is it read?
E068  261F                                    bne     jstvfy                                          ; no, adapt flags and exit
E06A  BDE20D                                  jsr     correct                                         ; correct LSN if needed
E06D  3574            verify  puls  b,x,y,u                                      ; restore received registers
E06F  3502                                    puls  a                                                         ; and first one pushed
E071  2506                                    bcs   failed                                            ; detect operation error
E073  2604                    bne   failed                                            ; detect incomplete read operation
E075  5F                                      clrb                                                                    ; operation succesful - no error code returned in B
E076  1CAF            outgo           andcc #eimask                                           ; enable interrupts
E078  39                                      rts                                                                     ; return to caller
E079  0DE7            failed   tst    <retry                                          ; is flag retry ON ($01)?
E07B  2706                                    beq     fail01                                          ; no, exit
E07D  2B04                                    bmi     fail01                                          ; if negative exit too
E07F  00E7                                    neg     <retry                                          ; to get value $FF
E081  2085                                    bra     entry                                                   ; try it again
E083  C6A6            fail01  ldb   #$a6                                                      ; set error code (unknown = ??)
E085  1A01                                    orcc    #%00000001                                      ; set Carry flag (error)
E087  20ED                                    bra   outgo                                                     ; return
E089  1A04            jstvfy  orcc    #%00000100                                      ; set flag Z=1          
E08B  20E0                                    bra     verify                                          ; exit via verify
                      ;********************************************************************************************************************************
                      ; This is used to calculate the LSN number from system variables $ec (track) and $ed (sector)
                      ; Controls the number of sides of the disks to change 18 s/t to 36
                      ; returns the calculated LSN in register Y
                      ;********************************************************************************************************************************
E08D  3416            calcy           pshs    d,x                                                     ; save working registers
E08F  8E06A6                                  ldx     #$6a6                                                   ; point to byte before sectors x track table
E092  D6EB                                    ldb     <$eb                                                    ; get drive number
E094  A685                                    lda     b,x                                                     ; get number of sectors x track for this drive
E096  D6EC            calcy01 ldb     <$ec                                                    ; get track number
E098  3D                                      mul                                                                     ; calculate sectors amount
E099  1F01                                    tfr     d,x                                                     ; X gets this value
E09B  96ED                                    lda     <$ed                                                    ; get sector number
E09D  4A                                      deca                                                                    ; decrement
E09E  3186                                    leay    a,x                                                     ; add to X and pass result to Y
E0A0  3596                                    puls    d,x,pc                                          ; restore registers and return
                      ;********************************************************************************************************************************
                      ; Aternative entry for DSKINIT that comes without data in $6a7-8-9-a. So this one relies on system variable $f4 (as numSides)
                      ;********************************************************************************************************************************
E0A2  3416            caldi           pshs    d,x                                                     ; save working registers
E0A4  8612                                    lda     #$12                                                    ; sectors x track (single side)
E0A6  0DF4                                    tst     <$f4                                                    ; is disk single sided?
E0A8  2701                                    beq     ecaldi                                          ; yes, skip next
E0AA  48                                      asla                                                                    ; double number of sectors x track
E0AB  20E9            ecaldi  bra     calcy01                                         ; jump to calculation routine
                      ;********************************************************************************************************************************
                      ; to apply index to a DW4 HD on drive 4 if it is the case
                      ;********************************************************************************************************************************
E0AD  3406            ctrlHD  pshs    a,b                                                     ; save register
E0AF  D6EB                                    ldb     <$eb                                                    ; get drive number
E0B1  C104                                    cmpb    #4                                                              ; is it 4?
E0B3  2613                                    bne     eCtrl                                                   ; no, exit
E0B5  F60115                                  ldb     idxHD                                                   ; get index of disc to be accesed
E0B8  C101                                    cmpb    #1                                                              ; is index 0 or 1?
E0BA  230C                                    bls     eCtrl                                                   ; yes, no offset
E0BC  5A                                      decb                                                                    ; else deduct one
E0BD  31A5                                    leay    b,y                                                     ; Y = LSN + index
E0BF  86F0                                    lda     #240                                                    ; maximum acceptable value that can be used to multiply the index
E0C1  3D                                      mul                                                                     ; D = 240 * Index  (if index=90, result 21360 -> $5370) always will be positive
E0C2  31AB                                    leay    d,y                                                     ; Y = LSN + 241*index
E0C4  31AB                                    leay    d,y                                                     ; Y = LSN + 481*index
E0C6  31AB                                    leay    d,y                                                     ; Y = LSN + 721*index  now we have calculated the real LSN to be used, no loops!
E0C8  3586            eCtrl           puls    a,b,pc                                          ; restore register and return
                      ;********************************************************************************************************************************
                      ; Process that substitutes the calls to "WriteTrack". It is used ONLY by DSKINIT
                      ; Writes the 18 sectors of one side in a loop
                      ; If sector number received is 0, will write to side 1, else ($ff) will do side 2 
                      ;********************************************************************************************************************************
E0CA  3401            inidsk  pshs    cc                                                              ; save flags
E0CC  BDE01D                                  jsr     phyvir                                          ; is drive physical or virtual?
E0CF  240B                                    bcc     virtual                                         ; virtual, process it
E0D1  4D                                      tsta                                                                    ; is linked value 0?
E0D2  2708                                    beq     virtual                                         ; yes, so DW4 slot 0
E0D4  3501                                    puls    cc                                                              ; get saved flags
E0D6  CC47F4                                  ldd   #$47f4                                            ; physical, do nothing. This was overwritten by the patch
E0D9  7EC328                                  jmp     $c328                                                   ; back to DOS
                                                                                                                      ; to avoid problems created by SOUND and PLAY over DW4
E0DC  0DEC            virtual tst     <$ec                                                    ; is track 0?
E0DE  260D                                    bne     iniD01                                          ; no, skip section
E0E0  3402                                    pshs    a                                                               ; save register
E0E2  8602                                    lda     #2                                                              ; get value to mark output
E0E4  B7FF20                                  sta     BBOUT                                                   ; to DW4
E0E7  4F                                      clra                                                                    ; set counter
E0E8  4A              iniL01  deca                                                                    ; decrement counter
E0E9  26FD                                    bne     iniL01                                          ; not zero, loopback
E0EB  3502                                    puls    a                                                               ; restore register
E0ED  3501            iniD01  puls    cc                                                              ; get saved flags
E0EF  1A50                                    orcc  #dimask                                           ; disable interrupts to work with dw4   
E0F1  4F                                      clra                                                                    ; value 0
E0F2  1F8B                                    tfr     a,dp                                                    ; set direct page to 0
E0F4                                          setdp   0                                                               ; notify compiler
E0F4  3476                                    pshs    d,x,y,u                                         ; save registers
E0F6  8D42                                    bsr     filbuf                                          ; fill buffer
E0F8  D6EC                                    ldb     <$ec                                                    ; get track number
E0FA  C114                                    cmpb    #$14                                                    ; is this the directory (T20 track)?
E0FC  2607                                    bne     wrt00                                                   ; no, init the track
E0FE  0DF3                                    tst     <$f3                                                    ; yes, is it side 2? (0 means side 1)
E100  2603                                    bne     wrt00                                                   ; init side 2!
                                                                                                                      ; yes, don't waste time, it will be filled later
E102  4F                                      clra                                                                    ; mark no error
E103  2025                                    bra     eIni01                                          ; exit
E105  4C              wrt00           inca                                                                    ; 1st track of side 1
E106  0DF3                                    tst     <$f3                                                    ; is it side 1?
E108  2702                                    beq     wrt01                                                   ; yes, skip next one
E10A  8613                                    lda     #19                                                     ; get 1st track of side 2
E10C  97ED            wrt01           sta     <$ed                                                    ; update system variable
E10E  8D92                                    bsr     caldi                                                   ; calculate first LSN to write to (into Y)
E110  8D9B                                    bsr     ctrlHD                                          ; control index of HD if needed
E112  3021                                    leax  1,y                               ; X points to the rigth dw4 sector (skipping the header) - [DW needs them inverted!)
E114  109EEE                                  ldy   <$ee                                              ; Y points to RAM buffer address
E117  C612                                    ldb     #18                                                     ; number of sectors to be initialized
E119  E7E2                                    stb     ,-s                                                     ; put counter in stack
E11B  BDE2C7          wrtmor  jsr     dowrit                                          ; write sector to DW4 
E11E  2510                                    bcs   failed2                                           ; detect operation error
E120  260E                    bne   failed2                                           ; detect incomplete read operation
E122  3001                                    leax    1,x                                                     ; point to next LSN
E124  6AE4                                    dec     ,s                                                              ; dec counter
E126  26F3                                    bne     wrtmor                                          ; if not yet 0, do next one
E128  3261                                    leas    1,s                                                     ; get rid of counter
E12A  3576            eIni01  puls    d,x,y,u                                    ; restore registers (0 will be pulled for A)
E12C  4D              eIniDsk tsta                                                                    ; update flags 
E12D  1CAF                                    andcc #eimask                                           ; enable interrupts
E12F  39                                      rts                                                                     ; return
E130  3261            failed2 leas    1,s                                                     ; get rid of counter
E132  3576                                    puls    d,x,y,u                                    ; restore registers
E134  86FF                                    lda     #$ff                                                    ; mark error for Dskinit function
E136  1A01                               orcc #$01                                                    ; set Carry flag (error)
E138  20F2                                    bra   eIniDsk                                           ; return to caller
                      ;********************************************************************************************************************************
                      ; Fill the buffer pointed by $ee all with value $e5
                      ;********************************************************************************************************************************
E13A  3454            filbuf  pshs    b,x,u                                                   ; save registers
E13C  9EEE                                    ldx     <$ee                                                    ; get buffer address
E13E  CEE5E5                                  ldu     #$e5e5                                          ; value to put
E141  8D32                                    bsr     cpu2xb                                          ; put 128 times U at X
E143  35D4                                    puls    b,x,u,pc                                                ; restore registers and return
                      ;********************************************************************************************************************************
                      ; Fill the directory track T20 for DW4 and floppy as well
                      ; Only used by DSKINIT, but to be sure, it controls a return address in the stack to work else returns
                      ;********************************************************************************************************************************
E145  3446            fildir  pshs    d,u                                                     ; save working registers
E147  CCC9EF                                  ldd     #$c9ef                                          ; return address when called to write directory tracks by DSKINIT
E14A  10A364                                  cmpd    4,s                                                     ; is it the write track call?
E14D  2607                                    bne     efildir                                         ; not the write track command, skip fill subroutines
E14F  8D0D                                    bsr     fil16s                                          ; fill buffer for the 16 Dir sectors
E151  8D2A                                    bsr     ffat1                                                   ; fill buffer for FAT 1st sector
E153  BDE1D2                                  jsr     ffat2                                                   ; fill buffer for FAT 2nd sector
E156  3546            efildir puls    d,u                                                     ; restore registers
E158  8E0800                                  ldx     #$800                                                   ; overwritten line by interceptor
E15B  7EC9F6                                  jmp     $c9f6                                                   ; back to DOS
                      ;********************************************************************************************************************************
                      ; Fill the buffer for the 16 directory sectors with 10 unused entries
                      ;********************************************************************************************************************************
E15E  8E0A00          fil16s  ldx     #$a00                                                   ; point to third buffer
E161  DE8A                                    ldu     <$8a                                                    ; get 16 bits zero
E163  8D10                                    bsr     cpu2xb                                          ; put 128 times U at X
E165  8689                                    lda     #$89                                                    ; code to mark not used entry
E167  8E0A00                                  ldx     #$a00                                                   ; point to the beginning again
E16A  A784            fild02  sta     ,x                                                              ; mark an entry
E16C  308819                                  leax    25,x                                                    ; point to next one
E16F  8C0AE8                                  cmpx    #$ae8                                                   ; done 10 entries?
E172  25F6                                    blo     fild02                                          ; no, loop back
E174  39                                      rts                                                                     ; return
                      ;********************************************************************************************************************************
                      ; Put value in U in bytes pointed by X, B times (each time a word)
                      ;********************************************************************************************************************************
E175  C680            cpu2xb  ldb     #128                                                    ; number of words to write
E177  EF81            cp01            stu     ,x++                                                    ; write one
E179  5A                                      decb                                                                    ; decrement counter
E17A  26FB                                    bne     cp01                                                    ; if not 0, loopback
E17C  39                                      rts                                                                     ; return
                      ;********************************************************************************************************************************
                      ; Fill the first FAT sector. Must mark as used only the 18 sectors of Tracks 20 and 16 if applies
                      ;********************************************************************************************************************************
E17D  8E0800          ffat1           ldx     #$800                                                   ; point to first buffer
E180  96F2                                    lda     <$f2                                                    ; get number of tracks  
E182  C612                                    ldb     #$12                                                    ; sectors x track (single side)
E184  0DF4                                    tst     <$f4                                                    ; is disk single sided?
E186  2701                                    beq     ffat100                                         ; yes, skip next
E188  58                                      aslb                                                                    ; double number of sectors x track
E189  DDFA            ffat100 std     <geodat                                         ; save geometry data for later use
E18B  C65A                                    ldb     #90                                                     ; there are 180 FAT bytes, so 90 words, but 180k discs need only HALF that!
E18D  DEFA                                    ldu     <geodat                                         ; get geometry data
E18F  11832812                                cmpu    #$2812                                          ; is it a Single Sided Single Density disk (180k)?
E193  2601                                    bne     ffat101                                         ; no, skip next
E195  56                                      rorb                                                                    ; yes, divide by 2 to get 45 words
E196  CEFFFF          ffat101 ldu     #$ffff                                          ; code to mark 16 free sectors
E199  8DDC                                    bsr     cp01                                                    ; copy B times U at X
E19B  3341                                    leau    1,u                                                     ; create 0 value
E19D  EF81            ffat102 stu     ,x++                                                    ; clear a word
E19F  8C08FC                                  cmpx    #$900-4                                         ; reached end of buffer minus 4 bytes to put geometry data there?
E1A2  25F9                                    blo     ffat102                                         ; no, loopback
E1A4  DCFA                                    ldd     <geodat                                         ; get geometry data
E1A6  ED81            ffat103 std     ,x++                                                    ; save it in 1st FAT sector
E1A8  43                                      coma                                                                    ; complement
E1A9  53                                      comb                                                                    ; both bytes
E1AA  ED84                                    std     ,x                                                              ; store again in FAT
E1AC  C624                                    ldb     #36                                                     ; This the FAT byte where dir T16 begins for single sided discs
E1AE  96FB                                    lda     <secxtrk                                                ; get sectors x track
E1B0  8112                                    cmpa    #$12                                                    ; is it single?
E1B2  2701                                    beq     ffat104                                         ; yes, skip next
E1B4  58                                      aslb                                                                    ; double to 72 (the same place for Double Side disks)
E1B5  3404            ffat104 pshs    b                                                               ; save value for later use
E1B7  86F0                                    lda     #$f0                                                    ; flag value for parameter '/' (leave T16 free for data storage)
E1B9  91E6                                    cmpa    <$e6                                                    ; has $e6 this value?
E1BB  2702                                    beq     ffat105                                         ; yes, don't mark sector T16 as used                                                            
E1BD  8D05                                    bsr     markfat                                         ; mark used sectors of T16
E1BF  D6FB            ffat105 ldb     <secxtrk                                                ; get sectors x track
                                                                                                                      ; distance between T16 and T20 is 4 * tracks * N_Sxt / 8 bits = NSxt / 2
E1C1  56                                      rorb                                                                    ; convert distance to FAT bytes to add to
E1C2  EBE0                                    addb    ,s+                                                     ; the previously pushed FAT byte for T16, clear stack
E1C4  8E0800          markfat ldx     #$800                                                   ; point to the begining of the FAT
E1C7  3085                                    leax    b,x                                                     ; add offset to 1st DIR sector byte of that track
E1C9  4F                                      clra                                                                    ; 16 bit zero
E1CA  5F                                      clrb                                                                    ; to mark 16
E1CB  ED81                                    std     ,x++                                                    ; used sectors by DIR track
E1CD  86FC                                    lda     #%11111100                                      ; and two zero bits more
E1CF  A784                                    sta     ,x                                                              ; to complete the 18 reserved sectors for any directory Track
E1D1  39                                      rts                                                                     ; return
                      ;********************************************************************************************************************************
                      ; Fill the second FAT sector. This is only used for 720k discs
                      ;********************************************************************************************************************************
E1D2  8E0900          ffat2           ldx     #$900                                                   ; point to second buffer
E1D5  DCFA                                    ldd     <geodat                                         ; get geometry data
E1D7  10835024                                cmpd    #$5024                                          ; is it a 720K disc?
E1DB  2607                                    bne     ffat201                                         ; no, do not put $FFs
E1DD  CEFFFF                                  ldu     #$ffff                                          ; to mark free sectors
E1E0  C65A                                    ldb     #90                                                     ; must mark 180 bytes
E1E2  8D93                                    bsr     cp01                                                    ; copy B times U at X
E1E4  DE8A            ffat201 ldu     <$8a                                                    ; get 16 bits zero
E1E6  EF81            ffat202 stu     ,x++                                                    ; put 2 zero bytes
E1E8  8C0A00                                  cmpx    #$a00                                                   ; reached end of buffer?
E1EB  25F9                                    blo     ffat202                                         ; no, loopback
E1ED  39                                      rts                                                                     ; return
                      ;********************************************************************************************************************************
                      ; Keep the good track number to avoid strange sequences. Called by DSKINIT and
                      ; called from other points, one of them the COPY routine and from SAVE as well ...
                      ; Has a control to detect that it is called from DSKINIT, else do nothing at all!
                      ; when applied to DW4 the track number suffers non lineal sequences. Here we correct that
                      ;********************************************************************************************************************************
E1EE  3406            modtrk  pshs    d                                                               ; save register
E1F0  CCC9E1                                  ldd     #$c9e1                                          ; the return address that identifies DSKINIT
E1F3  10A367                                  cmpd    7,s                                                     ; is the one corresponding to DSKINIT?
E1F6  2610                                    bne     notdsk                                          ; no, do nothing
E1F8  96EC                                    lda     <$ec                                                    ; get track number
E1FA  97F9                                    sta     <trknum                                         ; save in variable
E1FC  BDD36B                                  jsr     $d36b                                           ; (buf_read_sector) - this line was overwritten by patch
E1FF  96F9                                    lda     <trknum                                         ; get saved track number
E201  97EC                                    sta     <$ec                                                    ; back to its place
E203  3506            emodtrk puls    d                                                               ; restore register
E205  7ED1A7                                  jmp     $d1a7                                                   ; back to DOS
E208  BDD36B          notdsk  jsr     $d36b                                                   ; (buf_read_sector) - this line was overwritten by patch
E20B  20F6                                    bra     emodtrk                                         ; clean stack and back to DOS
                      ;********************************************************************************************************************************
                      ; Part to detect a bigger than 180k disc and update tables accordingly. The same if disc was changed
                      ; If read sector has no geometry data, we go to the other possible localization for directory ($168 - $2D0) (+1 for DW4)
                      ; X is updated to the new LSN to be read, if necessary
                      ; must be avoided ONLY when working with an HD at unit 4 (idxHD $115 >1)
                      ;********************************************************************************************************************************
E20D  3466            correct pshs    d,y,u                                                   ; save registers
E20F  96EB                                    lda     <$eb                                                    ; get unit number
E211  8104                                    cmpa    #4                                                              ; is it #4?
E213  2605                                    bne     cor00                                                   ; no, skip control
E215  7D0115                                  tst     idxHD                                                   ; working on a HD in a disk with index > 0?
E218  2639                                    bne     ecorok                                          ; yes, exit
E21A  DCEC            cor00           ldd     <$ec                                                    ; get track-sector numbers
E21C  10831401                                cmpd    #$1401                                          ; is it 1st FAT sector?
E220  2631                                    bne     ecorok                                          ; no, exit
E222  CCC869                                  ldd     #$c869                                          ; get return address from BACKUP command
E225  10A3E814                                cmpd    20,s                                                    ; is this the case?
E229  2728                                    beq     ecorok                                          ; yes, exit
E22B  DEEE                                    ldu     <$ee                                                    ; get read buffer address
E22D  ECC900FE                                ldd     $fe,u                                                   ; get complemented geodata bytes
E231  43                                      coma                                                                    ; complement both bytes
E232  53                                      comb                                                                    ; A=tracks; B=sectors x track ... possible values: 2812 - 2824 - 5012 - 5024
                                                                                                                              ; their 1st FAT sector will equal that one:        0169 - 02d1 - 0169 - 02d1 (+1 due to DW4) 
E233  10A3C900FC                              cmpd    $fc,u                                                   ; do they equal the not complemented ones?
E238  2717                                    beq     ecorok1                                         ; yes, we have a right FAT sector, update system tables and exit
                                                                                                                              ; Bad 1st FAT sector read, we must change single <--> double as required
E23A  8C0169                                  cmpx    #$0169                                          ; was sector read $0169 (single sided)
E23D  2605                                    bne     cor01                                                   ; no, change to single sided
E23F  8E02D1                                  ldx     #$02d1                                          ; yes, change to double sided
E242  2003                                    bra     cor02                                                   ; go save new 1st FAT LSN
E244  8E0169          cor01           ldx     #$0169                                          ; get single sided value
E247  8D10            cor02           bsr     updtab                                          ; update system tables (in some cases they are not yet filled)
E249  109EEE                                  ldy     <$ee                                                    ; get buffer address
E24C  BDFA0F                                  jsr     DoRead                                          ; read newly calculated T20 1st FAT LSN
E24F  2006                                    bra     ecor                                                    ; exit
E251  8D06            ecorok1 bsr     updtab                                          ; update system tables
E253  1A04            ecorok  orcc    #%00000100                                      ; set flag Z=1, result OK
E255  1CFE                    andcc   #%11111110                                      ; clear Carry flag, no errors
E257  35E6            ecor            puls    d,y,u,pc                                                ; restore registers and return                  
                      ;********************************************************************************************************************************
                      ; Updates system tables using param X as LSN for 1st DIR track (should get here with DW4 value)
                      ; Affects system tables at $61c (dir LSN) and $6a7 (sectors x track)
                      ;********************************************************************************************************************************
E259  CE0616          updtab  ldu     #$616                                                   ; point 6 bytes before drives info table                        
E25C  D6EB                                    ldb     <$eb                                                    ; get drive number
E25E  8606                                    lda     #6                                                              ; register length
E260  3D                                      mul                                                                     ; calculate offset
E261  311F                                    leay    -1,x                                                    ; system works with $0168 or $02d0 (DW4 adds one because of the header)
E263  10AFCB                                  sty     d,u                                                     ; put LSN at requested drive table
E266  8612                                    lda     #$12                                                    ; default value for single sided
E268  8C0169                                  cmpx    #$0169                                          ; is new value single sided?
E26B  2701                                    beq     cor03                                                   ; yes, skip next
E26D  48                                      asla                                                                    ; double value
E26E  CE06A6          cor03           ldu     #$6a6                                                   ; point 1 byte before sectors x track table
E271  D6EB                                    ldb     <$eb                                                    ; get drive number
E273  A7C5                                    sta     b,u                                                     ; update value in table for that drive
E275  39                                      rts                                                                     ; return
                      ;********************************************************************************************************************************
                      ; Updates table $6a6 if drive is virtual. Uses data from stack (only BACKUP)
                      ;********************************************************************************************************************************
E276  3416            patbck  pshs    d,x                                                     ; save registers
E278  8E00FB                                  ldx     #drvtab-1                                       ; point to 1 byte before the slots table
E27B  A6C4                                    lda     ,u                                                              ; get source drive from stack hole
E27D  8D0C                                    bsr     updsxt                                          ; update sectors x track
E27F  A649                                    lda     9,u                                                     ; get destination drive from stack hole
E281  8D08                                    bsr     updsxt                                          ; update sectors x track
E283  3516                                    puls    d,x                                                     ; restore registers
E285  A648                                    lda     8,u                                                     ; these 2 sentences were
E287  3D                                      mul                                                                     ; overwritten by patcher
E288  7EC839                                  jmp     $c839                                                   ; back to DOS
                      ;********************************************************************************************************************************
                      ; Updates sectors x track of that drive (comes in A)
                      ;********************************************************************************************************************************
E28B  E686            updsxt  ldb     a,x                                                     ; get associated slot number in drives table
E28D  2704                                    beq     upds01                                          ; if cero is a special DW4 case, process it
E28F  C104                                    cmpb    #4                                                              ; is it lower than 5?
E291  230B                                    bls     eupd                                                    ; yes, is a physicall drive, do not process it
E293  3410            upds01  pshs    x                                                               ; save register
E295  8E06A6                                  ldx     #$6a6                                                   ; point 1 byte before sectors x track table
E298  E648                                    ldb     8,u                                                     ; get number of sectors x track in stack hole
E29A  E786                                    stb     a,x                                                     ; into table for that drive
E29C  3510                                    puls    x                                                               ; restore register
E29E  39              eupd            rts                                                                     ; return
                      ;********************************************************************************************************************************
                      ; This part avoids calling a hardware routine when the drive is virtual
                      ; and when floppies allow it, in order to avoid that the FDC floppy light stays on forever
                      ;********************************************************************************************************************************
E29F  3403            noHdw           pshs    a,cc                                                    ; save A and flags
E2A1  BDE01D                                  jsr     phyvir                                          ; detect drive type
E2A4  240F                                    bcc     retHdw                                          ; if virtual, return
E2A6  4D                                      tsta                                                                    ; is linked value 0?
E2A7  270C                                    beq     retHdw                                          ; yes, so DW4 slot 0
E2A9  3503                                    puls    a,cc                                                    ; restore A and flags
                                                                                                                      ; so it goes for floppies or SDC
E2AB  1709EA                                  lbsr    flOrSd                                          ; will take care of destination hdw                     
E2AE  1A50                                    orcc  #dimask                                           ; first overwritten instruction
E2B0  66E4                                    ror     ,s                                                         ; second overwritten instruction
E2B2  7EC174                                  jmp     $c174                                                   ; back to DOS
E2B5  0DE7            retHdw  tst     <retry                                          ; is flag off (0)?
E2B7  2602                                    bne     ret01                                                   ; no, skip next
E2B9  0CE7                                    inc     <retry                                          ; turn flag ON ($01)
E2BB  3262            ret01           leas    2,s                                                     ; get rid of 2 pushed regs
E2BD  66E4                                    ror     ,s                                                              ; get rid of pushed carry value
E2BF  35FB                                    puls    a,cc,dp,x,y,u,pc                        ; restore registers and return
                      ;********************************************************************************************************************************
                      ; This intercepts the std DOS reset routine
                      ;********************************************************************************************************************************
E2C1  1702FD          dosReset lbsr   initask                                         ; put initial values
E2C4  7EB44F                                  jmp     wrmStrt                                         ; call basic warmstart
                      ;********************************************************************************************************************************
                      ; This part sends sectors to dw4 to be written. It is a block very similar to DoRead() but must work a bit different:
                      ; Must send three blocks of data: The dw4 command, the data and the checksum, all of them in just one chunk
                      ;********************************************************************************************************************************
E2C7  9612            dowrit  lda   <dnum                                                     ; get drive number (real dw4 number)
E2C9  5F                      clrb                                                     ; LSN bits 23-16
E2CA  3436                    pshs  a,b,x,y                            ; save registers / data                                        STACK: RETADRS, YL, YH, SectorL, SectorH, $0, dnum
E2CC  8E0000                  ldx     #0                                                              ; zero to calculate Checksum for that sector
E2CF  4F                      clra                                                                    ; byte counter (256)
E2D0  E6A0            calc    ldb     ,Y+                                                     ; read a byte
E2D2  3A                      abx                                                                     ; add to checksum
E2D3  4A                      deca                                                                    ; decrement counter
E2D4  26FA                    bne   calc                                                      ; if not all 256, keep on
E2D6  9F76                    stx   <chksum                                           ; save checksum
E2D8  8657                                    lda   #$57                      ; dw4 OP_WRITE (write one sector)
E2DA  3402            rewrit  pshs  a                                          ; save command onto stack                              STACK: RETADRS, YL, YH, SectorL, SectorH, $0, dnum, OP_WRITE
E2DC  30E4                    leax  ,s                                                ; point X to top of stack
E2DE  108E0005                ldy   #5                                ; 5 bytes
E2E2  BDFAE2                  jsr   DWWrite                           ; send 5 bytes to dw4:  OP_WRITE, dnum, $0, Sector(H-L)
E2E5  3502                    puls  a                                          ; discard command                                                      STACK: RETADRS, YL, YH, SectorL, SectorH, $0, dnum
E2E7  AE64                    ldx   4,s                                        ; get sector buffer pointer = received Y
E2E9  108E0100                ldy   #$0100                             ; to write 256 bytes
E2ED  BDFAE2                  jsr   DWWrite                           ; send sector (256 bytes)
E2F0  109E76                                  ldy   <chksum                                           ; get calculated checksum
E2F3  3420                    pshs  y                                          ; save it onto stack                                   STACK: RETADRS, YL, YH, SectorL, SectorH, $0, dnum, ChksumL, ChksumH
E2F5  30E4                    leax  ,s                                                ; point X to top of stack
E2F7  108E0002                ldy   #2                                                ; 2 bytes
E2FB  BDFAE2                  jsr   DWWrite                           ; send checksum to dw4
                                                                                                                      ; get answer from dw4 for the whole pack
E2FE  3121                                    leay  1,y                                        ; after a DWRITE Y returns zeroed, so this way it becomes equal to one
E300  BDFA5D                  jsr   DWRead                            ; read dw4 answer       - this overwrites dnum byte
E303  3262                    leas  2,s                                        ; discard checksum                                             STACK: RETADRS, YL, YH, SectorL, SectorH, $0, Returnedvalue
E305  2513                    bcs   writex                             ; detected framing error
E307  2611                    bne   writex                             ; detected not all bytes received
E309  A6E4                    lda   ,s                                                ; get server answer (in old dnum position)
E30B  270D                    beq   writex                             ; zero=OK, go out
E30D  81F3                    cmpa  #$F3                                       ; error type = #E_CRC?
E30F  2608                    bne   writer                             ; not this one, go out signaling error
E311  9612                                    lda   <dnum                           ; get drive number
E313  A7E4                    sta   ,s                                                ; put again in stack                                            STACK: RETADRS, YL, YH, SectorL, SectorH, $0, dnum
E315  8677                    lda   #$77                                                      ; dw4 code for OP_REWRITE
E317  20C1                    bra   rewrit                             ; go to rewrite same sector again
E319  53              writer  comb                                                     ; sets the carry flag (ERROR)
E31A  35B6            writex  puls  a,b,x,y,pc                ; restore received registers and return      STACK: - - -
                      ; ********************************************************************************************************************************
E31C  0D4457342653444320455854454E444552202856302E32352E3130290D psmsg           fcc     13,"DW4&SDC EXTENDER (V0.25.10)",13
E339  4259205045524520534552524154202832303230290D0D00                         fcn        "BY PERE SERRAT (2020)",13,13
                      ; --------------------------------------------------------------------------------------------------------------------------------                      
E351  204C4F4144494E4720524F4D0D00 ldtextr         fcn   / LOADING ROM/,13                                         ; initial message to the user
E35F  0D524F4D204C454E4754483A2000 endtxt  fcn   $0d, /ROM LENGTH: /                                       ; message to show rom length
E36D  0D524F4D20544F4F2042494700 bigrom  fcn   $0d, /ROM TOO BIG/                                        ; message for roms greater than 16k-256bytes
E37A  4E4F5420412044573420534C4F540D00 notdw4  fcn     /NOT A DW4 SLOT/,13                                     ; message for lrom without params and with default drive <> dw4
E38A  544F4B454E20544F4F204C4F570D00 low     fcn   /TOKEN TOO LOW/,13                                        ; message for not recognized token
E399  544F4B454E20544F4F20484947480D00 high    fcn   /TOKEN TOO HIGH/,13                                       ; message for not recognized token
E3A9  4452495645204D55535420455155414C20554E49540D00 drverr  fcn     /DRIVE MUST EQUAL UNIT/,13                      ; message for drive <> unit for physical drives
E3C0  4953204C494E4B454420544F20414E4F5448455220554E49540D00 alrput  fcn     /IS LINKED TO ANOTHER UNIT/,13  ; message for trying to use an already linked disc
E3DB  554E4954204C494E4B0D00 headlnk fcn     /UNIT LINK/,13                                                  ; header for LLINK
E3E6  2020202000      spaces  fcn     /    /                                                                  ; text for detail line in LLINK
E3EB  4C4F4144494E4720424F4F5420545241434B00 ldtext   fcn   /LOADING BOOT TRACK/                                     ; message for OS9boot
E3FE  0D4E4F5420414E204F5320424F4F5420545241434B00 noostxt  fcn   $0d, /NOT AN OS BOOT TRACK/              ; errro message for os9boot
E414  202300          xidxTxt fcn     / #/                                                                            ; text that precedes HD index number                    
E417  535452494E4720544F4F204C4F4E470D00 Str2Long        fcn     /STRING TOO LONG/,13                                    ; message for param string too long
E428  4E4F204D50492050524553454E540D00 NotAnMPI fcn    /NO MPI PRESENT/,13                                     ; message for no NPI and  ... next message too
E438  4E4F205344432050524553454E540D00 NotAnSDC        fcn     /NO SDC PRESENT/,13                                     ; message for no SDC present
E448                  NotActSDC
E448  4E4F20414354495645205344430D00                         fcn     /NO ACTIVE SDC/,13                                      ; mmesage for not selected SDC
E457                  BadFmtType
E457  464F524D41542056414C5545204E4F542056414C49440D00                         fcn     /FORMAT VALUE NOT VALID/,13             ; message for parameter invalid
E46F  494E44455820544F4F20484947480D00 idxHigh fcn     /INDEX TOO HIGH/,13                                     ; index higher than max allowed
E47F  43555252454E542044495220495320524F4F540D00 noCurDir        fcn   /CURRENT DIR IS ROOT/,13                  ; message to user
E494  4E4F5420504F535349424C450D00 canNot  fcn     /NOT POSSIBLE/,13                                               ; message for switching not possible
E4A2  3A2D2D4E4F20444953432D2D0D00 sdcNoDsc        fcn     /:--NO DISC--/,13                                               ; message for no disk in sdc drive
E4B0  3A544F54414C20534543544F52533A2000 totSect fcn     /:TOTAL SECTORS: /                                      ; text for detail in SDRIVE
E4C1  5344462000      datSDF  fcn     /SDF /                                                                  ; text for SDRIVE
E4C6  4449522000      datDIR  fcn     /DIR /                                                                  ; text for SDRIVE
E4CB  2020202000      datNUL  fcn     /    /                                                                  ; text for SDRIVE
E4D0  2020634F4E5400  msgCont fcn     /  cONT/                                                                        ; message to user CONTinue
E4D7  20206558495400  msgExit fcn     /  eXIT/                                                                        ; message to user EXIT
E4DE  494E56414C494420504152414D455445522056414C55450D00 wrongPar        fcn     /INVALID PARAMETER VALUE/,13            ; message
E4F7  535441434B20434F4E464C494354530D00 stackPrb        fcn     /STACK CONFLICTS/,13                                    ; error message 
E508  54494D454F55540D00 timeout         fcn     /TIMEOUT/,13                                                    ; error message for SDC access
E511                  targetInUse
E511  54415247455420494E205553450D00                         fcn     /TARGET IN USE/,13                                      ; error message for SDC
E520                  dirNotFound             
E520  4449524543544F5259204E4F5420464F554E440D00                         fcn     /DIRECTORY NOT FOUND/,13                        ; error message for SDC
E535                  pathNameInvalid         
E535  494E56414C494420504154484E414D450D00                         fcn     /INVALID PATHNAME/,13                           ; error message for SDC
E547                  miscHardware            
E547  4D4953432E204841524457415245204552524F520D00                         fcn     /MISC. HARDWARE ERROR/,13                       ; error message for SDC
E55D  554E4B4E4F574E204552524F520D00 unknown         fcn     /UNKNOWN ERROR/,13                                      ; error message for SDC
E56C                  targetNotFound  
E56C  544152474554204E4F5420464F554E440D00                         fcn     /TARGET NOT FOUND/,13                           ; error message for SDC
E57E  4E4F2036344B20434F4D50555445520D00 not64K  fcn     /NO 64K COMPUTER/,13                                    ; error message for SDC
                      
                      ; *******************************************************************************
                      ; Variables to control new added commands
                      ; *******************************************************************************
001A                  newtok  equ   $1a                               ; this works with dos cartridge plugged in!
0012                  tokens  equ   #18                               ; to be incremented each time words are added   = **************   NEW   COMMANDS   ***************** =
00B0                  usrptr  equ   $b0                               ; pointer to usr vector base
0134                  newstb  equ   $134                              ; base of new stub
                      ; *******************************************************************************
                      ; install new routines
                      ; Creates a new USR table and fills it with calls to 'FC ERROR'
                      ; usess the old USR table space to create a new STUB for the
                      ; new commands that will be added after the D.O.S. ones
                      ; *******************************************************************************
E58F  1A50            initext orcc    #dimask                                         ; disable interrupts
E591  308D1E55                                leax  newusr,pcr                        ; new address table for usr's
E595  9FB0                    stx   usrptr                            ; save at memory variable
E597  108E8B8D                ldy   #fcerr                            ; address for error message
E59B  860A                    lda   #10                               ; number of usr's
E59D  10AF81          nxtvec  sty   ,X++                              ; fill element table
E5A0  4A                      deca                                    ; decrement counter
E5A1  26FA                    bne   nxtvec                            ; not finished, go back
E5A3  8612                    lda   #tokens                           ; get number of new tokens
E5A5  B70134                  sta   newstb                            ; store at stub base byte
E5A8  308C67                  leax  newrds,pcr                        ; base of new reserved words table
E5AB  BF0135                  stx   newstb+1                          ; save on stub
E5AE  308C41                  leax  newdsp,pcr                        ; address of dispatching routine
E5B1  BF0137                  stx   newstb+3                          ; save on stub
E5B4  8EE31B                                  ldx     #psmsg-1                                                ; point to message string
E5B7  BD90E5                                  jsr     outstr                                          ; show to screen
E5BA  8D05                                    bsr     initask                                         ; put initial values
E5BC  1CAF                                    andcc   #eimask                                         ; enable interrupts
E5BE  7EF833                                  jmp     expert                                          ; goto auto-start experiments
                      
                      ; *******************************************************************************
E5C1  0FE7            initask clr     <retry                                          ; NO retry mark
E5C3  CC0102                                  ldd     #$0102                                          ; default values for the first two physical drives
E5C6  8E00FC                                  ldx     #$fc                                                    ; point to config area
E5C9  ED81                                    std     ,x++                                                    ; save these values
E5CB  CCF6FC                                  ldd     #$f6fc                                          ; default values for Units 3-4 pointing to DW4 slots
E5CE  ED84                                    std     ,x                                                              ; save these values
E5D0  DC8A                                    ldd     <$8a                                                    ; get 16 bits zero
E5D2  FD0114                                  std     floSDC                                          ; clear4 $114 and $115
                                                                                                                      ; pre-test on MPI                       
E5D5  1705E1                                  lbsr    tstMPI                                          ; is there an MPI?
E5D8  2604                                    bne     initsk01                                                ; no, go to individual checks
E5DA  17052B                                  lbsr    tstHARD                                         ; yes, verify connected hardware
E5DD  39                                      rts                                                                     ; return
                                                                                                                      ; else verify individual elements (only one active)
E5DE  1705F6          initsk01        lbsr    tstSDC                                          ; is SDC active?
E5E1  2709                                    beq     fndSDC                                          ; yes, skip next test
E5E3  170608                                  lbsr    tstFLO                                          ; is FDC active instead?
E5E6  2609                                    bne     fndNthg                                         ; no, go out, no harware present at all
E5E8  8604            fndFDC  lda     #%00000100                                      ; set FDC present and let both SDC drives OFF
E5EA  2002                                    bra     savHdwF                                         ; go save it
E5EC  860B            fndSDC  lda     #%00001011                                      ; mark SDC present and put both SDC drives ON
E5EE  B70114          savHdwF sta     floSDC                                          ; save result
E5F1  39              fndNthg rts                                                                     ; return
                      
                      ; *******************************************************************************
                      ; Command dispatcher called from BASIC
                      ; decodes the received token and passes it with the address
                      ; of the new dispatch table to the ROM command execution
                      ; *******************************************************************************
E5F2  811A            newdsp  cmpa  #newtok                           ; compare to first defined token
E5F4  250C                    blo   lowerr                            ; if lower show error
E5F6  812C                    cmpa  #newtok+tokens                    ; compare to last one
E5F8  240E                    bhs   higher                            ; if higher show error
E5FA  801A                    suba  #newtok                           ; first token converts to 0
E5FC  308C68                  leax  nwtb01,pcr                        ; get new commands dispatch table base address
E5FF  7E84ED                  jmp   romcmd                            ; pass to rom command execution
E602  308DFD84        lowerr  leax  low,pcr                           ; point to message
E606  2004                    bra   new1                                                      ; show error and return to interpreter
E608  308DFD8D        higher  leax  high,pcr                          ; point to message
E60C  BD90E5          new1    jsr   outstr                            ; show the selected error
E60F  7E89B4                  jmp   synerr                            ; rom will show syntax error message
                      ; *******************************************************************************
                      ; new reserved words
                      ; *******************************************************************************
E612  44574C4F41C4    newrds  fcc   /DWLOA/,$C4                       ; char 'D' OR $80               dwload  = DLOAD (the new into this ROM)
E618  4D4F554ED4              fcc   /MOUN/,$D4                        ; char 'T' OR $80               mounT           = Associate a VDK to a Drive Number
E61D  4C494ECB                fcc     /LIN/,$CB                                       ; char 'K' or $80               linK            = fill unit cell with received data
E621  4C4C494ECB              fcc     /LLIN/,$CB                                      ; char 'K' or $80               llinK           = show units links
E626  444FD3                  fcc     /DO/,$D3                                                ; char 'S' or $80               doS             = Boot a NitrOS-9 disc
E629  4C524FCD                fcc     /LRO/,$CD               ; char 'M' or $80               lroM            = Load or Run a ROM image
E62D  42414ECB                fcc     /BAN/,$CB                                       ; char 'K' or $80               banK            = change active SDC-bank
E631  534C4FD4                fcc     /SLO/,$D4                                       ; char 'T' OR $80               sloT            = change to another MPI slot ($ef)
E635  5344524956C5            fcc     /SDRIV/,$C5                                     ; char 'E' or $80               sdrivE  = multipurpose for SDC-card
E63B  534449D2                fcc     /SDI/,$D2                                       ; char 'R' or $80               sdiR            = multipurpose for SDC-card
E63F  48444944D8              fcc   /HDID/,$D8                                        ; char 'X' or $80               hdidX           = select the index disc in an HD array (max 90)
E644  534D4B4449D2            fcc     /SMKDI/,$D2                                     ; char 'R' or $80               smkdiR  = create a directory in the SDC-card
E64A  534B494CCC                         fcc  /SKIL/,$CC                                      ; char 'L' or $80               skilL           = kill a file or an empty directory in a SDC-card
E64F  535245CE                                fcc     /SRE/,$CE                                       ; char 'N' or $80               sreN            = rename a file in a SDC-card
E653  554E4C4F41C4                            fcc     /UNLOA/,$C4                     ; char 'D' OR $80               unloaD  = token to be used just as parameter ($f6)
E659  534348C4                                fcc     /SCH/,$C4                               ; char 'D' OR $80               schD            = to change ACTUAL Directory
E65D  57524954C5                              fcc     /WRIT/,$C5                                      ; char 'E' or $80               writE           = write to a flash bank
E662  53434F50D9                              fcc     /SCOP/,$D9                                      ; char 'Y' or $80    scopY    = copy ROM from bank or slot to RAM
                      ; *******************************************************************************               = **************   NEW   COMMANDS   ***************** =
                      ; Dispatch table. This is an indirect table
                      ; it contains the start address of every new command 
                      ; *******************************************************************************
E667  F882            nwtb01  fdb   dwload                                            ; routine for DWLOAD command
E669  E698                                    fdb   mount                                             ; routine for MOUNT command                
E66B  E706                                    fdb     link                                                    ; routine for LINK command
E66D  E785                                    fdb     llink                                                   ; routine for LLINK command
E66F  E7E2                                    fdb     os9boot                                         ; routine to boot NitrOS-9
E671  E847                                    fdb     xlrom                                                   ; routine to load - run a ROM image
E673  E9ED                                    fdb     xbank                                                   ; routine to switch SDC active bank
E675  EC0B                                    fdb     xslot                                                   ; routine to change of MPI slot
E677  ECFB                                    fdb     xsdrive                                         ; routine for SDC-card DRIVE functions
E679  F2A3                                    fdb     xsdir                                                   ; routine for SDC-card DIR functions
E67B  ECE1                                    fdb     xHDidx                                          ; routine to select index of HD
E67D  F1A0                                    fdb     xsmkdir                                         ; routine for SMKDIR command
E67F  F1CA                                    fdb     xskill                                          ; routine for SDELETE command
E681  F1D1                                    fdb     xsren                                                   ; routine for SRENAME command
E683  E76C                                    fdb     snError                                         ; simply show syntax error
E685  F268                                    fdb     xschd                                                   ; routine for SCHD command
E687  F449                                    fdb     xwrite                                          ; routine for WRITE command
E689  F60F                                    fdb     xscopy                                          ; routine for SCOPY command
                      ; *******************************************************************************               = **************   NEW   COMMANDS   ***************** =
                      ;  new USR table
                      ;  This is a reserved RAM area to reallocate the USR vector table 
                      ;  that is being used for the new STUB we just created
                      ;  CAREFULL, IT MUST RESIDE IN RAM!!
                      ; *******************************************************************************
03EA                  newusr  equ     $3ea                                                    ; will use $3ea - $3fd = 20 bytes
                      ; ===============================================================================
                      ; Dispatch routines for the new added commands
                      ; ===============================================================================
                      ; *******************************************************************************
                      ; Routine to mark as invalid the four I/O buffers
                      ; *******************************************************************************
E68B  4F              mrkmod  clra                                                                    ; get a 0
E68C  8E0636                                  ldx     #$636                                                   ; point to 1st I/O buffer flags data
E68F  C615                                    ldb     #21                                                     ; offset to last I/O
E691  A785            mo01            sta     b,x                                                     ; mark as invalid
E693  C007                                    subb    #7                                                              ; offset to previous I/O
E695  2AFA                                    bpl     mo01                                                    ; if not negative, loopback
E697  39                                      rts                                                                     ; return
                      ; *******************************************************************************
                      ; Command to ask DW4 to mount a VDK file and get the slot number back
                      ; *******************************************************************************
E698  1700BC          mount           lbsr    readunit                                                ; read unit number
E69B  D7F8                                    stb     <numUni                                         ; save it as number of Unit
E69D  BD89AA                                  jsr     ckComma                                         ; CkComma
E6A0  BDB7AA                                  jsr     getfnam                                         ; GetNam into $1d1 (nambuf)
E6A3  8E01D0                  ldx   #nambuf-1                                 ; byte before packet buffer start
E6A6  6F84                    clr   ,x                                                                ; zero MSB to make a 16 bit value for Y
E6A8  10AE81                  ldy   ,x++                                                      ; length of file name (16 bit)
E6AB  2756                    beq   failvdk                                           ; if len=0 no file entered
E6AD  E61F                    ldb   -1,x                                                      ; length of name (8 bit)
                                                                                                                      ; Adding .VDK if not termination entered, costs 31 bytes       
E6AF  3414                                    pshs  b,x                                                       ; save pointer to beginning of name and name length
E6B1  A680            isdot   lda   ,x+                                                       ; get a char from name
E6B3  812E                    cmpa  #'.'                                                      ; is it a dot?
E6B5  2718                    beq   alrdot                                            ; yes, stop search, quit loop
E6B7  5A                      decb                                                                    ; decrement counter
E6B8  26F7                    bne   isdot                                                     ; not end of string, get next char name
E6BA  CC2E56                  ldd   #$2e56                                            ; get chars ".V"
E6BD  ED81                    std   ,x++                                                      ; add to end of name
E6BF  CC444B                  ldd   #$444b                                            ; get chars "DK"
E6C2  ED81                    std   ,x++                                                      ; add to end of name
E6C4  E6E4                                    ldb   ,s                                                                ; get old name length
E6C6  CB04                                    addb  #4                                                                ; add 4 (because of .VDK added)
E6C8  E7E4                                    stb   ,s                                                                ; update new length in stack
E6CA  F701D1                  stb   nambuf                                            ; save new length in byte before string
E6CD  3124                    leay  4,y                                                       ; correction to name length - DW parameter
E6CF  3402            alrdot  pshs    a                                                               ; save register
E6D1  8602                                    lda     #2                                                              ; get value to mark output
E6D3  B7FF20                                  sta     BBOUT                                                   ; to DW4
E6D6  4F                                      clra                                                                    ; set counter
E6D7  4A              mount01 deca                                                                    ; decrement counter
E6D8  26FD                                    bne     mount01                                         ; not zero, loopback
E6DA  3502                                    puls    a                                                               ; restore register
E6DC  3514                                    puls  b,x                                               ; restore pointer and length
E6DE  6F85                                    clr   b,x                                                       ; zero terminate name string (for error)
E6E0  6C83                    inc   ,--x                                                      ; 1 = DriveWire OP_NAMEOBJ_MOUNT command
E6E2  3122                    leay  2,y                                                       ; length of DW packet = name length + 2
E6E4  1A50                    orcc    #dimask                                         ; disable interrupts to work with DW4
E6E6  BDFAE2                  jsr   DWWrite                                           ; send file request to server
E6E9  8E0012                  ldx   #dnum                                                     ; our drive number variable
E6EC  6F84                    clr   ,x                                                                ; clear it
E6EE  3121                    leay  1,y                                                       ; read one byte back
E6F0  BDFA5D                  jsr   DWRead                                            ; get drive number
E6F3  1CAF                                    andcc #eimask                                           ; enable interrupts again 
E6F5  6D84                    tst   ,x                                                                ; verify received value
E6F7  2605                    bne   vdkok                                                     ; successful mount                                       
E6F9  C619                    ldb   #$19                                                      ; code for MO ERROR
E6FB  7E8344                                  jmp   syserr                                            ; show error message (syserr)
E6FE  D612            vdkok           ldb     <dnum                                                   ; pass slot to B
E700  8D25                                    bsr     stval                                                   ; verify / store new value
E702  39                                      rts                                                                     ; return
E703  7EB84B          failvdk jmp   ioerror                                           ; I/O error message
                      ; *******************************************************************************
                      ; Command to link a Unit with a drive number or a DW4 slot number
                      ; *******************************************************************************
E706  8D4F            link            bsr     readunit                                                ; read unit number
E708  D7F8                                    stb     <numUni                                         ; save it as number of Unit
E70A  8D67                                    bsr     opt8bit                                         ; search another param 8 bit number
E70C  260E                                    bne     link01                                          ; if greater than 0, goto more tests
                                                                                                                      ; requested slot #0
E70E  D6F8                                    ldb     <numUni                                         ; get unit number
E710  C101                                    cmpb    #1                                                              ; is it 1?
E712  2604                                    bne     notalw                                          ; no, error. Only unit 1 is allowed to be linked to slot #0
E714  0FFC                                    clr     <$fc                                                    ; put 0 value in first cell (unit 1, value 0)
E716  200E                                    bra     link03                                          ; exit
E718  C608            notalw  ldb     #$08                                                    ; code for FC error
E71A  2047                                    bra     basyserr                                                ; show error message
E71C  C104            link01  cmpb    #4                                                              ; is new value greater than 4 (DW4)?
E71E  2204                                    bhi     link02                                          ; yes, skip test
E720  D1F8                                    cmpb    <numUni                                         ; equals new drive number the unit number? (mandatory!)
E722  262D                                    bne     numError                                                ; no, show unit error
E724  8D01            link02  bsr     stval                                                   ; verify and store new value
E726  39              link03  rts                                                                     ; return
                      ; *******************************************************************************
                      ; Verifies that the new value (B) is not linked to another unit. If good, saves it
                      ; *******************************************************************************
E727  8E00FB          stval           ldx     #$00fb                                          ; point one byte before drive/slot table
E72A  8604                                    lda     #4                                                              ; counter (number of units)
E72C  E186            stv02           cmpb    a,x                                                     ; equals new value the contents of that slot?
E72E  2715                                    beq     stv04                                                   ; yes, see if it is the same unit
E730  4A                                      deca                                                                    ; decrement counter
E731  26F9                                    bne     stv02                                                   ; if not 0, loopback
E733  8E00FB          estval  ldx     #$00fb                                          ; point to one byte before data table
E736  96F8                                    lda     <numUni                                         ; get unit number
E738  8104                                    cmpa    #4                                                              ; is unit 4?
E73A  2603                                    bne     stv03                                                   ; no, skip next
E73C  7F0115                                  clr     idxHD                                                   ; set index to HD to 0
E73F  E786            stv03           stb     a,x                                                     ; store slot into table
E741  17FF47                                  lbsr    mrkmod                                          ; modify data to flag drive changes to system   
E744  39                                      rts                                                                     ; everything alright, return
E745  91F8            stv04           cmpa    <numUni                                         ; is unit with this value the same as destination in LINK command?
E747  27EA                                    beq     estval                                          ; yes, acceptable, return
E749  3262                                    leas    2,s                                                     ; get rid of return address
E74B  2000                                    bra     already                                         ; show error. We do not admit the same slot number in two units!
                      ; *******************************************************************************
E74D  8EE3BF          already ldx     #alrput-1                                       ; point to error message
E750  10                                      fcb     #$10                                                    ; to jump over next one
E751  8EE3A8          numError ldx    #drverr-1                                       ; point to error message
E754  7E90E5                                  jmp     outstr                                          ; show message and return to interpreter
                      ; *******************************************************************************
E757  0FF8            readunit        clr     <numUni                                         ; clear unit number
E759  8D0D                                    bsr     getunit                                         ; get a one digit number
E75B  2704                                    beq     errNum                                          ; if zero jump to device error
E75D  C104                                    cmpb    #4                                                              ; is it greater than 4?
E75F  2305                                    bls     OKnext                                          ; no, goto OK
E761  C628            errNum  ldb     #$28                                                    ; device number error (bas_err_DN)
E763  7E8344          basyserr        jmp     syserr                                          ; bas_system_error
E766  0EA5            OKnext  jmp     <redLChr                                                ; bas_poll_char                 
                      ; *******************************************************************************
E768  9DA5            getunit jsr     <redLChr                                                ; get a char (bas_poll_char)
E76A  2603                                    bne     r8b01                                                   ; if there is something, skip next
E76C  7E89B4          snError jmp     synerr                                          ; show error message (BAS-SN-Error)
E76F  BDE77D          r8b01           jsr     get8bit                                         ; get a number (8bits)
E772  39                                      rts                                                                     ; return
                      ; *******************************************************************************
E773  9DA5            opt8bit jsr     <redLChr                                                ; get a char (bas_poll_char)
E775  2705                                    beq     eo8b                                                    ; if 0, return
E777  BD89AA                                  jsr     ckComma                                         ; check if there is a comma separator (CkComa)
E77A  8D01                                    bsr     get8bit                                         ; get a 8 bits number
E77C  39              eo8b            rts                                                                     ; return
                      ; *******************************************************************************
E77D  3460            get8bit pshs    y,u                                                     ; save working registers
E77F  BD8E51                                  jsr     asc2Num                                         ; convert ASCII to number (1 byte)
E782  5D                                      tstb                                                                    ; update flags upon result
E783  35E0                                    puls    y,u,pc                                          ; restore registers and return
                      ; *******************************************************************************
                      ; Command to show the links of every unit
                      ; *******************************************************************************
E785  8EE3DA          llink           ldx     #headlnk-1                                      ; point to message header
E788  BD90E5                                  jsr     outstr                                          ; show it
E78B  8E00FB                                  ldx     #$00fb                                          ; point one byte before first data byte
E78E  C601                                    ldb     #1                                                              ; unit counter
E790  A685            llink01 lda     b,x                                                     ; get linked value
E792  8D06                                    bsr     showLine                                                ; show data
E794  5C                                      incb                                                                    ; increment counter
E795  C104                                    cmpb    #4                                                              ; done all units?
E797  23F7                                    bls     llink01                                         ; no, loopback
E799  39                                      rts                                                                     ; return to caller
                      ; *******************************************************************************
E79A  3416            showLine        pshs    d,x                                                     ; save working registers
E79C  8EE3E7                                  ldx     #spaces+1                                       ; point to print 2 spaces
E79F  BD90E5                                  jsr     outstr                                          ; print them
E7A2  E661                                    ldb     1,s                                                     ; get unit number
E7A4  4F                                      clra                                                                    ; clear high byte
E7A5  BD957A                                  jsr     outnum                                          ; print number
E7A8  E6E4                                    ldb     ,s                                                              ; get linked value
E7AA  4F                                      clra                                                                    ; clear high byte
E7AB  C10A                                    cmpb    #10                                                     ; is value lower than 10?
E7AD  2506                                    blo     sL01                                                    ; yes, goto print
E7AF  4C                                      inca                                                                    ; one space less
E7B0  C164                                    cmpb    #100                                                    ; is value lower than 100?
E7B2  2501                                    blo     sL01                                                    ; yes, goto print
E7B4  4C                                      inca                                                                    ; one space less
E7B5  8EE3E5          sL01            ldx     #spaces-1                                       ; point to print 4 spaces
E7B8  3086                                    leax    a,x                                                     ; skip the calculated ones upon the value
E7BA  BD90E5                                  jsr     outstr                                          ; print required spaces
E7BD  E6E4                                    ldb     ,s                                                              ; get linked value
E7BF  4F                                      clra                                                                    ; clear high byte                                                       
E7C0  BD957A                                  jsr     outnum                                          ; print number
E7C3  E661                                    ldb     1,s                                                     ; get unit number
E7C5  C104                                    cmpb    #4                                                              ; is it 4?
E7C7  2612                                    bne     sL02                                                    ; no, stop printing
E7C9  F60115                                  ldb     idxHD                                                   ; get index for HD
E7CC  270D                                    beq     sL02                                                    ; if zero, skip printing it
E7CE  8EE413                                  ldx     #xidxTxt-1                                      ; point to text
E7D1  BD90E5                                  jsr     outstr                                          ; show it
E7D4  F60115                                  ldb     idxHD                                                   ; get unit number
E7D7  4F                                      clra                                                                    ; zero to high byte
E7D8  BD957A                                  jsr     outnum                                          ; call ROM to print integer in D
E7DB  860D            sL02            lda     #$0d                                                    ; enter
E7DD  BDBCAB                                  jsr     outRegA                                         ; print A 
E7E0  3596                                    puls    d,x,pc                                          ; restore registers and return
                                              
                      ; *******************************************************************************
                      ; Command to Boot a NitrOS-9 disc
                      ; *******************************************************************************
E7E2  5F              os9boot  clrb                                                                   ; default drive 0
E7E3  9DA5                     jsr   <redLChr                 ; peek at next char
E7E5  2703                     beq   defnum                                           ; not found, use default
E7E7  BD8E51                   jsr   asc2Num                                  ; Get 1 byte
E7EA  D712            defnum  stb   <dnum                                                     ; save drive number
E7EC  8EE3EA                                  ldx   #ldtext-1                                 ; point to 'loading' message
E7EF  BD90E5                   jsr   outstr                                           ; show it
E7F2  1A50                                    orcc  #dimask                                           ; disable interrupts
E7F4  BDE835                                  jsr     trick                                                   ; make sure DW4 connection is on
                                                                                                                              ; DriveWire bootstrap code. Get sectors 612-629 to $2600
E7F7  8E0264                        ldx   #612                                                        ; starting sector number to load
E7FA  108E2600                 ldy   #$2600                                           ; destination in memory
E7FE  BDFA0F          DOSLoop  jsr   DoRead                                           ; read sector (into Y)
E801  252F                     bcs   fail2                                                    ; start all over
E803  262D                     bne   fail2                                                    ; show error
E805  8C0264                   cmpx  #612                                                     ; our first sector?
E808  2608                     bne   not1st                                           ; branch if not
E80A  ECA4                     ldd   ,y                                                               ; else get 1st two bytes at $2600
E80C  10834F53                 cmpd  #$4f53                                           ; OS?
E810  261A                     bne   noos                                                     ; no, show error
E812  866A            not1st   lda   #'*'+$40                                         ; print star
E814  31A90100                 leay  256,y                                                    ; move for next sector
E818  3001                     leax  1,x                                                      ; increment number of sector to be read
E81A  A789023B                 sta   571,x                                                    ; print progress (1024+32*5-612-1)
E81E  8C0276                   cmpx  #630                                                     ; are we at last sector?
E821  2DDB                     blt   DOSLoop                                          ; branch if not
E823  8661                     lda   #'!'+$40                                         ; print bang
E825  A789027B                 sta   635,x                                                    ; over last '*' (636-1)
E829  7E2602                   jmp   $2602                                                    ; jump into Boot code
E82C  8EE3FD          noos     ldx   #noostxt-1                                       ; point to message
E82F  BD90E5                   jsr   outstr                                           ; print string at x+1
E832  7EB84B          fail2    jmp   ioerror                                          ; IO_error equ $b84b
                      ; *******************************************************************************
                      ; Routine that marks on the DW4 line for a while, to avoid fails at 1st access
                      ; *******************************************************************************
E835  3414            trick    pshs   b,x                                                     ; save working registers
E837  8EFF20                                  ldx   #PIA1Base                         ; address of PIA1
E83A  C602                     ldb   #$02                                                     ; put value #2 (bit1 to high level)
E83C  E784                     stb   ,x                               ; make RS-232 output marking
                                                                                                                              ; Spin for a while so that the RS-232 bit stays hi for a time
E83E  8EA000          waw             ldx   #$a000                                            ; wait counter
E841  301F            waw1     leax  -1,x                                                     ; decrement counter
E843  26FC                     bne   waw1                                                     ; if not 0, loopback
E845  3594                     puls   b,x,pc                                          ; restore registers and return
                      ; *******************************************************************************
                      ; Command to load - run a ROM image
                      ; *******************************************************************************
0076                  endadr  equ   $76                                                       ; last loaded address - probably no conflict but collides with  chksum  equ   $76       
0089                  curlow  equ   $89                                                       ; low byte of cursor position
00E6                  bigfil  equ   $e6                                                       ; too big file flag - probably no conflict but collides with internal use of $e6 in format by dplus
015E                  hokini  equ   $15e                                                      ; first system hook
01A8                  hokend  equ   $1a8                                                      ; last system hook
3000                  ramini  equ   $3000                                                     ; where to copy the rom
6EFF                  rammax  equ   $6eff                                                     ; ram limit to send to highram (16k - 256 bytes) as a rom program to be run
6FFF                  rmaxld  equ   $6fff                                                     ; ram limit to load into low memory (for basic roms)
                      
E847  1A50            xlrom           orcc  #dimask                   ; disable firq, irq
E849  8DEA                                    bsr     trick                                                   ; make sure dw4 connection is on
E84B  9DA5            lromrty         jsr   <redLChr                                          ; peek at next character (after "rom")
E84D  804E                    suba  #'N'                                                      ; to detect if it was an 'n' - if so don't exec it after loading
E84F  9710                    sta   <noexec                                           ; save difference. 0 = noexec
E851  2602                    bne   savvec                                            ; if not 'n' jump next instruccion
E853  9D9F                    jsr   <getNChr                                          ; jump over that char (the 'n')
E855  8EE350          savvec  ldx   #ldtextr-1                                        ; byte before message
E858  8D76                                    bsr   prtstr                                            ; print string
E85A  BDB7AA                          jsr   getfnam                                           ; read rom file name and length into $1d1 = nambuf
E85D  7D01D1                  tst     nambuf                                          ; is filename length equal 0?
E860  261D                    bne     rom01                                                   ; no, process it
E862  F6060A                  ldb     $60a                                                    ; get default drive number
E865  8E00FB                  ldx     #$00fb                                          ; point to table of links x unit
E868  A685                    lda     b,x                                                     ; search slot number in table
E86A  2604                    bne     ver01                                                   ; if not 0, do more tests
E86C  C101                    cmpb    #1                                                              ; is slot 0 linked to unit 1? - to avoid tricks with poke unit's slot numbers
E86E  2704                    beq     ver02                                                   ; yes, accept it and go on. the only legal one!
E870  8105            ver01           cmpa    #5                                                              ; is it a dw4 slot?
E872  2504                    blo     errdw4                                          ; no, show error
E874  9712            ver02           sta     <dnum                                                   ; save slot number at variable
E876  2064                    bra     rdecb                                                   ; go process rom at default drive
E878  8EE379          errdw4  ldx     #notdw4-1                                       ; point to message error
E87B  BD90E5                                  jsr   outstr                                            ; show message
E87E  39                                      rts                                                                     ; return
E87F  8E01D0          rom01           ldx   #nambuf-1                                 ; byte before packet buffer start
E882  6F84                    clr   ,x                                                                ; zero msb to make a 16bit value for y
E884  10AE81                  ldy   ,x++                                                      ; length of file name (16 bit)
E887  2742                    beq   failedr                                           ; if len=0 no file entered
E889  E61F                    ldb   -1,x                                                      ; length of name (8 bit)
                                                                                                                              ; adding .rom if not termination entered      
E88B  3414                                    pshs  b,x                                                       ; save pointer to beginning of name and name length
E88D  A680            isdotr          lda   ,x+                                                       ; get a char from name
E88F  812E                    cmpa  #'.'                                                      ; is it a dot?
E891  2718                    beq   alrdotr                                           ; yes, stop search, quit loop
E893  5A                      decb                                                                    ; decrement counter
E894  26F7                    bne   isdotr                                            ; not end of string, get next char name
E896  CC2E52                  ldd   #$2e52                                            ; get chars ".r"
E899  ED81                    std   ,X++                                                      ; add to end of name
E89B  CC4F4D                  ldd   #$4f4d                                            ; get chars "om"
E89E  ED81                    std   ,x++                                                      ; add to end of name
E8A0  E6E4                                    ldb   ,s                                                                ; get old name length
E8A2  CB04                                    addb  #4                                                                ; add 4 (because of .rom added)
E8A4  E7E4                                    stb   ,s                                                                ; update new length in stack
E8A6  F701D1                  stb   nambuf                                            ; save new length in byte before string
E8A9  3124                    leay  4,y                       ; correction to name length - dw parameter
E8AB  3514            alrdotr         puls  b,x                                               ; restore pointer and length
E8AD  6F85                    clr   b,x                                                       ; zero terminate name string (for error)
E8AF  6C83                    inc   ,--x                                                      ; 1 = drivewire op_nameobj_mount command
E8B1  3122                    leay  2,y                                                       ; length of dw packet = name length + 2
E8B3  BDFAE2                  jsr   DWWrite                                           ; send file request to server
E8B6  8E0012                  ldx   #dnum                                                     ; our drive number variable
E8B9  6F84                    clr   ,x                                                                ; clear it
E8BB  3121                    leay  1,y                                                       ; read one byte back
E8BD  BDFA5D                  jsr   DWRead                                            ; get drive number
E8C0  6D84                    tst   ,x                                                                ; verify received value
E8C2  2618                    bne   rdecb                                                     ; successful mount. if z has not been set = error (coundn't mount). jump to readin
E8C4  C619            moError         ldb   #$19                                                      ; errror code for mo error
E8C6  1CAF                    andcc #eimask                                   ; enable interrupts
E8C8  7E8344                  jmp   syserr                                            ; show error message
E8CB  1CAF            failedr         andcc #eimask                                   ; enable interrupts
E8CD  7EB84B                                  jmp   ioerror                                           ; i/o errror message
E8D0  D689            prtstr  ldb     <curlow                                                 ; get low byte of cursor position
E8D2  C41F                    andb  #$1f                                                      ; is column zero?
E8D4  2602                    bne   shwmsg                                    ; no, print whith leading enter
E8D6  3001                                    leax  1,x                                       ; jump leading enter 
E8D8  BD90E5          shwmsg  jsr   outstr                                       ; print string
E8DB  39                                      rts                                                                     ; return
E8DC  8E0000          rdecb   ldx   #0                                                                ; begin at sector zero
E8DF  108E3000                ldy   #ramini                                           ; destination address
E8E3  9F76                    stx   <endadr                                           ; set last load address to zero 
E8E5  0FE6                    clr   <bigfil                                           ; flag for long file to false
E8E7  1700AB          romlop  lbsr  doreadr                                   ; read sector (into y) 
E8EA  25DF                    bcs   failedr                                           ; detect frame error
E8EC  26DD                    bne   failedr                                           ; detect incomplete read
E8EE  2D15                    blt   nomore                    ; detect end of file. finish
E8F0  862A                    lda   #'*'                                                      ; get 'star' ascii code
E8F2  BDB54A                                  jsr   outchr                                            ; show on screen, we have read a sector
E8F5  31A90100                leay  256,y                                                     ; point to next sector address
E8F9  108C6FFF                cmpy  #rmaxld                                           ; got to the ram limit?
E8FD  2404                    bhs   bigend                                            ; no, keep on copying
E8FF  3001                    leax  1,x                                                       ; increment sector number
E901  20E4                    bra   romlop                                            ; go read next sector
E903  0CE6            bigend  inc   <bigfil                                           ; big file flag to true
E905  109F76          nomore  sty   <endadr                                           ; save as last loaded address
E908  8608                    lda   #$08                                                      ; get backspace ascii code
E90A  BDB54A                                  jsr      outchr                                         ; print on screen (go back one char)
E90D  8621                    lda   #'!'                                                      ; get 'bang' ascii code
E90F  BDB54A                                  jsr   outchr                                            ; put it on screen
E912  0DE6                                    tst   <bigfil                                           ; is this a too long file?
E914  2705                                    beq   final                                                     ; no, jump message
E916  8EE36C                                  ldx   #bigrom-1                                 ; message for rom greater than 16k-256bytes
E919  8DB5                                    bsr   prtstr                                            ; print message
E91B  0D10            final   tst   <noexec                                           ; has to be run?
E91D  2610                    bne   runit                                                     ; yes, do it
E91F  8EE35E                  ldx   #endtxt-1                                 ; point to end text
E922  8DAC                                    bsr   prtstr                                            ; print message
E924  DC76                                    ldd   <endadr                                           ; get last loaded ram address
E926  833000                                  subd  #ramini                                           ; subtract init ram address
E929  BD957A                                  jsr   outnum                                            ; print length 
E92C  1CAF                            andcc #eimask                                           ; enable interrupts
E92E  39                      rts                                                                     ; return to the caller
E92F  FCC000          runit   ldd   >romini                                           ; get two bytes from rom area
E932  8E9669                                  ldx   #$9669                                            ; create a pattern 1001 0110 0110 1001
E935  BFC000                                  stx   >romini                                           ; store into two dos-rom bytes
E938  BCC000                                  cmpx  >romini                                           ; have been modified?
E93B  2603                                    bne   basicup                                           ; no, must switch to map1 and copy basic to upper ram
E93D  FDC000                                  std   >romini                                           ; put back two saved bytes
E940  8EE952          basicup  ldx    #basic                                          ; point to beginning of code to be moved
E943  CE2A00                                  ldu     #$2a00                                          ; destination in low ram
E946  EC81            basup01 ldd     ,x++                                                    ; get a word
E948  EDC1                                    std     ,u++                                                    ; put at destination
E94A  8CE995                                  cmpx    #doreadr                                                ; got to the end?
E94D  23F7                                    bls     basup01                                         ; no, loopback
E94F  7E2A00                                  jmp     $2a00                                                   ; transfer control to copied code in ram
                      ; -------------------------------------------------------------------------------------------
E952  8E8000          basic   ldx   #basini                                           ; get basic initial address
E955  B7FFDE          bas1    sta   $ffde                                                     ; go map0 (rom)
E958  EC84                    ldd   ,x                                                                ; get 2 bytes
E95A  B7FFDF                  sta   $ffdf                                                     ; go map1 (ram)
E95D  ED81                    std   ,x++                                                      ; store 2 bytes, incr. pointer
E95F  8CC000                  cmpx  #romini                                           ; reached end of basic?
E962  25F1                    blo   bas1                                                      ; no, keep on copying basic
E964  8E3000          map1            ldx     #ramini                                         ; get rom origin address at ram area
E967  CEC000                  ldu   #romini                                           ; get rom destination address
E96A  B7FFDF                  sta   $ffdf                                                     ; go map1 (ram) and stay there
E96D  EC81            move    ldd   ,x++                                                      ; get two bytes, inc pointer
E96F  EDC1                    std   ,u++                                                      ; store two bytes, inc pointer
E971  9C76                    cmpx  <endadr                                           ; reached last loaded address?
E973  2405                                    bhs   nohook                                            ; yes, quit loop
E975  8C6EFF                                  cmpx  #rammax                                           ; reached max rom possible (16k-256bytes)?
E978  25F3                                    blo      move                                                   ; no, keep on copying rom
E97A  8639            nohook  lda   #$39                                                      ; opcode for rts
E97C  8E015E                                  ldx   #hokini                                           ; get initial address of system hooks
E97F  A780            hooks   sta   ,x+                                                       ; overwrite them
E981  8C01A8                  cmpx  #hokend                                           ; got hooks end?
E984  25F9                    blo   hooks                                                     ; no, keep on overwriting them
E986  CC444B                                  ldd     #$444b                                          ; sign for dos roms
E989  8EC000                                  ldx     #romini                                         ; beginning of the rom
E98C  10A384                                  cmpd    ,x                                                              ; is a dos what we have loaded?
E98F  2602                                    bne     bas2                                                    ; no, execute it
E991  3002                                    leax    2,x                                                     ; skip over dos signature
E993  6E84            bas2            jmp   ,x                                                                ; start loaded rom
                      ; -------------------------------------------------------------------------------------------
E995  9612            doreadr         lda   <dnum                                                     ; our drive number
E997  5F                      clrb                                                                    ; lsn bits 23-16
E998  3436                    pshs  a,b,x,y                                           ; save registers / data
E99A  86D2                    lda   #$d2                                                      ; dw op_readex (read one sector)
E99C  3402            rereadr         pshs  a                                                         ; save command to stack
E99E  30E4                    leax  ,s                                                                ; point x to top of stack
E9A0  108E0005                ldy   #5                                                                ; 5 bytes
E9A4  BDFAE2                  jsr   DWWrite                                           ; send 5 bytes to dw      - readex, dnum, $0, sector
E9A7  3502                    puls  a                                                         ; discard command
E9A9  AE64                    ldx   4,s                                                       ; get read buffer pointer (old y)
E9AB  108E0100                ldy   #256                                                      ; read 256 bytes
E9AF  CC0085                  ldd   #133                                                      ; 1 second timeout
E9B2  BDFA5D                  jsr   DWRead                                            ; receive 256 bytes
E9B5  2534                    bcs   readex2                                           ; detect framing error
E9B7  2632                    bne   readex2                                           ; detect not all bytes received
E9B9  3420                    pshs  y                                                         ; save checksum
E9BB  30E4                    leax  ,s                                                                ; point x to top of stack
E9BD  108E0002                ldy   #2                                                                ; 2 bytes
E9C1  BDFAE2                  jsr   DWWrite                                           ; send checksum
E9C4  3121                    leay  1,y                                                       ; after a dwrite 'y' returns zeroed, so this way it becomes equal to one
E9C6  CC0085                  ldd   #133                                                      ; 1 second timeout
E9C9  BDFA5D                  jsr   DWRead                                            ; read dw answer
E9CC  3262                    leas  2,s                                                       ; discard checksum
E9CE  251B                    bcs   readex2                                           ; detect framing error
E9D0  2619                    bne   readex2                                           ; detect not all bytes received
E9D2  A6E4                    lda   ,s                                                                ; get server answer
E9D4  2715                    beq   readex2                                           ; zero=ok, return
E9D6  81F3                    cmpa  #$f3                                                      ; error code is #e_crc?
E9D8  2606                    bne   reader                                            ; not this one, go verify eof condition
E9DA  86F2                    lda   #$f2                                                      ; command reread. #op_rereadex
E9DC  6FE4                    clr   ,s                                                                ; clear received answer
E9DE  20BC                    bra   rereadr                                           ; read again
E9E0  81F4            reader  cmpa  #$f4                                      ; error code is 'read past end' (eof)?
E9E2  2606                                    bne   noteof                                            ; no, mark read error
E9E4  1CFC                    andcc #$fc                    ; unsets the 'c' and 'v' flags
E9E6  1A0C                    orcc  #$0c                                                      ; sets 'n' and 'z' flags. both prepare the flags to be positive at 'blt' test
E9E8  2001                    bra   readex2                                           ; jump over next instruction, not a read error
E9EA  53              noteof  comb                                                                    ; sets the carry flag (error)
E9EB  35B6            readex2         puls  a,b,x,y,pc                                        ; restore registers and return
                      
                      ; *******************************************************************************
                      ; Command to start a SDC-bank
                      ; *******************************************************************************
00F9                  swMap1  equ     $f9                                                     ; flag to switch to map1 (re-used)
                      
E9ED  BDE768          xbank           jsr     getunit                                         ; read a number between 0-7
E9F0  C107                                    cmpb    #7                                                              ; is it greater than 7?
E9F2  2305                                    bls     OKbank                                          ; no, OK
E9F4  C611            badnum  ldb     #$11                                                    ; SD error
E9F6  7E8344                                  jmp     syserr                                          ; bas_system_error
E9F9  1A50            OKbank  orcc    #dimask                                         ; disable interrupts
E9FB  D7F8                                    stb     <numUni                                         ; save received number in first sentence (B)
E9FD  0FF9                                    clr     swMap1                                          ; by default don't change map type
E9FF  9DA5                                    jsr     <redLChr                                                ; see next char after BANK
EA01  274B                                    beq     seeSlot                                         ; no extra chars, skip extra param search
EA03  812C                                    cmpa    #','                                                    ; is it a comma?
EA05  1026FD63                                lbne    snError                                         ; no, show error
EA09  9D9F                                    jsr     <getNChr                                                ; get next char
EA0B  8152                                    cmpa    #'R'                                                    ; is it R (for RAM mode = MAP1)?
EA0D  1026FD5B                                lbne    snError                                         ; no, show error
                      ; - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
                                                                                                                      ; copy and execute from RAM the 64k tester
EA11  8EEA2B                                  ldx     #tst64k                                         ; point to beginning of data to be copied
EA14  CE01DA                                  ldu     #$1da                                                   ; point to destination
EA17  EC81            xb001           ldd     ,x++                                                    ; get a wrod
EA19  EDC1                                    std     ,u++                                                    ; put at destination
EA1B  8CEA4C                                  cmpx    #emod1B+2                                       ; got to the end?
EA1E  25F7                                    blo     xb001                                                   ; no, loopback
EA20  BD01DA                                  jsr     $1da                                                    ; jsr that copied code from RAM
EA23  2627                                    bne     xb002                                                   ; if not zero, we have 64K
EA25  8EE57D                             ldx  #not64K-1                                       ; else show message NO 64K COMPUTER                     
EA28  7E90E5                                  jmp     outstr                                          ; show message
                      ; - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
                                                                                                                      ; validate if there are 64k - This must be executed from RAM
EA2B  8D0B            tst64k  bsr     mod1Byt                                         ; can modify rom area?
EA2D  2608                                    bne     etst64k                                         ; yes, return result
EA2F  B7FFDF                                  sta     $ffdf                                                   ; no, try switching to map1                     
EA32  8D04                                    bsr     mod1Byt                                         ; can modify area?
EA34  F7FFDE                                  stb     $ffde                                                   ; back to map0 where it was
EA37  39              etst64k rts                                                                     ; return result
EA38  5F              mod1Byt clrb                                                                    ; default 32K
EA39  B6C000                                  lda     $c000                                                   ; get first DOS ROM char 
EA3C  43                                      coma                                                                    ; invert bits
EA3D  B7C000                                  sta     $c000                                                   ; update value
EA40  B1C000                                  cmpa    $c000                                                   ; has been updated?
EA43  2605                                    bne     emod1B                                          ; no, return
EA45  43                                      coma                                                                    ; invert bits again
EA46  B7C000                                  sta     $c000                                                   ; restore value
EA49  5C                                      incb                                                                    ; found 64k
EA4A  5D              emod1B  tstb                                                                    ; update flags
EA4B  39                                      rts                                                                     ; return
                      ; - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
EA4C  0CF9            xb002           inc     swMap1                                          ; mark flag to change to map1
                                                                                                                      ; see if we have a SDC
EA4E  B60114          seeSlot lda     floSDC                                          ; get hdw control byte
EA51  8508                                    bita    #%00001000                                      ; is there an SDC?
EA53  102700A8                                lbeq    NotSDCPresent                           ; no, show error
EA57  84C0                                    anda    #%11000000                                      ; use bits 7-6 (slot where it is)
EA59  44                                      lsra                                                                    ; move them to
EA5A  44                                      lsra                                                                    ; bits 5-4 (lower bits of high nibble)
EA5B  8AC0                                    ora     #%11000000                                      ; set bits 7-6. High nibble OK
EA5D  3402                                    pshs    a                                                               ; save high nibble
EA5F  44                                      lsra                                                                    ; low nibble   
EA60  44                                      lsra                                                                    ; gets high nibble
EA61  44                                      lsra                                                                    ; that will end  
EA62  44                                      lsra                                                                    ; as zero
EA63  AAE0                                    ora     ,s+                                                     ; add high nibble (now they are the same value)
EA65  0DF9                                    tst     swMap1                                          ; should it be run in MAP1?
EA67  2616                                    bne     swMbank                                         ; yes, go to next section
EA69  3402            swbank  pshs    a                                                               ; save SDC slot number
EA6B  8EEA95                                  ldx     #special                                                ; point to the beginning of code to execute from RAM
EA6E  CE01DA                                  ldu     #$1da                                                   ; point at destination
EA71  EC81            swB01           ldd     ,x++                                                    ; get a word
EA73  EDC1                                    std     ,u++                                                    ; copy at destination
EA75  8CEAB9                                  cmpx    #exbank+3                                       ; got to the end?
EA78  25F7                                    blo     swB01                                                   ; no, loopback
EA7A  3502                                    puls    a                                                               ; restore slot number
EA7C  7E01DA                                  jmp     $1da                                                    ; jumto to RAM copied code
EA7F  3402            swMbank pshs    a                                                               ; save SDC slot number
EA81  8EEAB9                                  ldx     #allRAM                                         ; point to the beginning of code to execute from RAM
EA84  CE01DA                                  ldu     #$1da                                                   ; point at destination
EA87  EC81            swMB01  ldd     ,x++                                                    ; get a word
EA89  EDC1                                    std     ,u++                                                    ; copy at destination
EA8B  8CEAF3                                  cmpx    #m1L03+3                                                ; got to the end?
EA8E  25F7                                    blo     swMB01                                          ; no, loopback
EA90  3502                                    puls    a                                                               ; restore slot number
EA92  7E01DA                                  jmp     $1da                                                    ; jumto to RAM copied code
                                                                                                                      ; this TWO parts MUST be run from RAM because we are going
                                                                                                                      ; to lose control when switching the ROM slot and/or the bank!
                                                                                                                      ; code to start the bank in ROM mode (MAP0)
EA95  1A50            special orcc    #dimask                                         ; disable interrupts
EA97  7FFFDE                                  clr     $ffde                                                   ; switch to MAP0 (ROM-RAM mode)
EA9A  B7FF7F                                  sta     $ff7f                                                   ; switch to the slot where the SDC is
EA9D  D6F8                                    ldb     <numUni                                         ; get requested bank number
EA9F  F7FF4B                                  stb     $ff4b                                                   ; point to it
EAA2  8E444B                                  ldx     #$444b                                          ; get mark for DOS system
EAA5  BCC000                                  cmpx    romini                                          ; is there a DOS ROM?
EAA8  270A                                    beq     spec01                                          ; yes, issue cold start
EAAA  8E7F37                                  ldx     #$7f37                                          ; get std value for stack
EAAD  3284                                    leas    ,x                                                              ; assign it
EAAF  C4AF                                    andb    #eimask                                         ; enable interrupts
EAB1  7EC000                                  jmp     romini                                          ; no, execute ROMPACK
EAB4  0F71            spec01  clr     <$71                                                    ; force a cold boot
EAB6  7EB3B4          exbank  jmp     resetRt                                         ; execute the system reset routine
                                                                                                                      ; code to start the bank but having copied everything into MAP1
EAB9  1A50            allRAM  orcc  #dimask                           ; disable interrupts
EABB  B7FF7F                                  sta     $ff7f                                                   ; switch to the slot where the SDC is
EABE  D6F8                                    ldb     <numUni                                         ; get requested bank number
EAC0  F7FF4B                                  stb     $ff4b                                                   ; point to it
                                                                                                                      ; copy ROM to RAM
EAC3  8E8000                                  ldx   #basini                           ; point to beginning of Basic
EAC6  B7FFDE          m1L01           sta   $ffde                             ; switch to ROM map
EAC9  EC84                                    ldd   ,x                                ; get a word
EACB  B7FFDF                                  sta   $ffdf                             ; switch to RAM map
EACE  ED81                                    std   ,x++                              ; copy at RAM destination
EAD0  8CFEFF                                  cmpx  #$feff                            ; end of ROM?
EAD3  25F1                                    bcs   m1L01                                     ; no, loopback
EAD5  8639                                    lda   #$39                              ; get RTS opcode
EAD7  8E015E                                  ldx   #$015e                            ; point to beginning of User vectors (hooks)
EADA  A780            m1L02           sta   ,x+                               ; put an RTS in each one
EADC  8C01A8                                  cmpx  #$01a8                            ; end of user vectors?
EADF  25F9                                    bcs   m1L02                                             ; no, loopback
EAE1  BDBA77                                  jsr     $ba77                                                   ; clear screen
EAE4  CC444B                                  ldd     #$444b                                          ; get mark for O.S.
EAE7  10B3C000                                cmpd    romini                                          ; is there an O.S. in RAM now?
EAEB  2703                                    beq     m1L03                                                   ; yes, start it
EAED  7EC000                                  jmp     romini                                          ; else start ROM from the beginning
EAF0  7EC002          m1L03           jmp   romini+2                          ; start ROM (copied into RAM)
                      
                      ;--------------------------------------------------------------------------
                                                                                                                              ; MESSAGES. They all will return with carry set.
EAF3                  BadFormat
EAF3  8EE456                                  ldx     #BadFmtType-1                           ; point to message
EAF6  10                                      fcb     #$10                                                    ; skip next ldx
EAF7  8EE416          TooLong ldx     #Str2Long-1                                     ; point to message
EAFA  10                                      fcb     #$10                                                    ; skip next ldx
EAFB                  NotMPIPresent
EAFB  8EE427                                  ldx     #NotAnMPI-1                                     ; point to message
EAFE  10                                      fcb     #$10                                                    ; skip next ldx
EAFF                  NotSDCPresent
EAFF  8EE437                                  ldx     #NotAnSDC-1                                     ; point to message
EB02  BD90E5                                  jsr     outstr                                          ; show it
EB05  1A01                                    orcc    #%00000001                                      ; set carry flag
EB07  39                                      rts                                                                     ; return
                      
                      ;--------------------------------------------------------------------------
                      ; This routine will copy the hardware tests routines at $1da and jump there 
                      ; to avoid hanging the system while switching of MPI slot
                      ;--------------------------------------------------------------------------
EB08  8EEB1B          tstHARD ldx     #tstHDW                                         ; point to beginning of program to be copied
EB0B  CE01DA                                  ldu     #$1da                                                   ; point to destination
EB0E  EC81            tstL01  ldd     ,x++                                                    ; get a word
EB10  EDC1                                    std     ,u++                                                    ; put at destination
EB12  8CEC0B                                  cmpx    #xslot                                          ; got to end of program?
EB15  25F7                                    blo     tstL01                                          ; no, loopback
EB17  BD01DA                                  jsr     $1da                                                    ; execute copied program in low RAM
EB1A  39                                      rts                                                                     ; return
                      
                      ;--------------------------------------------------------------------------
                      ; test for the presence of MPI, floppies, SDC and update $114 accordingly
                      ;--------------------------------------------------------------------------
EB1B  5F              tstHDW  clrb                                                                    ; get zero value
EB1C  F70114                                  stb     floSDC                                          ; control byte default value (no hdw found)
EB1F  170097                                  lbsr    tstMPI                                          ; is there an MPI present?
EB22  10260092                                lbne    etstHDW                                         ; no, exit
EB26  D776                                    stb     hdwFlo                                          ; no Floppies found
EB28  D777                                    stb     hdwSdc                                          ; no SDC found
                                                                                                                      ; search for SDC and/or floppies
EB2A  B6FF7F                                  lda     $ff7f                                                   ; get active MPI slot                                           --- $FF slot 4 (discos)
EB2D  8433                                    anda    #$33                                                    ; save only real numbers                                        --- $33
EB2F  B7FF7F                                  sta     $ff7f                                                   ; working value                                                         --- $ff7f = $33
EB32  C604                                    ldb     #4                                                              ; ready to check four slots
                                                                                                                      ; first look for SDC
EB34  0D77            tst01           tst     hdwSdc                                          ; have we found it before?
EB36  2613                                    bne     tst02                                                   ; yes, go to floppy test
EB38  17009C                                  lbsr    tstSDC                                          ; is there an SDC in this MPI slot?             
EB3B  260E                                    bne     tst02                                                   ; no, go to floppy test
EB3D  8ACC                                    ora     #$cc                                                    ; add higher bits to 1                                                                          --- $CC
EB3F  9777                                    sta     hdwSdc                                          ; save slot where SDC has been found                                    --- found at $CC
EB41  B60114                                  lda     floSDC                                          ; get hdw control byte
EB44  8A08                                    ora     #%00001000                                      ; add flag for SDC present
EB46  B70114                                  sta     floSDC                                          ; update it
EB49  2015                                    bra     tst03                                                   ; go process another slot
                                                                                                                      ; if not SDC, then for floppy
EB4B  0D76            tst02           tst     hdwFlo                                          ; have we found it before?
EB4D  2611                                    bne     tst03                                                   ; go process another slot
EB4F  17009C                                  lbsr    tstFLO                                          ; is there a floppy controller in this MPI slot?
EB52  260C                                    bne     tst03                                                   ; no, process another slot
EB54  8ACC                                    ora     #$cc                                                    ; add higher bits to 1                                                                          --- $FF
EB56  9776                                    sta     hdwFlo                                          ; save slot where floppy has been found                         --- found at $FF
EB58  B60114                                  lda     floSDC                                          ; get hdw control byte
EB5B  8A04                                    ora     #%00000100                                      ; add flag for Floppies present
EB5D  B70114                                  sta     floSDC                                          ; update it
                                                                                                                      ; look at another slot
EB60  B6FF7F          tst03           lda     $ff7f                                                   ; get analized slot                                                                                                                                                     ---     A = $33/$22/$11/$00
EB63  8011                                    suba    #$11                                                    ; calculate previous slot                                                                                                                                       ---   A = $22/$11/$00/$EF
EB65  2A02                                    bpl     tst04                                           ; if not negative, skip next
EB67  8633                                    lda     #$33                                                    ; get upper value                                                                                                                                                               ---   A = $33
EB69  3402            tst04           pshs    a                                                               ; save new calculated slot number                                                                                                       ---   pushed $22/$11/$00/$33
EB6B  B6FF7F                                  lda     $ff7f                                                   ; get last tested slot value                                                                                                                    ---        A = $33/$22/$11/$00
EB6E  84CC                                    anda    #$cc                                                    ; clean bits 0-1 in both nibbles                                                                                                                ---        A = $00/$00/$00/$00
EB70  AAE0                                    ora     ,s+                                                     ; add new calculated slot number ($00-$11-$22-$33)                                                              ---        A = $22/$11/$00/$33
EB72  B7FF7F                                  sta     $ff7f                                                   ; change of I/O slot (not keeping the same ROM slot)                                                    ---  $FF7F = $22/$11/$00/$33
EB75  5A                                      decb                                                                    ; decrement slot counter                                                                                                                                ---      B =   3/  2/  1/  0
EB76  26BC                                    bne     tst01                                                   ; if not zero, loopback
                                                                                                                      ; if zero: end of tests. $ff7f must be restored to original value
EB78  B6FF7F                                  lda     $ff7f                                                   ; get calculated value (initial slot)                                                                                           ---                                                      $33
EB7B  8ACC                                    ora     #$cc                                                    ; add high bits to 1                                                                                                                                            ---                                                      $ff
EB7D  B7FF7F                                  sta     $ff7f                                                   ; update active slot                                                                                                                                            ---  $FF7F = $CC
EB80  B60114          tst05           lda     floSDC                                          ; get hdw control byte
EB83  8508                                    bita    #%00001000                                      ; is bit3 ON (SDC present)?
EB85  271C                                    beq     tst07                                                   ; no, look for floppies 
EB87  9677                                    lda     hdwSdc                                          ; get SDC slot                                                                                                                                                                  --- A = $CC
EB89  B1FF7F                                  cmpa    $ff7f                                                   ; equals $ff7f? 
EB8C  2608                                    bne     tst06                                                   ; no, don't modify floSDC
EB8E  F60114                                  ldb     floSDC                                          ; get hdw control byte                                                                                  
EB91  CA03                                    orb     #%00000011                                      ; set floSDC.bits1-0 to #%11 (SDC ON)
EB93  F70114                                  stb     floSDC                                          ; update control byte
EB96  8403            tst06           anda    #%00000011                                      ; use just slot number                                                                                                                                          --- A = $00
EB98  5F                                      clrb                                                                    ; result = 0                                                                                                                                                                    --- B = 0
EB99  44                                      lsra                                                                    ; send bit 0 to carry                                                                                                                                           --- A = 0, C = 0
EB9A  56                                      rorb                                                                    ; receive bit0 from carry in bit7                                                                                                               --- B = 0
EB9B  44                                      lsra                                                                    ; send bit 1 (now 0) to carry                                                                                                                           --- A = 0, C = 0
EB9C  56                                      rorb                                                                    ; receive bit1 (now 0) from carry in bit7 and bit6 goes to bit7                         --- B = 0
EB9D  FA0114                                  orb     floSDC                                          ; update bits 7-6 of $114 with slot used by SDC                                                                         --- floSDC.bits5-4 = %00
EBA0  F70114                                  stb     floSDC                                          ; save new value
EBA3  B60114          tst07           lda     floSDC                                          ; get hdw control byte
EBA6  8504                                    bita    #%00000100                                      ; is bit2 ON (floppies present)?
EBA8  270E                                    beq     etstHDW                                         ; no, exit
EBAA  9676                                    lda     hdwFlo                                          ; get floppies slot                                                                                                                                                     --- A = $FF
EBAC  8403                                    anda    #%00000011                                      ; get just slot number                                                                                                                                          --- A = $03
EBAE  48                                      lsla                                                                    ; move bits 0-1                                                                                                                                                         --- A = $06
EBAF  48                                      lsla                                                                    ; to bits 5-4                                                                                                                                                                   --- A = $0C
EBB0  48                                      lsla                                                                    ;                                                                                                                                                                                                       --- A = $18
EBB1  48                                      lsla                                                                    ;                                                                                                                                                                                                       --- A = $30
EBB2  BA0114                                  ora     floSDC                                          ; update bits 5-4 of $114 with slot used by Floppies                                                            --- floSDC.bits7-6 = %11
EBB5  B70114                                  sta     floSDC                                          ; save new value
EBB8  39              etstHDW rts                                                                     ; return
                      
                      ;--------------------------------------------------------------------------
                      ; Tests for the presence of an MPI
                      ; returns zero flag Z=1 if present else to zero
                      ; by Darren Atkinson (2015)
                      ;--------------------------------------------------------------------------
EBB9  3402            tstMPI  pshs    a                                                               ; save register
EBBB  B6FF7F                                  lda     $ff7f                                           ; get MPI settings                                                                                                      --- A = $FF
EBBE  880C                                    eora  #$0c                                              ; toggle two unused bits in SCS nibble                                                  --- A = $F3
EBC0  B7FF7F                                  sta   $ff7f                                             ; update MPI settings                                                                                           --- $FF7F = $F3
EBC3  B1FF7F                                  cmpa  $ff7f                                             ; did it accept the change?
EBC6  2704                                    beq   yesMPI                                            ; branch if yes (MPI present), will return with Z=1
EBC8  8601                                    lda     #1                                                              ; will return with Z=0
EBCA  2009                                    bra     noMPI                                                   ; exit
EBCC  B6FF7F          yesMPI  lda     $ff7f                                                   ; get modified value                                                                                                    --- A = $F3
EBCF  880C                                    eora    #$0c                                                    ; recover possible modified bits                                                                        --- A = $FF
EBD1  B7FF7F                                  sta     $ff7f                                                   ; back to its old value                                                                                         --- $FF7F = $FF
EBD4  4F                                      clra                                                                    ; make Z=1
EBD5  3582            noMPI           puls    a,pc                                                    ; restore register and return
                      ;--------------------------------------------------------------------------
                      ; Tests for the presence of a CoCo SDC on active slot
                      ; returns zero flag Z=1 if present else to zero
                      ; by Darren Atkinson (2015)
                      ;--------------------------------------------------------------------------
EBD7  3412            tstSDC  pshs    a,x                                                     ; save registers
EBD9  8EFF4A                          ldx   #$ff4a                                            ; point X at Flash registers (FF42 for CoCo)
EBDC  861A                            lda   #$1a                                              ; test pattern
EBDE  A784                            sta   ,x                                                ; store in data register
EBE0  A601                            lda   1,x                                               ; get value from control register
EBE2  6F84                            clr   ,x                                                ; clear data register
EBE4  A801                            eora  1,x                                               ; get changed bits from control register
EBE6  8118                            cmpa  #$18                                              ; did the expected bits change?
EBE8  2702                            beq   yesSDC                                            ; branch if SDC present, will return with Z=1
EBEA  8601                                    lda     #1                                                              ; will return with Z=0
EBEC  3592            yesSDC  puls    a,x,pc                                          ; restore registers and return
                      
                      ;--------------------------------------------------------------------------
                      ; Test for presence of FDC in active MPI slot.
                      ; Returns with Z=1 if FDC has been found
                      ; by Darren Atkinson (2015)
                      ;--------------------------------------------------------------------------
EBEE  3416            tstFLO   pshs  d,x                              ; save registers
EBF0  8EFF40                   ldx   #$ff40                           ; address for Dragon DOS controller
EBF3  86D0                     lda   #$d0                             ; FORCE INTERRUPT command
EBF5  A784                     sta   ,x                               ; send to FDC command register
EBF7  8D0D                     bsr   fdcDelay                         ; delay time
EBF9  A684                     lda   ,x                               ; read FDC status to clear INTRQ
EBFB  EC02                     ldd   2,x                              ; get FDC sector and data registers
EBFD  43                       coma                                   ; invert sector bits only
EBFE  ED02                     std   2,x                              ; put both registers back
EC00  8D04                     bsr   fdcDelay                         ; delay time
EC02  A302                     subd  2,x                              ; set Z if read matches write
EC04  3596                     puls  d,x,pc                           ; restore registers and return
EC06  3437            fdcDelay pshs  y,x,d,cc                         ; push registers
EC08  3D                       mul                                    ; consume 11 CPU cycles
EC09  35B7                     puls  cc,d,x,y,pc                      ; restore registers and return
                      
                      ;--------------------------------------------------------------------------
                      ; Command to change of MPI-slot and start it 
                      ; it could happen that the new slot has no FDC nor SDC
                      ; but if it had any of them, then floSDC must be adjusted accordingly
                      ;--------------------------------------------------------------------------
EC0B  9DA5            xslot           jsr   <redLChr                                          ; peek at next character after "SLOT"
EC0D  804E                    suba  #'N'                                                      ; to detect if it was an 'N' - if so do not switch to MAP0
EC0F  9710                    sta   <noexec                                           ; save difference. 0 = do not switch
EC11  2602                    bne   xslot1                                            ; if not 'N' skip next
EC13  9D9F                    jsr   <getNChr                                          ; jump over char 'N'
EC15  BDE768          xslot1  jsr     getunit                                         ; read a number between 1-4
EC18  2704                                    beq     nogood                                          ; if 0 error
EC1A  C104                                    cmpb    #4                                                              ; is it greater than 4?
EC1C  2305                                    bls     OKslot                                          ; no, OK
EC1E  C60D            nogood  ldb     #$0d                                                    ; MU error
EC20  7E8344                                  jmp     syserr                                          ; bas_system_error
EC23  8D94            OKslot  bsr     tstMPI                                          ; is there an MPI present?
EC25  2704                                    beq     OKslot01                                                ; yes, skip section
EC27  17FED1                                  lbsr    NotMPIPresent                           ; no, show message
EC2A  39                                      rts                                                                     ; return
EC2B  5A              OKslot01        decb                                                                    ; slots range 0-3. We use reg B read at function getunit
EC2C  D7F8                                    stb     <numUni                                         ; save low nibble --- $00 - $03 if command was SLOTN1 - SLOTN4
EC2E  58                                      lslb                                                                    ; move it                       --- $00 - $06
EC2F  58                                      lslb                                                                    ; to the                        --- $00 - $0c
EC30  58                                      lslb                                                                    ; high                          --- $00 - $18
EC31  58                                      lslb                                                                    ; nibble                                --- $00 - $30
EC32  DAF8                                    orb     <numUni                                         ; add low nibble ($nn)          --- $00 - $33
EC34  CACC                                    orb     #$cc                                                    ; to get $CC-$DD,$EE,$FF        --- $cc - $ff
EC36  D7F8                                    stb     <numUni                                         ; save calculated code          --- $cc - $ff
                                                                                                                      ; now must compare the new slot with the one for FDC and SDC and modify floSDC accordingly
                                                                                                                      ; by definition, switching the slot MUST enable the drives of that slot, so $114 must vary
EC38  C430                                    andb    #%00110000                                      ; get just low bits of high nibble = new slot number    --- $00 - $30
EC3A  3404                                    pshs    b                                                               ; save value                                                                                                            --- $00 - $30
EC3C  B60114                                  lda     floSDC                                          ; get hdw control byte                                                                                  --- $3C - $3C if we are in the FDC
EC3F  8430                                    anda    #%00110000                                      ; get the slot where the Floppies are                                           --- $30 - $30
EC41  A1E0                                    cmpa    ,s+                                                     ; is new slot where the floppies are?                                           --- NO  - YES
EC43  2607                                    bne     xslot2                                          ; no, compare SDC                                                                                                       --- jmp - mod
EC45  B60114                                  lda     floSDC                                          ; get control value
EC48  84FC                                    anda    #%11111100                                      ; set both floppies ON
EC4A  2016                                    bra     xslot3                                          ; skip section - no need to verify SDC
EC4C  96F8            xslot2  lda     <numUni                                         ; get new slot number                                                                                   --- $cc - $ff
EC4E  8430                                    anda    #%00110000                                      ; get just low bits of high nibble = new slot number    --- $00 - $30
EC50  48                                      asla                                                                    ; move 2 bits to 
EC51  48                                      asla                                                                    ; the left                                                                                                                      --- $00 - $C0
EC52  3402                                    pshs    a                                                               ; save value                                                                                                            --- $00 - $C0
EC54  B60114                                  lda     floSDC                                          ; get control bte
EC57  84C0                                    anda    #%11000000                                      ; get the slot where SD is                                                                              --- $00 - $00
EC59  A1E0                                    cmpa    ,s+                                                     ; is the select one?                                                                                            --- YES - NO
EC5B  2608                                    bne     xslot4                                          ; no, skip two next                                                                                             --- mod - jump
EC5D  B60114                                  lda     floSDC                                          ; get control value
EC60  8A03                                    ora     #%00000011                                      ; set both floppies OFF
EC62  B70114          xslot3  sta     floSDC                                          ; update hdw control byte
                                                                                                                              ; this part MUST be run from RAM because we are going
                                                                                                                              ; to lose control when switching the ROM slot!
EC65  8EEC77          xslot4  ldx     #specia2                                                ; point to beginning of code to execute from RAM
EC68  CE01DA                                  ldu     #$1da                                                   ; point at destination
EC6B  EC81            xsL01           ldd     ,x++                                                    ; get a word
EC6D  EDC1                                    std     ,u++                                                    ; put at destination
EC6F  8CEC98                                  cmpx    #strROM+3                                       ; got to the end?
EC72  25F7                                    blo     xsL01                                                   ; no, loopback
EC74  7E01DA                                  jmp     $1da                                                    ; jump to RAM copied code
EC77  1A50            specia2 orcc    #dimask                                         ; disable interrupts
EC79  0D10                                    tst     <noexec                                         ; should switch to MAP0?                                                                --- noexec = 0
EC7B  2703                                    beq     specia3                                         ; no, skip next                                                                                 --- apply jump
EC7D  7FFFDE                                  clr     $ffde                                                   ; switch to MAP0 (ROM mode)                                                     --- skipped
EC80  D6F8            specia3 ldb     <numUni                                         ; get calculated code                                                                   --- B = $cc
EC82  F7FF7F                                  stb     $ff7f                                                   ; switch to that slot                                                                   --- $ff7f = $cc
EC85  0D10                                    tst     <noexec                                         ; called with param N?
EC87  2601                                    bne     eOKSlot                                         ; no, start slot
EC89  39                                      rts                                                                     ; return
EC8A  CC444B          eOKSlot ldd     #$444b                                          ; signature for O.S.
EC8D  10B3C000                                cmpd    romini                                          ; is an O.S. at $C000?
EC91  2602                                    bne     strROM                                          ; no, start RomPack
EC93  0F71                                    clr     <$71                                                    ; force coldstart
EC95  7EB3B4          strROM  jmp     resetRt                                         ; reset machine
                      
                      ;********************************************************************************************************************************
                      ; to access floppies or SDC-card
                      ; uses byte floSDC ($114) to determine what it must access
                      ; byte structure: 
                      ;       bit 7-6 - slot number where is the CoCo-SDC  (0-3)
                      ;       bit 5-4 - slot number where are the floppies (0-3)
                      ;       bit 3           - 1 if SDC present else 0
                      ;       bit 2           - 1 if Floppy controller present else 0
                      ;       bit 1           - for Unit #2: if 0 use floppy, if 1 use SDC
                      ;       bit 0           - for Unit #1: if 0 use floppy, if 1 use SDC
                      ;********************************************************************************************************************************
EC98  3401            flOrSd  pshs    cc                                                              ; save register
EC9A  B60114                                  lda     floSDC                                          ; get hdw control byte
EC9D  840C                                    anda    #%00001100                                      ; test bits3-2 (SDC and Floppies present)
EC9F  810C                                    cmpa    #%00001100                                      ; are both present?
ECA1  263C                                    bne     eflOr                                                   ; no, exit
ECA3  96EB                                    lda     <$eb                                                    ; get unit number
ECA5  8102                                    cmpa    #2                                                              ; is greater than 2?  --- If 3-4 and SDC present, should change to FDC slot!
ECA7  221F                                    bhi     flo2flo                                         ; yes, switch to floppies
ECA9  B60114                                  lda     floSDC                                          ; get hdw control byte
ECAC  94EB                                    anda    <$eb                                                    ; test corresponding bit 0-1 (affected drive)
ECAE  2718                                    beq     flo2flo                                         ; if zero, switch to floppies
                                                                                                                      ; has to switch to slot with SDC (bits 7-6 of floSDC)
ECB0  B60114          flo2SDC lda     floSDC                                          ; get hdw control byte
ECB3  84C0                                    anda    #%11000000                                      ; use bits 7-6 to get slot number
ECB5  44                                      lsra                                                                    ; send them
ECB6  44                                      lsra                                                                    ; to bits 5-4 (2 low bits of high nibble)
ECB7  8AC0                                    ora     #$c0                                                    ; set bits 7-6. High nibble OK
ECB9  3402                                    pshs    a                                                               ; save high nibble
ECBB  44                                      lsra                                                                    ; move it
ECBC  44                                      lsra                                                                    ; 4 times to the right
ECBD  44                                      lsra                                                                    ; to have it
ECBE  44                                      lsra                                                                    ; as low nibble
ECBF  AAE0                                    ora     ,s+                                                     ; add high nibble and clean stack
ECC1  B1FF7F                                  cmpa    $ff7f                                                   ; already in that slot?
ECC4  2719                                    beq     eflOr                                                   ; yes, do nothing
ECC6  2014                                    bra     flOr01                                          ; go update value
                                                                                                                      ; has to switch to slot with floppies (bits 5-4- of floSDC)
ECC8  B60114          flo2flo lda     floSDC                                          ; get hdw control byte
ECCB  8430                                    anda    #%00110000                                      ; use bits 5-4 (2 low bits of high nibble)
ECCD  8AC0                                    ora     #$c0                                                    ; set bits 7-6. High nibble OK
ECCF  3402                                    pshs    a                                                               ; save high nibble
ECD1  44                                      lsra                                                                    ; move it
ECD2  44                                      lsra                                                                    ; 4 times to the right
ECD3  44                                      lsra                                                                    ; to have it
ECD4  44                                      lsra                                                                    ; as low nibble
ECD5  AAE0                                    ora     ,s+                                                     ; add high nibble and clean stack
ECD7  B1FF7F                                  cmpa    $ff7f                                                   ; already in that slot?
ECDA  2703                                    beq     eflOr                                                   ; yes, do nothing
ECDC  B7FF7F          flOr01  sta     $ff7f                                                   ; change active MPI slot
ECDF  3581            eflOr           puls    cc,pc                                                   ; restore register and return
                      
                      ;--------------------------------------------------------------------------
                      ; accept a value to be used as index disc in a HD in unit4 (for DW4)
                      ;--------------------------------------------------------------------------
ECE1  BDE768          xHDidx  jsr     getunit                                         ; read a number between 1-90
ECE4  C101                                    cmpb    #1                                                              ; value under minimum accepted?
ECE6  1025080E                                lblo    badParam                                                ; yes, show message
ECEA  C15A                                    cmpb    #90                                                     ; exceeds the max index number?
ECEC  2207                                    bhi     idxBad                                          ; yes, show message
ECEE  F70115                                  stb     idxHD                                                   ; save disc index number
ECF1  BDE68B                                  jsr     mrkmod                                          ; mark as modified
ECF4  39              exHD            rts                                                                     ; return
ECF5  8EE46E          idxBad  ldx     #idxHigh-1                                      ; poit to message
ECF8  7E90E5                                  jmp     outstr                                          ; show it
                      ;--------------------------------------------------------------------------
                      ; Command to deal with SDC-card DRIVE commands (1,2)
                      ;--------------------------------------------------------------------------
00FA                  valPar  equ     $fa                                                     ; (re-used) value for parameter to send into B
00FB                  cmdNum  equ     $fb                                                     ; (re-used) command code to be sent
0076                  bufPtr  equ     $76                                                     ; (re-used) pointer to buffer end for new name     *** ### ***
0076                  dsktyp  equ     $76                                                     ; (re-used) only for SDRIVE creating file. 0=DSK, 1=VDK
00F8                  cpyPtr  equ     $f8                                                     ; (re-used) pointer to buffer end for SDIR
03D9                  fST16           equ     $3d9                                                    ; first sector of Track 16 (backup DIR)
                      ; - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -                                                           
ECFB  9DA5            xsdrive jsr     <redLChr                                                ; peek at next char after SDRIVE
ECFD  10270473                                lbeq    sdcLST                                          ; is the SDRIVE alone
                                                                                                                      ; validate syntax and drive number
ED01  BDE768                                  jsr     getunit                                         ; read a number between 1-2
ED04  1027FCEC                                lbeq    badnum                                          ; 0 not accepted
ED08  C102                                    cmpb    #2                                                              ; is it greater than 4?
ED0A  1022FCE6                                lbhi    badnum                                          ; no, ok
ED0E  5A                                      decb                                                                    ; internally will work with 0-1
ED0F  D7F8                                    stb     <numUni                                         ; save received number
ED11  0FFA                                    clr     <valPar                                         ; value param is 0
ED13  C6E0                                    ldb     #$e0                                                    ; more used value
ED15  D7FB                                    stb     <cmdNum                                         ; store it
ED17  BD89AA                                  jsr     ckComma                                         ; ckcomma. if not comma found syntax error
                      ; - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -                                                           
                                                                                                                      ; decode token
ED1A  81C2                                    cmpa    #$c2                                                    ; is this the token for off?
ED1C  10270348                                lbeq    flopON                                          ; yes, go process it
ED20  8188                                    cmpa    #$88                                                    ; is this the token for on?
ED22  10270339                                lbeq    flopOF                                          ; yes, go process it
ED26  81B3                                    cmpa    #$b3                                                    ; is this the token for get?
ED28  10270369                                lbeq    sdcGET                                          ; yes, go process it
ED2C  818C                                    cmpa    #$8c                                                    ; is this the token for dim?
ED2E  10270355                                lbeq    sdcDIM                                          ; yes, go process it
ED32  81F6                                    cmpa    #$f6                                                    ; is this the token for unload?
ED34  260B                                    bne     nxtParam                                                ; no, go to params section
ED36  9D9F                                    jsr     <getNChr                                                ; yes, it is the 'unload' command version. must skip the token to avoid a sn error when done
ED38  CC0100          reuseUL ldd     #$0100                                          ; get 1 byte with value equal null
ED3B  FD01D1                                  std     nambuf                                          ; put in nambuf area
ED3E  8D60            dTok01  bsr     putCmd                                          ; skip parameters section
ED40  39                                      rts                                                                     ; return
                      ; - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -                                                           
                                                                                                                      ; search for more parameters
ED41  BDB7AA          nxtParam        jsr   getfnam                                           ; read string and length into $1d1 = nambuf
ED44  F601D1                  ldb     nambuf                                          ; is string length 0?
ED47  1027FA21                lbeq    snError                                         ; yes, show error
ED4B  C1FD                    cmpb    #253                                                    ; is string longer than maximum accepted? (253 + m: + final null)
ED4D  1024FDA6                                lbcc    TooLong                                         ; show error message
ED51  9DA5                                    jsr     <redLChr                                                ; get last read char
ED53  812C                                    cmpa    #$2c                                                    ; is it a comma?
ED55  26E7                                    bne     dTok01                                          ; no, end of command. this is 'mount' command version
ED57  9D9F                                    jsr     <getNChr                                                ; get next token
ED59  1027FA0F                                lbeq    snError                                         ; if no token after comma, show error
ED5D  5F                                      clrb                                                                    ; bit 7 value 0 means dsk format
ED5E  8197                                    cmpa    #$97                                                    ; is it the token for new?
ED60  2708                                    beq     crtDisc                                         ; yes, create a dsk file
ED62  CA80                                    orb     #$80                                                    ; bit 7 value 1 means sdf format
ED64  8198                                    cmpa    #$98                                                    ; is it the token for def?
ED66  1026FA02                                lbne    snError                                         ; no, show error message
                      ; - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -                                                           
                                                                                                                      ; create disc
ED6A  D710            crtDisc stb     <noexec                                         ; save format flag (bit7)
ED6C  1704C3                                  lbsr    addExt                                          ; add extension to the file name
ED6F  9D9F                                    jsr     <getNChr                                                ; get next char
ED71  812C                                    cmpa    #$2c                                                    ; is it a comma?
ED73  2707                                    beq     crt01                                                   ; yes, test size parameter
ED75  8601                                    lda     #1                                                              ; by default type=1
ED77  8E00B4                                  ldx     #180                                                    ; and will be 180k (720 sectors)
ED7A  201A                                    bra     crtstep                                         ; exit cheks
ED7C  9D9F            crt01           jsr     <getNChr                                                ; advance one char. mandatory for next routine
ED7E  BD8E83                                  jsr     get16bN                                         ; get a 16 bit number into x
ED81  8601                                    lda     #1                                                              ; 1 will mean 180k dragon-like disc (40 tracks 1 side)
ED83  8C00B4                                  cmpx    #180                                                    ; equals read value 180?
ED86  270E                                    beq     crtstep                                         ; yes, exit checks
ED88  4C                                      inca                                                                    ; 2 will mean 360k dragon-like disc (40 tracks 2 sides)
ED89  8C0168                                  cmpx    #360                                                    ; equals read value 360?
ED8C  2708                                    beq     crtstep                                         ; yes, exit checks
ED8E  4C                                      inca                                                                    ; 3 will mean 720k dragon-like disc (80 tracks 2 sides)
ED8F  8C02D0                                  cmpx    #720                                                    ; equals read value 720?
ED92  1026FD5D                                lbne    BadFormat                                       ; not, bad format number, show error
ED96  9B10            crtstep adda    <noexec                                         ; add disc type to format flag
ED98  9710                                    sta     <noexec                                         ; save calculated value ($01-02-03 or $81-82-83)
                      ; - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -                                                           
                                                                                                                      ; entries to send command to the sdc
                                                                                                                      ; entry for create new disc comands
ED9A  CC4E3A                                  ldd     #"N:"                                                   ; command to be sent to create a file
ED9D  8D06                                    bsr     putCmd2                                         ; send command
ED9F  39              noMsgEr rts                                                                     ; return
                                                                                                                      ; entry for mount - unmount like commands
EDA0  0F10            putCmd          clr     <noexec                                         ; no params needed (no disk to be created)
EDA2  CC4D3A                                  ldd     #"M:"                                                   ; command to be sent to mount or unmount a drive
                      ; - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -                                                           
                                                                                                                      ; detect if sdc is already active
EDA5  0F77            putCmd2 clr     hdwSdc                                          ; no need to go back to old active slot
EDA7  17FE2D                                  lbsr    tstSDC                                          ; is there an active sdc?
EDAA  2732                                    beq     xsd01                                                   ; yes, go on
                      ; - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -                                                           
                                                                                                                      ; if present but not active change to that slot, save old one                   
                                                                                                                      ; if sdc present, active it and mark somewhere that we must go back to the active one!
EDAC  3406                                    pshs    d                                                               ; save registers
EDAE  B60114                                  lda     floSDC                                          ; get hdw control byte
EDB1  8508                                    bita    #%00001000                                      ; is a sdc in the system?
EDB3  260E                                    bne     vfySD01                                         ; yes, jump section
EDB5  3264                                    leas    4,s                                                     ; get rid of d and 1 return address
EDB7  AEE4                                    ldx     ,s                                                              ; see if return address is $0900 (from sdir?)
EDB9  8C0900                                  cmpx    #$0900                                          ; comes from sdir?
EDBC  2602                                    bne     vfySD00                                         ; no, skip next
EDBE  3262                                    leas    2,s                                                     ; clean stack                                                                   
EDC0  16FD3C          vfySD00 lbra    NotSDCPresent                           ; no, show message
                                                                                                                      ; calculate sdc slot
EDC3  B60114          vfySD01 lda     floSDC                                          ; get hdw control byte
EDC6  84C0                                    anda    #%11000000                                      ; use bits 7-6 (slot where sdc is)
EDC8  44                                      lsra                                                                    ; move them to
EDC9  44                                      lsra                                                                    ; bits 5-4 (lower bits of high nibble)
EDCA  8AC0                                    ora     #%11000000                                      ; set bits 7-6. high nibble ok
EDCC  3402                                    pshs    a                                                               ; save high nibble
EDCE  44                                      lsra                                                                    ; low nibble   
EDCF  44                                      lsra                                                                    ; gets high nibble
EDD0  44                                      lsra                                                                    ; that will end  
EDD1  44                                      lsra                                                                    ; as zero
EDD2  AAE0                                    ora     ,s+                                                     ; add high nibble (now they are the same value)
                                                                                                                      ; save active slot in hdwsdc
EDD4  F6FF7F                                  ldb     $ff7f                                                   ; get active slot
EDD7  D777                                    stb     hdwSdc                                          ; save it for later use                                         
EDD9  B7FF7F                                  sta     $ff7f                                                   ; activate sdc
EDDC  3506                                    puls    d                                                               ; restore registers
                      ; - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -                                                           
                                                                                                                      ; pass string to data buffer
EDDE  8E01D2          xsd01           ldx   #nambuf+1                                 ; point to 1st char of the input string
EDE1  CE0800                                  ldu     #$0800                                          ; point to 1st dos buffer
EDE4  EDC1                                    std     ,u++                                                    ; put at the beginning of the command buffer
EDE6  F601D1                                  ldb     nambuf                                          ; get string length
EDE9  A680            str2buf lda     ,x+                                                     ; get a char from input string
EDEB  A7C0                                    sta     ,u+                                                     ; copy in the buffer
EDED  5A                                      decb                                                                    ; decrement counter
EDEE  26F9                                    bne     str2buf                                         ; not yet done, loopback
EDF0  6FC4                                    clr     ,u                                                              ; put a final zero
EDF2  0D10                                    tst     <noexec                                         ; was sdf format?
EDF4  2B06                                    bmi     addParms                                                ; yes, put params: cylinders and sides into b:x
                                                                                                                              ; the other varians send these registers with zero value                        
                      ; - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -                                                           
                                                                                                                      ; params to be passed (b:x)
EDF6  D6FA                                    ldb     <valPar                                         ; get b parameter value
EDF8  9E8A                                    ldx     <$8a                                                    ; 16 bits zero                                  
EDFA  2015                                    bra     sendIt                                          ; send the command
                      ; - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -                                                           
                                                                                                                      ; special parameters for command 'create disc' (b:x)
EDFC  9610            addParms        lda     <noexec                                         ; get disc type
EDFE  8403                                    anda    #$03                                                    ; use lowest nibble
EE00  C628                                    ldb     #$28                                                    ; minimum value 40 cylinders
EE02  8103                                    cmpa    #3                                                              ; was type 3 (2880 sectors)?
EE04  2601                                    bne     addP01                                          ; no, skip next
EE06  58                                      aslb                                                                    ; double cylinders number
EE07  8E0100          addP01  ldx     #$0100                                          ; one side by default
EE0A  8101                                    cmpa    #1                                                              ; was type 0 or 1 (630 - 720 sectors)?
EE0C  2303                                    bls     sendIt                                          ; yes, skip next
EE0E  8E0200                                  ldx     #$0200                                          ; set double sided
                      ; - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -                                                           
                                                                                                                      ; send command to the sdc
                                                                                                                              ; b:x have zeros or tracks:sides (at x-high byte)
EE11  96FB            sendIt  lda     <cmdNum                                         ; code for extended command with data block
EE13  9BF8                                    adda    <numUni                                         ; add drive number
EE15  CE0800                                  ldu     #$0800                                          ; point to beginning of data block
EE18  BDF7A4                                  jsr     CommSDC                                         ; send command to sdc-card with (a,b:x,u)
EE1B  2438                                    bcc     send01                                          ; if ok, skip section
                                                                                                                      ; if error $10 and sdir command, show message (not error)
EE1D  C510                                    bitb    #$10                                                    ; is bit 4 set?
EE1F  2714                                    beq     nxtCEr                                          ; no, goto next control
EE21  3406                                    pshs    d                                                               ; save registers
EE23  96FA                                    lda     <valPar                                         ; get parameter used
EE25  8143                                    cmpa    #'C'                                                    ; was it a 'C' (sdir)?
EE27  2608                                    bne     send00                                          ; no, exit
EE29  8D19                                    bsr     currentDirIsRoot                        ; show message
EE2B  3506                                    puls    d                                                               ; restore registers
EE2D  1A01                                    orcc    #%00000001                                      ; set carry to mark error
EE2F  2027                                    bra     send02                                          ; exit
EE31  3506            send00  puls    d                                                               ; clean stack
EE33  2020                                    bra     send01                                          ; exit
                                                                                                                      ; if error $08 and get-dim command, show message (not error)
EE35  C508            nxtCEr  bitb    #$08                                                    ; is bit 3 set?
EE37  271C                                    beq     send01                                          ; no, skip section
EE39  84F0                                    anda    #$f0                                                    ; use high nibble
EE3B  81C0                                    cmpa    #$c0                                                    ; was command $cn (get-dim)?
EE3D  2616                                    bne     send01                                          ; no, go on
EE3F  1A01                                    orcc    #%00000001                                      ; set carry to mark error
EE41  16021A                                  lbra    exsdrv1                                         ; exit
EE44                  currentDirIsRoot                                                                ; current dir = root!
EE44  96F9                                    lda     <numDir                                         ; get number of directories requested in a series
EE46  8101                                    cmpa    #1                                                              ; is first one?
EE48  260A                                    bne     curNoMsg                                                ; no, do not show message
EE4A  3450                                    pshs    x,u                                                     ; save registers
EE4C  8EE47E                                  ldx     #noCurDir-1                                     ; point to message
EE4F  BD90E5                                  jsr     outstr                                          ; show it
EE52  3550                                    puls    x,u                                                     ; restore registers
EE54  39              curNoMsg        rts                                                                     ; return
EE55  17091C          send01  lbsr    vfyStatus                                       ; verify operation result
EE58  0D10            send02  tst     <noexec                                         ; test format type
EE5A  10270200                                lbeq    exsdrv1                                         ; if coco-like 160k, all done, exit. should never happen
EE5E  102B01FC                                lbmi    exsdrv1                                         ; if sdf type, all done, exit   
                      ; - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -                                                           
                                                                                                                      ; additional swrite command for dragon-like disks
                                                                                                                      ; these three need a swrite,39(79),18(36),"",""
                                                                                                                      ; x must have the lsn number: 719,1439,2879
EE62  8E02CF                                  ldx     #719                                                    ; to have 720 sectors
EE65  9610                                    lda     <noexec                                         ; get disc type
EE67  8403                                    anda    #3                                                              ; use lowest 2 bits
EE69  8101                                    cmpa    #1                                                              ; is it a 180k (720 sectors) disc? 
EE6B  270A                                    beq     sendWrt                                         ; yes, send write order
EE6D  8E059F                                  ldx     #1439                                                   ; to force 1440 sectors
EE70  8102                                    cmpa    #2                                                              ; is it a 360k (1440 sectors) disc?
EE72  2703                                    beq     sendWrt                                         ; yes, send write order
EE74  8E0B3F                                  ldx     #2879                                                   ; to force 2880 sectors
                                                                                                      ; *** ADDED CODE TO SUPPORT VDK FORMAT CREATING FILES   
                                                                                                                      ; here we must test the file extension to be ".VDK"
                      ;sendWrt        clr     ,-s                                                     ; make a hole in the stack to contain a VDK flag. Now equals 0 = NO
EE77  0F76            sendWrt clr     <dsktyp                                         ; Now equals 0 = DSK
EE79  CE01D1                                  ldu     #nambuf                                         ; point to filename buffer
EE7C  E6C4                                    ldb     ,u                                                              ; get filename length
EE7E  C003                                    subb    #3                                                              ; to point to the possible dot (.) if .VDK was entered
EE80  33C5                                    leau    b,u                                                     ; point to it
EE82  CC2E56                                  ldd     #".V"                                                   ; get dot and letter V
EE85  10A3C1                                  cmpd    ,u++                                                    ; have we received it?
EE88  260C                                    bne     sWrt01                                          ; no, skip section
EE8A  CC444B                                  ldd     #"DK"                                                   ; get string DK
EE8D  10A3C4                                  cmpd    ,u                                                              ; have we received it?
EE90  2604                                    bne     sWrt01                                          ; no, skip section
                                                                                                                      ; now we know the user wants to create a virtual VDK file
EE92  3001                                    leax    1,x                                                     ; so that the disk has an extra sector (for the header)                                                                 
EE94  0C76                                    inc     <dsktyp                                         ; now is 1 = VDK
                                                                                                      ; *** END_ADDED CODE TO SUPPORT VDK FORMAT CREATING FILES       
EE96  CE0800          sWrt01  ldu     #$0800                                          ; point to beginning of buffer
EE99  4F                                      clra                                                                    ; create a  
EE9A  5F                                      clrb                                                                    ; 16 bits zero
EE9B  EDC1            cleanBuf        std     ,u++                                                    ; clean a word
EE9D  11830900                                cmpu    #$0900                                          ; reached buffers end?
EEA1  25F8                                    blo     cleanBuf                                                ; no, loopback
EEA3  CE0800                                  ldu     #$0800                                          ; point to buffers beginning
EEA6  86A0                                    lda     #$a0                                                    ; command to be sent: SDWRITE
EEA8  9BF8                                    adda    <numUni                                         ; add drive number
EEAA  BDF7A4                                  jsr     CommSDC                                         ; send command to SDC-card      (LSN = B:X)
EEAD  1708C4                                  lbsr    vfyStatus                                       ; verify operation result
                                                                                                      ; *** ADDED CODE TO FORMAT ON THE FLY DSK and VDK FILES
                                                                                                                      ; as the buffer is clean, we just need to fill bytes 0-19
                                                                                                                              ; for a VDK file, we will write a suitable header to sector #0
                                                                                                                              ; it should contain:
                                                                                                                              ;  position             length contents                                  value
EEB0  CC646B                                  ldd     #"dk"                                                   ;       offset 0     2     signature               ('dk')
EEB3  EDC1                                    std     ,u++
EEB5  CC0001                                  ldd     #$0001                                          ;       offset 2     2     Header length           ($00, $01) = $100
EEB8  EDC1                                    std     ,u++
EEBA  CC1010                                  ldd     #$1010                                          ;       offset 4     1     VDK version - actual    ($10)
EEBD  EDC1                                    std     ,u++                                                    ;       offset 5     1     VDK version - compat    ($10)
EEBF  CC8800                                  ldd     #$8800                                          ;       offset 6     1     Source id               ($88)
EEC2  EDC1                                    std     ,u++                                                    ;       offset 7     1     Source version          ($00)
EEC4  108E2801                                ldy     #$2801                                          ; 40 tracks 01 side (180k)
EEC8  9610                                    lda     <noexec                                         ; get disc type
EECA  8403                                    anda    #3                                                              ; use lowest 2 bits
EECC  8101                                    cmpa    #1                                                              ; is it a 180k disc? 
EECE  270A                                    beq     noMore2                                         ; yes, send write order
EED0  3121                                    leay    1,y                                                     ; to force double sided ($2802)
EED2  8102                                    cmpa    #2                                                              ; is it a 360k disc?
EED4  2704                                    beq     noMore2                                         ; yes, send write order
EED6  108E5002                                ldy     #$5002                                          ; it is a 720k disk
EEDA  1F20            noMore2 tfr     y,d                                                     ; pass geodata to regD
EEDC  97F2                                    sta     <$f2                                                    ; number of tracks to DOS system variable
EEDE  5A                                      decb                                                                    ; because 0 means 1 side
EEDF  D7F4                                    stb     <$f4                                                    ; number of sides to DOS system variable
                                                                                                                              ;       offset 8     1     Number of tracks        ($28)          40 or 80 --> ($28  $50)
EEE1  10AFC1                                  sty     ,u++                                                    ;       offset 9     1     Number of sides         ($01)          1 or 2
EEE4  CC0040                                  ldd     #$0040                                          ;       offset 10    1     Flags                   ($00)
EEE7  EDC1                                    std     ,u++                                                    ;       offset 11    1     Comp & Name length      ($40)          bits 0-2 = compression [0=off, >0=TBD]
                                                                                                                              ;                                                                           bits 3-7 = filename length [min 0, max 31]
EEE9  8E01D1                                  ldx     #nambuf                                         ;       offset 12    0-31  Disk name               (...)          (OPTIONAL) filename in ASCII. Doesn't need a zero terminator
EEEC  E680                                    ldb     ,x+                                                     ; get filename length
EEEE  C004                                    subb    #4                                                              ; subtract 4 to have the actual name without dot and extension
EEF0  A680            cpy2buf lda     ,x+                                                     ; get a char from name
EEF2  A7C0                                    sta     ,u+                                                     ; put into buffer
EEF4  5A                                      decb                                                                    ; decrement counter
EEF5  26F9                                    bne     cpy2buf                                         ; if not done, loopback, regB ends with value Zero
EEF7  CE0800                                  ldu     #$0800                                          ; point to buffers beginning
EEFA  86A0                                    lda     #$a0                                                    ; command to be sent: SDWRITE
EEFC  9BF8                                    adda    <numUni                                         ; add drive number
EEFE  8E0000                                  ldx     #$0000                                          ; to point to sector 0 (header)
EF01  BDF7A4                                  jsr     CommSDC                                         ; send command to SDC-card      (LSN = B:X)
EF04  17086D                                  lbsr    vfyStatus                                       ; verify operation result
                                                                                                                      ; Once we have the header, we must add 20 tracks of $e5 bytes
EF07  8E0800                                  ldx     #$0800                                          ; to use this buffer
EF0A  9FEE                                    stx     <$ee                                                    ; update system pointer
EF0C  BDE13A                                  jsr     filbuf                                          ; will fill it with bytes value $E5
EF0F  CC1412                                  ldd     #$1412                                          ; get number of tracks to write (20) and numsectors x track
EF12  0DF4                                    tst     <$f4                                                    ; is single sided?
EF14  2701                                    beq     noDouble                                                ; yes, do not double
EF16  58                                      aslb                                                                    ; double number of sectors x track
EF17  3D              noDouble        mul                                                                     ; calculate total number of sector to write
EF18  1F02                                    tfr     d,y                                                     ; pass value to regY
EF1A  8E0001                                  ldx     #1                                                              ; 1st sector (LSN) to be written for a VDK file
EF1D  0D76                                    tst     <dsktyp                                         ; are we creating a VDK?
EF1F  2602                                    bne     nSector                                         ; yes, skip next
EF21  301F                                    leax    -1,x                                                    ; 1st LSN for a DSK file (and so rewrite the header!)
EF23  3430            nSector pshs    x,y                                                     ; save variables
EF25  CE0800                                  ldu     #$0800                                          ; point to buffers beginning
EF28  86A0                                    lda     #$a0                                                    ; command to be sent: SDWRITE
EF2A  9BF8                                    adda    <numUni                                         ; add drive number
EF2C  5F                                      clrb                                                                    ; higher byte to zero
EF2D  BDF7A4                                  jsr     CommSDC                                         ; send command to SDC-card      (LSN = B:X)
EF30  170841                                  lbsr    vfyStatus                                       ; verify operation result
EF33  3530                                    puls    x,y                                                     ; restore variables
EF35  3001                                    leax    1,x                                                     ; point to next LSN
EF37  313F                                    leay    -1,y                                                    ; decrement counter
EF39  26E8                                    bne     nSector                                         ; if not done, loopback 
EF3B  3410                                    pshs    x                                                               ; put next LSN on stack
                                                                                                                      ; now we must add 1st FAT sector
EF3D  BDE17D                                  jsr     ffat1                                                   ; buffer: $800. fills 1st FAT sector upon geodata
EF40  CE0800                                  ldu     #$0800                                          ; point to buffers beginning
EF43  86A0                                    lda     #$a0                                                    ; command to be sent: SDWRITE
EF45  9BF8                                    adda    <numUni                                         ; add drive number
EF47  5F                                      clrb                                                                    ; higher byte to zero. X is the good number
EF48  AEE4                                    ldx     ,s                                                              ; get LSN from stack
EF4A  BDF7A4                                  jsr     CommSDC                                         ; send command to SDC-card      (LSN = B:X)
EF4D  170824                                  lbsr    vfyStatus                                       ; verify operation result
                                                                                                                      ; let's try the same for track 16 (backup DIR)
                                                                                                                      ; we have to calculate the first sector for that track
EF50  8610                                    lda     #16                                                     ; number of track to write on
EF52  C612                                    ldb     #18                                                     ; number of sectors x track if single sided disk
EF54  0DF4                                    tst     <$f4                                                    ; is double sided?
EF56  2701                                    beq     1f                                                              ; no, skip next
EF58  58                                      aslb                                                                    ; double sectors x track
EF59  3D              1                       mul                                                                     ; to get sectors before sector 1 track 16
EF5A  0D76                                    tst     <dsktyp                                         ; is it a VDK?
EF5C  2703                                    beq     2f                                                              ; no, skip next
EF5E  C30001                                  addd    #1                                                              ; add a sector (header)
EF61  DDD9            2                       std     <fST16                                          ; save that value for later use
                                                                                                                      ; now write 1st sector Track 16
EF63  CE0800                                  ldu     #$0800                                          ; point to buffers beginning
EF66  86A0                                    lda     #$a0                                                    ; command to be sent: SDWRITE
EF68  9BF8                                    adda    <numUni                                         ; add drive number
EF6A  5F                                      clrb                                                                    ; higher byte to zero. X is the good number
EF6B  9ED9                                    ldx     <fST16                                          ; get LSN from variable
EF6D  BDF7A4                                  jsr     CommSDC                                         ; send command to SDC-card      (LSN = B:X)
EF70  170801                                  lbsr    vfyStatus                                       ; verify operation result
                                                                                                                      ; now we must add 2nd FAT sector
EF73  BDE1D2                                  jsr     ffat2                                                   ; buffer: $900. fills 2nd FAT sector upon geodata
EF76  CE0900                                  ldu     #$0900                                          ; point to buffers beginning
EF79  86A0                                    lda     #$a0                                                    ; command to be sent: SDWRITE
EF7B  9BF8                                    adda    <numUni                                         ; add drive number
EF7D  5F                                      clrb                                                                    ; higher byte to zero. X is the good number
EF7E  AEE4                                    ldx     ,s                                                              ; get previous LSN from stack
EF80  3001                                    leax    1,x                                                     ; increment number
EF82  AFE4                                    stx     ,s                                                              ; update it
EF84  BDF7A4                                  jsr     CommSDC                                         ; send command to SDC-card      (LSN = B:X)
EF87  1707EA                                  lbsr    vfyStatus                                       ; verify operation result
                                                                                                                      ; and do the same for Track 16
EF8A  CE0900                                  ldu     #$0900                                          ; point to buffers beginning
EF8D  86A0                                    lda     #$a0                                                    ; command to be sent: SDWRITE
EF8F  9BF8                                    adda    <numUni                                         ; add drive number
EF91  5F                                      clrb                                                                    ; higher byte to zero. X is the good number
EF92  9ED9                                    ldx     <fST16                                          ; get previous LSN from variable
EF94  3001                                    leax    1,x                                                     ; increment number
EF96  9FD9                                    stx     <fST16                                          ; update it
EF98  BDF7A4                                  jsr     CommSDC                                         ; send command to SDC-card      (LSN = B:X)
EF9B  1707D6                                  lbsr    vfyStatus                                       ; verify operation result
                                                                                                                      ; now we go for the 16 directory sectors with entries ...
EF9E  BDE15E                                  jsr     fil16s                                          ; buffer: $a00. fills a sector to be used as diretory sector (3-18)
EFA1  C610                                    ldb     #16                                                     ; number of sectors to write
EFA3  3510                                    puls    x                                                               ; get previous LSN from stack
EFA5  3001                                    leax    1,x                                                     ; increment LSN number
EFA7  3414            nDSect  pshs    b,x                                                     ; save variables
EFA9  CE0A00                                  ldu     #$0a00                                          ; point to buffers beginning
EFAC  86A0                                    lda     #$a0                                                    ; command to be sent: SDWRITE
EFAE  9BF8                                    adda    <numUni                                         ; add drive number
EFB0  5F                                      clrb                                                                    ; higher byte to zero. X is the good number
EFB1  BDF7A4                                  jsr     CommSDC                                         ; send command to SDC-card      (LSN = B:X)
EFB4  1707BD                                  lbsr    vfyStatus                                       ; verify operation result
EFB7  3514                                    puls    b,x                                                     ; restore variables
EFB9  3001                                    leax    1,x                                                     ; increment LSN number
EFBB  5A                                      decb                                                                    ; decrement sector counter
EFBC  26E9                                    bne     nDSect                                          ; not done, loop
EFBE  3410                                    pshs    x                                                               ; save next LSN (rest of the disk)
                                                                                                                      ; and do the same for Track16 too
EFC0  C610                                    ldb     #16                                                     ; number of sectors to write
EFC2  9ED9                                    ldx     <fST16                                          ; get previous LSN from variable
EFC4  3001                                    leax    1,x                                                     ; increment LSN number
EFC6  3414            n16Sect pshs    b,x                                                     ; save variables
EFC8  CE0A00                                  ldu     #$0a00                                          ; point to buffers beginning
EFCB  86A0                                    lda     #$a0                                                    ; command to be sent: SDWRITE
EFCD  9BF8                                    adda    <numUni                                         ; add drive number
EFCF  5F                                      clrb                                                                    ; higher byte to zero. X is the good number
EFD0  BDF7A4                                  jsr     CommSDC                                         ; send command to SDC-card      (LSN = B:X)
EFD3  17079E                                  lbsr    vfyStatus                                       ; verify operation result
EFD6  3514                                    puls    b,x                                                     ; restore variables
EFD8  3001                                    leax    1,x                                                     ; increment LSN number
EFDA  5A                                      decb                                                                    ; decrement sector counter
EFDB  26E9                                    bne     n16Sect                                         ; not done, loop
                                                                                                                      ; finally we will write the rest of the sectors with $e5
                                                                                                                      ; X points to next LSN. end od disk might be: 720, 1440, 2880 (due to the header)
EFDD  8E0800                                  ldx     #$0800                                          ; to use this buffer again
EFE0  9FEE                                    stx     <$ee                                                    ; update system pointer
EFE2  BDE13A                                  jsr     filbuf                                          ; will fill it with bytes value $E5
EFE5  3510                                    puls    x                                                               ; get LSN for rest of the disk from stack
EFE7  108E02D1                                ldy     #721                                                    ; value to detect end of disk for 180k disk
EFEB  9610                                    lda     <noexec                                         ; get disc type
EFED  8403                                    anda    #3                                                              ; use lowest 2 bits
EFEF  8101                                    cmpa    #1                                                              ; is it a 180k (720 sectors) disk? 
EFF1  270C                                    beq     1f                                                              ; yes, send write order
EFF3  108E05A1                                ldy     #1441                                                   ; to detect end of 360 disk
EFF7  8102                                    cmpa    #2                                                              ; is it a 360k (1440 sectors) disk?
EFF9  2704                                    beq     1f                                                              ; yes, send write order
EFFB  108E0B41                                ldy     #2881                                                   ; to detect end of 720k disk
EFFF  0D76            1                       tst     <dsktyp                                         ; are we creating a VDK file?
F001  2602                                    bne     nRDisk                                          ; yes, skip next
F003  313F                                    leay    -1,y                                                    ; subtract 1 (no header for DSK)
F005  3430            nRDisk  pshs    x,y                                                     ; save registers (LSN, endOfFile)
F007  CE0800                                  ldu     #$0800                                          ; point to buffers beginning
F00A  86A0                                    lda     #$a0                                                    ; command to be sent: SDWRITE
F00C  9BF8                                    adda    <numUni                                         ; add drive number
F00E  5F                                      clrb                                                                    ; higher byte to zero. X is the good number
F00F  BDF7A4                                  jsr     CommSDC                                         ; send command to SDC-card      (LSN = B:X)
F012  17075F                                  lbsr    vfyStatus                                       ; verify operation result
F015  3510                                    puls    x                                                               ; restore LSN number
F017  3001                                    leax    1,x                                                     ; increment LSN number
F019  ACE4                                    cmpx    ,s                                                              ; is X greater than Y (in the stack)
F01B  3520                                    puls    y                                                               ; to balance stack
F01D  26E6                                    bne     nRDisk                                          ; not done, loop
                                                                                                                      ; call SDRIVE n,UNLOAD
                                                                                                                      ; first we must save the filename from nambuf or it will be lost
                                                                                                                      ; we need from nambuf to nambuf + 12 (8 plus dot plus VDK)
F01F  8E0A00                                  ldx     #$a00                                                   ; point to last DOS buffer
F022  CE01D1                                  ldu     #nambuf                                         ; point to filename buffer
F025  C60D                                    ldb     #13                                                     ; number of bytes to copy
F027  A6C0            1                       lda     ,u+                                                     ; get a byte from filename buffer
F029  A780                                    sta     ,x+                                                     ; put into DOS buffer
F02B  5A                                      decb                                                                    ; decrement counter
F02C  26F9                                    bne     1b                                                              ; not done, loop
F02E  0FFA                                    clr     <valPar                                         ; value param is 0
F030  C6E0                                    ldb     #$e0                                                    ; more used value
F032  D7FB                                    stb     <cmdNum                                         ; store it
F034  BDED38                                  jsr     reuseUL                                         ; call final part of SDRIVEx,UNLOAD
F037  8E0A00                                  ldx     #$a00                                                   ; point to last DOS buffer
F03A  CE01D1                                  ldu     #nambuf                                         ; point to filename buffer
F03D  C60D                                    ldb     #13                                                     ; number of bytes to copy
F03F  A680            1                       lda     ,x+                                                     ; get a byte from DOS buffer
F041  A7C0                                    sta     ,u+                                                     ; put into filename buffer
F043  5A                                      decb                                                                    ; decrement counter
F044  26F9                                    bne     1b                                                              ; not done, loop
F046  0FFA                                    clr     <valPar                                         ; value param is 0
F048  C6E0                                    ldb     #$e0                                                    ; more used value
F04A  D7FB                                    stb     <cmdNum                                         ; store it
F04C  BDED3E                                  jsr     dTok01                                  ; call SDRIVE n,"received FILENAME.VDK"
F04F  9677                                    lda     hdwSdc                                          ; is here the old slot?
F051  2703                                    beq     1f                                                              ; no, exit
F053  B7FF7F                                  sta     $ff7f                                                   ; active it
F056  39              1                       rts                                                                     ; return
                                                                              ; *** END_ADDED CODE TO SUPPORT VDK AND DSK FORMAT WHILE CREATING FILES 
                      ; - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -                                                           
                                                                                                                      ; restore old slot if needed
F057  9677            exsdrv  lda     hdwSdc                                          ; is here the old slot?
F059  2703                                    beq     exsdrv1                                         ; no, exit
F05B  B7FF7F                                  sta     $ff7f                                                   ; active it
F05E  39              exsdrv1 rts                                                                     ; return
                      ; = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = =                                                           
                                                                                                                      ; special command floppy off
F05F  8D0F            flopOF  bsr     flopCom                                         ; see if it could be done
F061  FA0114                                  orb     floSDC                                          ; add to hdw control byte (a bit to 1 disables floppy and enables sdc)
F064  F70114          eflop           stb     floSDC                                          ; update hdw control byte
F067  39                                      rts                                                                     ; return
                      ; = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = =                                                           
                                                                                                                      ; special command floppy on
F068  8D06            flopON  bsr     flopCom                                         ; see if it could be done
F06A  53                                      comb                                                                    ; invert all bits: %00000001 - %00000010 convert to %11111110 - %11111101
F06B  F40114                                  andb    floSDC                                          ; a bit 0 enables floppy and disables sdc
F06E  20F4                                    bra   eflop                                                     ; update and return
                      ; - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -                                                           
                                                                                                                      ; common part for floppy on-off
F070  9D9F            flopCom jsr     <getNChr                                                ; skip token to avoid sn error when returning
F072  F60114                                  ldb     floSDC                                          ; get hdw control byte
F075  C40C                                    andb    #%00001100                                      ; see if switching is possible
F077  C10C                                    cmpb    #%00001100                                      ; are both floppies and sdc present?
F079  2604                                    bne     notAllow                                                ; no, show error message
F07B  D6F8                                    ldb     <numUni                                         ; get drive number (0-1)
F07D  5C                                      incb                                                                    ; convert to 1-2
F07E  39                                      rts                                                                     ; return
F07F  3262            notAllow leas   2,s                                                     ; clean stack (is inside one subroutine)
F081  8EE493                                  ldx     #canNot-1                                       ; point to message
F084  7E90E5                                  jmp     outstr                                          ; show it and return
                      ; = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = =                                                           
                                                                                                                      ; special command dim
F087  8651            sdcDIM  lda     #'Q'                                                    ; param to send into b
F089  8D14                                    bsr     sdcDAT                                          ; process it
F08B  2527                                    bcs     noDiscEr                                                ; if got "not disc" error, explain it
F08D  1700F7                                  lbsr    shwDIM                                          ; show received data
F090  8DC5            rstSlot bsr     exsdrv                                          ; restore slot if needed
F092  1CFE                                    andcc   #%11111110                                      ; clear carry flag
F094  39                                      rts                                                                     ; return
                      ; = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = =                                                           
                                                                                                                      ; special command get
F095  8649            sdcGET  lda     #'I'                                                    ; param to send into b
F097  8D06                                    bsr     sdcDAT                                          ; process it
F099  2519                                    bcs     noDiscEr                                                ; if got "not disc" error, explain it
F09B  8D24                                    bsr     shwGET                                          ; show received data
F09D  20F1                                    bra     rstSlot                                         ; restore slot if needed
                      ; = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = =                                                           
                                                                                                                      ; common part for get-dim
F09F  97FA            sdcDAT  sta     <valPar                                         ; save it
F0A1  9D9F                                    jsr     <getNChr                                                ; skip read token
F0A3  86C0                                    lda     #$c0                                                    ; command to send
F0A5  97FB                                    sta     <cmdNum                                         ; save it
F0A7  CC0100                                  ldd     #$0100                                          ; get 1 byte with value equal null
F0AA  FD01D1                                  std     nambuf                                          ; put in nambuf area (no name)
F0AD  0F10                                    clr     <noexec                                         ; no additional command needed
F0AF  DC8A                                    ldd     <$8a                                                    ; get 16 bits zero
F0B1  7EEDA5                                  jmp     putCmd2                                         ; send command
                      ; - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -                                                           
                                                                                                                      ; message no disc in drive
F0B4  96F8            noDiscEr        lda     <numUni                                         ; gets 0-1
F0B6  8B31                                    adda    #$31                                                    ; converts to '1'-'2'
F0B8  BDB54A                                  jsr     outchr                                          ; show drive number
F0BB  8EE4A1                                  ldx     #sdcNoDsc-1                                     ; point to message
F0BE  7E90E5                                  jmp     outstr                                          ; show it
                      ; - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -                                                           
                                                                                                                      ; show received 'get' info
F0C1  96F8            shwGET  lda     <numUni                                         ; get drive number
F0C3  8B31                                    adda    #$31                                                    ; converts to '1'-'2'
F0C5  BDB54A                                  jsr     outchr                                          ; show drive number
F0C8  863A                                    lda     #':'                                                    ; get char ':'
F0CA  BDB54A                                  jsr     outchr                                          ; show it
                                                                                                                      ; show drive status
F0CD  866F                                    lda     #'o'                                                    ; get fixed 1st letter
F0CF  BDB54A                                  jsr     outchr                                          ; show it
F0D2  866E                                    lda     #'n'                                                    ; get on as default
F0D4  D6F8                                    ldb     <numUni                                         ; get drive unit
F0D6  5C                                      incb                                                                    ; convert  0-1 into 1-2
F0D7  F40114                                  andb    floSDC                                          ; test corresponding bit of hdw control byte
F0DA  2607                                    bne     sdcS00                                          ; if not zero, skip next
F0DC  8666                                    lda     #'f'                                                    ; get value for off
F0DE  BDB54A                                  jsr     outchr                                          ; put first 'f'
F0E1  200D                                    bra     sdcS01                                          ; go put 2nd one instead of space
F0E3  BDB54A          sdcS00  jsr     outchr                                          ; show result
F0E6  8620                                    lda     #32                                                     ; get space as inverted block to poke it
F0E8  9E88                                    ldx     <$88                                                    ; get cursor address
F0EA  A780                                    sta     ,x+                                                     ; put inverse space there
F0EC  9F88                                    stx     <$88                                                    ; update cusor address
F0EE  2003                                    bra     sdcS02                                          ; skip next
F0F0  BDB54A          sdcS01  jsr     outchr                                          ; show it
F0F3  8D30            sdcS02  bsr     shwFdat                                         ; show filename and type
F0F5  33C811                                  leau    17,u                                                    ; jump to length bytes
F0F8  8D06                                    bsr     shwFL                                                   ; show file length
F0FA  860D                                    lda     #$0d                                                    ; get enter value
F0FC  BDB54A                                  jsr     outchr                                          ; print it
F0FF  39                                      rts                                                                     ; return
                      ; - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -                                                           
                                                                                                                      ; show file length - structure of the fac
                                                                                                                      ;       $50 highest byte        --- always $00
                                                                                                                      ;       $51 high byte           --- high byte of u
                                                                                                                      ;       $52 mid  byte           --- low  byte of u
                                                                                                                      ;       $53 lower byte          --- reg. a ($00 = multiple of 256)
F100  E6C4            shwFL           ldb     ,u                                                              ; get mid (2nd byte)
F102  A641                                    lda     1,u                                                     ; get 3rd byte (higher than 2nd)
F104  1F03                                    tfr     d,u                                                     ; pass to u
F106  4F                                      clra                                                                    ; lower byte = 0 (always n x 256 bytes!)
F107  BDDD7D          shwNum  jsr     $dd7d                                                   ; integer to fac and normalize it
F10A  BD9587                                  jsr     $9587                                                   ; fac to ascii (x points to string ended with $00
                                                                                                                      ; calculate length and add spaces to have total length = 7
F10D  3101                                    leay    1,x                                                     ; point to first digit
F10F  C608                                    ldb     #8                                                              ; desired length + final null
F111  5A              shwN01  decb                                                                    ; decrement counter
F112  A6A0                                    lda     ,y+                                                     ; get a digit
F114  26FB                                    bne     shwN01                                          ; if not 0, loopback
F116  5D                                      tstb                                                                    ; needs extra spaces?
F117  2708                                    beq     shwN02                                          ; no, skip section
F119  8620                                    lda     #32                                                     ; get space
F11B  BDB54A          shwN03  jsr     outchr                                          ; print it
F11E  5A                                      decb                                                                    ; decrement counter
F11F  26FA                                    bne     shwN03                                          ; if not zero, loopback
F121  BD90E5          shwN02  jsr     outstr                                          ; print lengths' string
F124  39                                      rts                                                                     ; return
                      ; - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -           
                                                                                                                      ; show filename and extension                   
F125  CE0800          shwFdat ldu     #$800                                                   ; point to beginning of data
F128  C608            shwFd01 ldb     #8                                                              ; 8 chars for name
F12A  A6C0            sdc01           lda     ,u+                                                     ; get a char
F12C  BDB54A                                  jsr     outchr                                          ; show it
F12F  5A                                      decb                                                                    ; decrement counter
F130  26F8                                    bne     sdc01                                                   ; not all, loopback
F132  862E                                    lda     #'.'                                                    ; get a dot
F134  BDB54A                                  jsr     outchr                                          ; show it
F137  C603                                    ldb     #3                                                              ; 3 chars for extension
F139  A6C0            sdc02           lda     ,u+                                                     ; get a char
F13B  BDB54A                                  jsr     outchr                                          ; show it
F13E  5A                                      decb                                                                    ; decrement counter
F13F  26F8                                    bne     sdc02                                                   ; not all, loopback
                                                                                                                      ; show file type and extra info
F141  E6C0                                    ldb     ,u+                                                     ; get byte 11 (type)
F143  8620                                    lda     #32                                                     ; get space
F145  C102                                    cmpb    #$02                                                    ; is byte hidden?
F147  2602                                    bne     sdc05                                                   ; no, skip next 
F149  8668                                    lda     #'h'                                                    ; get h char
F14B  BDB54A          sdc05           jsr     outchr                                          ; show it
F14E  8620                                    lda     #32                                                     ; get space
F150  C101                                    cmpb    #$01                                                    ; is it locked?
F152  2602                                    bne     sdc06                                                   ; no, skip next 
F154  866C                                    lda     #'l'                                                    ; get l char
F156  BDB54A          sdc06           jsr     outchr                                          ; show it
F159  C110                                    cmpb    #$10                                                    ; is it a directory?
F15B  270C                                    beq     sdc04                                                   ; yes, get that text
F15D  C104                                    cmpb    #$04                                                    ; is it sdf type?
F15F  2704                                    beq     sdc03                                                   ; yes, get that text
F161  8EE4CA                                  ldx     #datNUL-1                                       ; else get spaces
F164  10                                      fcb     #$10                                                    ; to skip next
F165  8EE4C0          sdc03           ldx     #datSDF-1                                       ; point to message
F168  10                                      fcb     #$10                                                    ; to skip next
F169  8EE4C5          sdc04           ldx     #datDIR-1                                       ; point to message
F16C  3440                                    pshs    u                                                               ; save pointer
F16E  BD90E5                                  jsr     outstr                                          ; put the three chars
F171  3540                                    puls    u                                                               ; restore pointer
F173  39                                      rts                                                                     ; return
                      ; - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -                                                           
                                                                                                                      ; command sdrive alone
F174  0FF8            sdcLST  clr     <numUni                                         ; set drive 1
F176  17FF1C                                  lbsr    sdcGET                                          ; get and show info for that drive
                      ;                       bcs     esdcLST                                         ; if carry set, exit (no sdc present)
F179  210B                                    brn     esdcLST                                         ; if carry set, exit (no sdc present)
F17B  9EA6                                    ldx     <$a6                                                    ; get input pointer
F17D  301E                                    leax    -2,x                                                    ; move it back (will be moved forward later)
F17F  9FA6                                    stx     <$a6                                                    ; update it
F181  0CF8                                    inc     <numUni                                         ; det drive 2                   
F183  17FF0F                                  lbsr    sdcGET                                          ; get and show info for that drive
F186  39              esdcLST rts                                                                     ; return
                      ; - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -                                                           
                                                                                                                      ; show received 'dim' info
F187  96F8            shwDIM  lda     <numUni                                         ; get drive number
F189  8B31                                    adda    #$31                                                    ; converts to '1'-'2'
F18B  BDB54A                                  jsr     outchr                                          ; show drive number
F18E  8EE4AF                                  ldx     #totSect-1                                      ; point to message
F191  BD90E5                                  jsr     outstr                                          ; show it
F194  FEFF41                                  ldu     $ff41                                                   ; get two higher bytes
F197  B6FF43                                  lda     $ff43                                                   ; get lower byte
F19A  17FF6A                                  lbsr    shwNum                                          ; process and show it
F19D  7E90A5                                  jmp     sendCR                                          ; send cr
                      ; = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = =                                                           
                                                                                                                      ; command smkdir
F1A0  8D09            xsmkdir bsr     others                                          ; read filename, get params
F1A2  CC4B3A                                  ldd     #"K:"                                                   ; command to send
F1A5  BDEDA5          xsmk01  jsr     putCmd2                                         ; send command
F1A8  16FEE5                                  lbra    rstSlot                                         ; restore slot if needed
                      ; - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -                                                           
                                                                                                                      ; read filename and set parameters                      
F1AB  BDB7AA          others  jsr   getfnam                                           ; read string and length into $1d1 = nambuf
F1AE  F601D1                  ldb     nambuf                                          ; is string length 0?
F1B1  1027F5B7                lbeq    snError                                         ; yes, show error
F1B5  C1FD                    cmpb    #253                                                    ; is string longer than maximum accepted? (253 + m: + final null)
F1B7  2505                                    bcs     other01                                         ; no, skip section
F1B9  3262                                    leas    2,s                                                     ; clean stack (we are inside a subroutine)
F1BB  16F939                                  lbra    TooLong                                         ; show error message
F1BE  5F              other01 clrb                                                                    ; set to zero
F1BF  D7FA                                    stb     <valPar                                         ; parameter for b
F1C1  D710                                    stb     <noexec                                         ; no more parmas
F1C3  D7F8                                    stb     <numUni                                         ; drive 0
F1C5  86E0                                    lda     #$e0                                                    ; command to send
F1C7  97FB                                    sta     <cmdNum                                         ; save it
F1C9  39                                      rts                                                                     ; return
                      ; = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = =                                                           
                                                                                                                      ; command skill
F1CA  8DDF            xskill  bsr     others                                          ; read filename, get params
F1CC  CC583A                                  ldd     #"X:"                                                   ; command to send
F1CF  20D4                                    bra     xsmk01                                          ; go send command
                      ; = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = =                                                           
                                                                                                                      ; command sren
F1D1  BDB7AA          xsren           jsr     getfnam                                         ; get old name
F1D4  F601D1                  ldb     nambuf                                          ; is string length 0?
F1D7  1027F591                lbeq    snError                                         ; yes, show error
F1DB  C17D                    cmpb    #125                                                    ; is string longer than maximum accepted? (125 for each one)
F1DD  1024F916                lbcc    TooLong                                         ; show error message
                                                                                                                      ; send it to $800 ended with $00
F1E1  8E01D2                                  ldx   #nambuf+1                                 ; point to 1st char of the input string
F1E4  CE0800                                  ldu     #$0800                                          ; point to 1st dos buffer
F1E7  F601D1                                  ldb     nambuf                                          ; get string length
F1EA  A680            xsren01 lda     ,x+                                                     ; get a char from input string
F1EC  A7C0                                    sta     ,u+                                                     ; copy in the buffer
F1EE  5A                                      decb                                                                    ; decrement counter
F1EF  26F9                                    bne     xsren01                                         ; not yet done, loopback
F1F1  6FC0                                    clr     ,u+                                                     ; put a final zero
F1F3  DF76                                    stu     <bufPtr                                         ;save pointer
F1F5  BDB7AA                                  jsr     getfnam                                         ; get new name
F1F8  F601D1                  ldb     nambuf                                          ; is string length 0?
F1FB  1027F56D                lbeq    snError                                         ; yes, show error
F1FF  C17D                    cmpb    #125                                                    ; is string longer than maximum accepted? (125 for each one)
F201  1024F8F2                lbcc    TooLong                                         ; show error message
                                                                                                                      ; add after old name
F205  8E01D2                                  ldx   #nambuf+1                                 ; point to 1st char of the input string
F208  DE76                                    ldu     <bufPtr                                         ; get old pointer
F20A  F601D1                                  ldb     nambuf                                          ; get newname string length
F20D  A680            xsren02 lda     ,x+                                                     ; get a char from input string
F20F  A7C0                                    sta     ,u+                                                     ; copy in the buffer
F211  5A                                      decb                                                                    ; decrement counter
F212  26F9                                    bne     xsren02                                         ; not yet done, loopback
F214  DF76                                    stu     <bufPtr                                         ; save last byte received
                                                                                                                      ; send it back to nambuf+1 - put length in nambuf 
F216  8E01D2                                  ldx     #nambuf+1                                       ; point to 1st char in input buffer                                                                             
F219  CE0800                                  ldu     #$800                                                   ; point to 1st byte of dos buffer
F21C  5F                                      clrb                                                                    ; counter to zero
F21D  A6C0            xsren03 lda     ,u+                                                     ; get a byte
F21F  A780                                    sta     ,x+                                                     ; put on the input buffer
F221  5C                                      incb                                                                    ; increment counter
F222  119376                                  cmpu    <bufPtr                                         ; got to the end?
F225  25F6                                    blo     xsren03                                         ; no, loopback
F227  F701D1                                  stb     nambuf                                          ; save total string length 
F22A  8D92                                    bsr     other01                                         ; set parameters
F22C  CC523A                                  ldd     #"R:"                                                   ; command to send
F22F  16FF73                                  lbra    xsmk01                                          ; go send command
                      
                      ;--------------------------------------------------------------------------
                      ; routine to add extension to the file name if not present
                      ; param: noexec=$00 to add .DSK  --- noexec=$80 to add .SDF
                      ;--------------------------------------------------------------------------
F232  3416            addExt  pshs    d,x                                                     ; save pointers
F234  8E01D1                                  ldx     #nambuf                                         ; poit to name buffer
F237  E680                                    ldb   ,x+                                                       ; length of name
                                                                                                                      ; Check for existing extension
F239  A680            isadot          lda   ,x+                                                       ; get a char from name
F23B  812E                    cmpa  #'.'                                                      ; is it a dot?
F23D  2727                    beq   eAddExt                                           ; yes, exit
F23F  5A                      decb                                                                    ; decrement counter
F240  26F7                    bne   isadot                                            ; not end of string, get next char name
                                                                                                                      ; so, we must add an extension
F242  0D10                    tst     <noexec                                         ; is for a SDF file?
F244  2B0C                    bmi     addSDF                                          ; yes, jump section
F246  CC2E44          addDSK          ldd   #$2e44                                            ; get chars ".D"
F249  ED81                    std   ,x++                                                      ; add to end of name
F24B  CC534B                  ldd   #$534b                                            ; get chars "SK"
F24E  ED81                    std   ,x++                                                      ; add to end of name
F250  200A                                    bra     add01                                                   ; skip next section
F252  CC2E53          addSDF          ldd   #$2e53                                            ; get chars ".S"
F255  ED81                    std   ,x++                                                      ; add to end of name
F257  CC4446                  ldd   #$4446                                            ; get chars "DF"
F25A  ED81                    std   ,x++                                                      ; add to end of name
F25C  6F84            add01           clr     ,x                                                              ; put a null at next byte
F25E  F601D1                                  ldb   nambuf                                            ; get old name length
F261  CB04                                    addb  #4                                                                ; add 4 (because of dot and extension added)
F263  F701D1                  stb   nambuf                                            ; save new length in byte before string
F266  3596            eAddExt puls  d,x,pc                                            ; restore pointers and return
                      ;--------------------------------------------------------------------------
                      ; Section to deal with SDC-card SCHD commands
                      ;--------------------------------------------------------------------------
F268  9DA5            xschd           jsr     <redLChr                                                ; peek at next char after SCHD
F26A  1027F4FE                                lbeq    snError                                         ; if null, syntax error
F26E  8122                                    cmpa    #$22                                                    ; is this the double quote?  = new directory
F270  10270123                                lbeq    setCD01                                         ; do the same as SDIR = (by now!)
F274  81C6                                    cmpa    #$c6                                                    ; is it the token for /?
F276  271C                                    beq     chgRoot                                         ; process it
F278  812E                                    cmpa    #'.'                                                    ; is it a dot '.'?
F27A  1026F4EE                                lbne    snError                                         ; no, show syntax error
F27E  9D9F                                    jsr     <getNChr                                                ; get next char
F280  812E                                    cmpa    #'.'                                                    ; is it another dot '.'?
F282  1026F4E6                                lbne    snError                                         ; no, show syntax error
                                                                                                                      ; this is to set parent dir as actual
F286  9D9F            chgParD jsr     <getNChr                                                ; skip token
F288  8602            chwPD01 lda     #$02                                                    ; only one char in nambuf
F28A  C62E                                    ldb     #'.'                                                    ; char to design root directory (twice)
F28C  FD01D1                                  std     nambuf                                          ; put signs
F28F  F701D3                                  stb     nambuf+2                                                ; into input buffer
F292  2009                                    bra     chgR01                                          ; process it
                                                                                                                      ; this is to set root as actual dir
F294  9D9F            chgRoot jsr     <getNChr                                                ; skip token
F296  8601                                    lda     #$01                                                    ; only one char in nambuf
F298  C62F                                    ldb     #'/'                                                    ; char to design root directory
F29A  FD01D1                                  std     nambuf                                          ; put data in input buffer area
F29D  17FF1E          chgR01  lbsr    other01                                         ; set parameters
F2A0  1600F7                                  lbra    setCD02                                         ; go to process the requested directory
                      ;--------------------------------------------------------------------------
                      ; Section to deal with SDC-card SDIR commands
                      ;--------------------------------------------------------------------------
00FA                  itsRead equ     $fa                                                     ; number of received items in a dir page (re-used)
00FB                  itsDone equ     $fb                                                     ; number of items shown of a page
                      
F2A3  9DA5            xsdir           jsr     <redLChr                                                ; peek at next char after sdir
F2A5  102700E6                                lbeq    shwCAct                                         ; is the sdir alone = show current dir contents
                                                                                                                      ; decode token - sign
F2A9  81B3                                    cmpa    #$b3                                                    ; is the token for get?
F2AB  102700AF                                lbeq    shwCDir                                         ; yes, show current dir name
F2AF  8187                                    cmpa    #$87                                                    ; is it the token for print (question mark)?
F2B1  102700EB                                lbeq    shwFull                                         ; yes, show the full path name of actual dir (maybe substitute sdir get!)
F2B5  81C4                                    cmpa    #$c4                                                    ; is this the token for ('-')? = show contents of root dir
F2B7  102700C5                                lbeq    shwRoot                                         ; yes, go process it
F2BB  81CB                                    cmpa    #$cb                                                    ; is this the token for ('=')? = set current dir
F2BD  102700D4                                lbeq    setCDir                                         ; yes, go process it
F2C1  8122                                    cmpa    #$22                                                    ; is this the sign ('"')?  = show contents of the requested dir
F2C3  1026F4A5                                lbne    snError                                         ; no, show syntax error
                      ; - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -                                                           
                                                                                                                      ; show contents of requested dir
                                                                                                                      ; send a initiate directory listing command
F2C7  17FEE1          shwRDir lbsr    others                                          ; read filename, get params
F2CA  CC4C3A          shwEDir ldd     #"L:"                                                   ; command to send
F2CD  BDEDA5                                  jsr     putCmd2                                         ; send command
                      ; - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -                                                           
                                                                                                                      ; ask for directory pages to the sdc
F2D0  17FEEB          shwRD00 lbsr    other01                                         ; set some parameters
F2D3  C63E                                    ldb     #'>'                                                    ; special parameter
F2D5  D7FA                                    stb     <valPar                                         ; save it
F2D7  86C0                                    lda     #$c0                                                    ; command to send
F2D9  97FB                                    sta     <cmdNum                                         ; save it
F2DB  CC0100                                  ldd     #$0100                                          ; mark 1 byte with value equal null
F2DE  FD01D1                                  std     nambuf                                          ; put in nambuf area (no name)
F2E1  BDEDA5                                  jsr     putCmd2                                         ; send command
                      ; - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -                                                           
F2E4  8E0800                                  ldx     #$0800                                          ; point to buffer beginning
F2E7  5F                                      clrb                                                                    ; counter to zero
F2E8  A684            shwRD01 lda     ,x                                                              ; get first char (1st name character)
F2EA  2709                                    beq     shwRD02                                         ; if zero, no more items
F2EC  5C                                      incb                                                                    ; increment received items counter
F2ED  308810                                  leax    16,x                                                    ; point to next item
F2F0  8C0900                                  cmpx    #$0900                                          ; got to end of buffer?
F2F3  25F3                                    blo     shwRD01                                         ; no, loopback
F2F5  5D              shwRD02 tstb                                                                    ; have we read any item?
F2F6  2762                                    beq     shwRD09                                         ; no, print ok and exit
                      ; - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -                                                           
F2F8  D7FA                                    stb     <itsRead                                                ; save number of items we received
F2FA  CE0800                                  ldu     #$0800                                          ; point to beginning of received data
F2FD  DFF8                                    stu     <cpyPtr                                         ; save it
F2FF  0FFB                                    clr     <itsDone                                                ; for each item (clear counter)
F301  0CFB            shwRD03 inc     <itsDone                                                ; add 1 to counter
F303  17FE22                                  lbsr    shwFd01                                         ; show item name and type
F306  30C4                                    leax    ,u                                                              ; point to msb highest byte of item length (not used)
F308  EE01                                    ldu     1,x                                                     ; get 2 next bytes
F30A  A603                                    lda     3,x                                                     ; get lowest byte
F30C  17FDF8                                  lbsr    shwNum                                          ; show item Length (U:A)
F30F  9EF8                                    ldx     <cpyPtr                                         ; get beginning of this record
F311  308810                                  leax    16,x                                                    ; point to next one
F314  9FF8                                    stx     <cpyPtr                                         ; update pointer
F316  D6FA                                    ldb     <itsRead                                                ; get number of items read
F318  C110                                    cmpb    #16                                                     ; was 16?                                                               
F31A  262C                                    bne     shwRD07                                         ; NO, so partial page. Send CR and go for next one
F31C  96FB                                    lda     <itsDone                                                ; YES, it was a full page. Get item number just printed
F31E  810F                                    cmpa    #15                                                     ; was penultimate row?
F320  2608                                    bne     shwRD04                                         ; no, see if it was the last one
F322  8EE4CF                                  ldx     #msgCont-1                                      ; yes point to message CONT
F325  BD90E5                                  jsr     outstr                                          ; show message                  
F328  201E                                    bra     shwRD07                                         ; and send CR and go for next item      
F32A  8110            shwRD04 cmpa    #16                                                     ; is the last one?
F32C  261A                                    bne     shwRD07                                         ; no, send CR and go for next item
F32E  8EE4D6                                  ldx     #msgExit-1                                      ; yes, point to message EXIT
F331  BD90E5                                  jsr     outstr                                          ; show message                  
F334  BD8006          shwRD05 jsr     $8006                                                   ; wait for user keypress
F337  27FB                                    beq     shwRD05                                         ; none, loopback
F339  8143                                    cmpa    #$43                                                    ; is 'C' for Continue?
F33B  2605                                    bne     shwRD06                                         ; no, verify if is 'E'
F33D  BD90A5                                  jsr     sendCR                                          ; send a CR (will scroll 1 line up)
F340  208E                                    bra   shwRD00                                           ; BREAK, ask for a new page to the SDC
F342  8145            shwRD06 cmpa    #$45                                                    ; is 'E' for eXIT?
F344  26EE                                    bne     shwRD05                                         ; no, loopback and wait for a correct key
F346  200D                                    bra     shwRD08                                         ; yes, send CR, show OK and exit
F348  BD90A5          shwRD07 jsr     sendCR                                          ; send CR
F34B  96FB                                    lda     <itsDone                                                ; get number of items shown
F34D  91FA                                    cmpa    <itsRead                                                ; work done?
F34F  2709                                    beq     shwRD09                                         ; yes, send OK and exit
F351  DEF8                                    ldu     <cpyPtr                                         ; no, get saved buffer pointer
F353  20AC                                    bra     shwRD03                                         ; loopback for next item
F355  860D            shwRD08 lda     #$0d                                                    ; get enter value
F357  BDB54A                                  jsr     outchr                                          ; print it
F35A  17FCFA          shwRD09 lbsr    exsdrv                                          ; restore slot if needed
F35D  39                                      rts                                                                     ; return
                      ; - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -                                                           
                                                                                                                      ; show name of current dir
F35E  9D9F            shwCDir jsr     <getNChr                                                ; skip token
F360  8601                                    lda     #1                                                              ; put value 1
F362  97F9                                    sta     <numDir                                         ; as first level requested
F364  17FE57                                  lbsr    other01                                         ; put some parameters
F367  86C0                                    lda     #$c0                                                    ; command to be sent
F369  97FB                                    sta     <cmdNum                                         ; save it
F36B  C643                                    ldb     #'C'                                                    ; param for b
F36D  D7FA                                    stb     <valPar                                         ; save it
F36F  CC0100                                  ldd     #$0100                                          ; get 1 byte with value equal null
F372  FD01D1                                  std     nambuf                                          ; put in nambuf area (no name)
F375  BDEDA5                                  jsr     putCmd2                                         ; send command
F378  2503                                    bcs     eshwCDir                                                ; if error, don't show garbage
F37A  17FDA8                                  lbsr    shwFdat                                         ; show dirname and type
F37D  16FCD7          eshwCDir        lbra    exsdrv                                          ; restore slot if needed
                      ; - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -           
                                                                                                                      ; show contents of root dir
F380  9D9F            shwRoot jsr     <getNChr                                                ; skip token
F382  8601                                    lda     #$01                                                    ; only one char in nambuf
F384  C62F                                    ldb     #'/'                                                    ; char to design root directory
F386  FD01D1          shwRt01 std     nambuf                                          ; put data in input buffer area
F389  17FE32                                  lbsr    other01                                         ; set parameters
F38C  16FF3B                                  lbra    shwEDir                                         ; go to process the requested directory
                      ; - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -                                                           
                                                                                                                      ; show contents of current dir
F38F  8601            shwCAct lda     #$01                                                    ; only one char in nambuf
F391  C62A                                    ldb     #'*'                                                    ; char to design current directory
F393  20F1                                    bra     shwRt01                                         ; complete data and process command
                      ; - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -                                                           
                                                                                                                      ; set current dir to requested name
F395  9D9F            setCDir jsr     <getNChr                                                ; get next entered char
F397  17FE11          setCD01 lbsr    others                                          ; read filename, get params
F39A  CC443A          setCD02 ldd     #"D:"                                                   ; command to send
F39D  16FE05                                  lbra    xsmk01                                          ; go send command
                      ; - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -                                                           
                                                                                                                      ; show actual dir full path
00F9                  numDir  equ     $f9                                                     ; number of received dirs (re-used)
F3A0  9D9F            shwFull jsr     <getNChr                                                ; skip token
F3A2  327E                                    leas    -2,s                                                    ; create a 3 bytes hole for variables
0000                  destPtr equ     0                                                       ; ofsset to get to the pointer to destination buffer (2 bytes)
F3A4  0FF9                                    clr     numDir                                          ; set items received to 0
F3A6  8E0900                                  ldx     #$0900                                          ; get 2nd dos buffer address
F3A9  AFE4                                    stx     destPtr,s                                       ; to receive results
F3AB  17FE10          shwF01  lbsr    other01                                         ; put some parameters
F3AE  86C0                                    lda     #$c0                                                    ; command to be sent
F3B0  97FB                                    sta     <cmdNum                                         ; save it
F3B2  C643                                    ldb     #'C'                                                    ; param for b
F3B4  D7FA                                    stb     <valPar                                         ; save it
F3B6  CC0100                                  ldd     #$0100                                          ; get 1 byte with value equal null
F3B9  FD01D1                                  std     nambuf                                          ; put in nambuf area (no name)
F3BC  0CF9                                    inc     numDir                                          ; increment received items counter
F3BE  BDEDA5                                  jsr     putCmd2                                         ; send command to get actual directory
                                                                                                                      ; processing single answer
F3C1  253D                                    bcs     shwF80                                          ; if error (got to root) skip section  
                                                                                                                      ; save received data to $900 buffer
F3C3  8E0800                                  ldx     #$800                                                   ; point to beginning of received data
F3C6  EEE4                                    ldu     destPtr,s                                       ; point ot destination record
F3C8  C608                                    ldb     #8                                                              ; max name length 
F3CA  A680            shwF02  lda     ,x+                                                     ; get a received char
F3CC  8120                                    cmpa    #$20                                                    ; is it space
F3CE  2705                                    beq     shwF03                                          ; yes, stop copying name
F3D0  A7C0                                    sta     ,u+                                                     ; pass char to record
F3D2  5A                                      decb                                                                    ; decrement counter
F3D3  26F5                                    bne     shwF02                                          ; if not zero, loopback
F3D5  E680            shwF03  ldb     ,x+                                                     ; get 1st termination char
F3D7  C120                                    cmpb    #$20                                                    ; is it a space?
F3D9  2714                                    beq     shwF04                                          ; yes, goto next dir element
F3DB  862E                                    lda     #'.'                                                    ; get dot char
F3DD  EDC1                                    std     ,u++                                                    ; add dot and 1st ext char
F3DF  A680                                    lda     ,x+                                                     ; get 2nd extension char
F3E1  8120                                    cmpa    #$20                                                    ; is it a space?
F3E3  270A                                    beq     shwF04                                          ; yes, goto next dir element
F3E5  A7C0                                    sta     ,u+                                                     ; add 2nd extension char
F3E7  A680                                    lda     ,x+                                                     ; get 3rd extension char
F3E9  8120                                    cmpa    #$20                                                    ; is it a space?
F3EB  2702                                    beq     shwF04                                          ; yes, skip next
F3ED  A7C0                                    sta     ,u+                                                     ; add 3rd extension char
F3EF  CC2F00          shwF04  ldd     #$2f00"                                         ; get a slash and zero terminator
F3F2  EDC4                                    std     ,u                                                              ; add to destination buffer
F3F4  EEE4                                    ldu     destPtr,s                                       ; get beginning just used
F3F6  33C810                                  leau    16,u                                                    ; add 16 to buffer pointer
F3F9  EFE4                                    stu     destPtr,s                                       ; update it for next path directory
F3FB  17FE8A                                  lbsr    chwPD01                                         ; ask SDC to go to parent dir (SCHD"..")
F3FE  20AB                                    bra     shwF01                                          ; go back to ask for new actual directory
                                                                                                                      ; got to the root, NO MORE data to request. 
F400  96F9            shwF80  lda     <numDir                                         ; get number of received items
F402  8101                                    cmpa    #1                                                              ; only one item received?
F404  273E                                    beq     eshwFull                                                ; yes, it was ROOT, so exit
                                                                                                                      ; process received data to show full path
F406  CE01D2                                  ldu     #nambuf+1                                       ; destination for full path name (input buffer)
F409  862F                                    lda     #$2f                                                    ; get a slash to denote root
F40B  A7C0                                    sta     ,u+                                                     ; put into buffer as 1st char
                                                                                                                      ; add all the received dirs in reverse order
F40D  AEE4                                    ldx     destPtr,s                                       ; get last record address (not used ... it was root)
F40F  3010            shwF81  leax    -16,x                                                   ; point to previous used entry
F411  8C0900                                  cmpx    #$0900                                          ; gone back past the beginning?
F414  250E                                    bcs     shwF90                                          ; yes, end of putting items together
F416  3410                                    pshs    x                                                               ; save register
F418  A680            shwF82  lda     ,x+                                                     ; get a char from origin
F41A  2704                                    beq     shwF83                                          ; if null, end of string
F41C  A7C0                                    sta     ,u+                                                     ; else put at destination
F41E  20F8                                    bra     shwF82                                          ; go get next char
F420  3510            shwF83  puls    x                                                               ; restore pointer
F422  20EB                                    bra     shwF81                                          ; back to get next directory
F424  6FC4            shwF90  clr     ,u                                                              ; mark end of string
F426  1F30                                    tfr     u,d                                                     ; send poiter to D
F428  8301D2                                  subd    #nambuf+1                                       ; to calculate length of string
F42B  F701D1                                  stb     nambuf                                          ; save it
F42E  8E01D1                                  ldx     #nambuf                                         ; point to whole path name
F431  BD90E5                                  jsr     outstr                                          ; show the full mounted directory path
F434  860D                                    lda     #$0d                                                    ; code for Enter
F436  BDB54A                                  jsr     outchr                                          ; print it
                                                                                                                      ; issue a SCHD"" with the whole path to restore situation!!
F439  3262                                    leas    2,s                                                     ; clean stack
F43B  17FD80                                  lbsr    other01                                         ; set some parameters
F43E  CC443A                                  ldd     #"D:"                                                   ; command to send       
F441  16FD61                                  lbra    xsmk01                                          ; go back to the actual directory we had before
F444  3262            eshwFull        leas    2,s                                                     ; clean stack
F446  16FC0E                                  lbra    exsdrv                                          ; exit
                      
                      ;--------------------------------------------------------------------------
                      ; process for command WRITE MEM @bank, start, destination, length
                      ; this code is made based on the FLASHER.BIN program by Darren Atkinson
                      ;--------------------------------------------------------------------------
FFD8                  R1CLR           equ     $ffd8                                                   ; to clear fast poke FFD7, for Dragon should be FFD6 !!
FFDE                  RAMROM  equ     $ffde                                                   ; to go MAP0
00EB                  DCDRV           equ     $eb                                                     ; bank number
0086                  TargBank equ   $86                                                      ; bank to be flashed
0050                  FPA0            equ     $50                                                     ; number of bytes to flash
0003                  ErasCtr  equ   $03                                                      ; sctor count in eraser
00EE                  DCBPT           equ     $ee                                                     ; pointer to subroutines
FF4A                  FLSHREG  equ   $ff4a                                                    ; it was $42 for CoCo
FF4B                  BANKREG  equ   $ff4b                                                    ; it was $43 for CoCo
0076                  dSource equ     $76                                                     ; 2 bytes (re-used)
00FA                  dDest           equ     $fa                                                     ; 2 bytes (re-used)
00F9                  sdcSlot equ     $f9                                                     ; 1 byte (re-used)
                      
F449  9D9F            xwrite  jsr     <getNChr                                                ; get next char
F44B  819B                                    cmpa    #$9b                                                    ; is the token for MEM?
F44D  1026F31B                                lbne    snError                                         ; no, show error message
F451  9D9F                                    jsr     <getNChr                                                ; get next byte
F453  8140                                    cmpa    #64                                                     ; is it sign '@'?
F455  1026F313                                lbne    snError                                         ; no, show error message
F459  9D9F                                    jsr     <getNChr                                                ; get number into A
F45B  8130                                    cmpa    #$30                                                    ; is greater or equal 0?
F45D  2504                                    bcs     xwrt01                                          ; no, show error
F45F  8137                                    cmpa    #$37                                                    ; is less or equal 7?
F461  2305                                    bls     xwrt02                                          ; yes, goto next parameters
F463  C611            xwrt01  ldb     #$11                                                    ; SD error
F465  7E8344                                  jmp     syserr                                          ; bas_system_error
F468  8407            xwrt02  anda    #$07                                                    ; use only valid bits
F46A  97EB                                    sta     <DCDRV                                          ; save requested bank number
F46C  9D9F                                    jsr     <getNChr                                                ; advance one char
F46E  260F                                    bne     morParms                                                ; if more chars, read them
                                                                                                                      ; default parameters for writting a ROMPACK
F470  8E3000                                  ldx     #$3000                                          ; get origin of data
F473  9F76                                    stx     <dSource                                                ; save start into variable
F475  8EC000                                  ldx     #$c000                                          ; get destination in bank
F478  9FFA                                    stx     <dDest                                          ; save destination into variable
F47A  8E3F00                                  ldx     #$3f00                                          ; get max. length (16k-256bytes)
F47D  203F                                    bra     savLen                                          ; skip section
F47F  BD89AA          morParms        jsr     ckComma                                         ; CkComma
F482  BD8E83                                  jsr     get16bN                                         ; get a 16 bit number into X
F485  8C7FFF                                  cmpx    #basini-1                                       ; is it on the upper 32k?
F488  226E                                    bhi     badParam                                                ; yes, bad format number, show error
F48A  9F76                                    stx     <dSource                                                ; save start into variable
F48C  9DA5                                    jsr     <redLChr                                                ; advance one char
F48E  BD89AA                                  jsr     ckComma                                         ; CkComma
F491  BD8E83                                  jsr     get16bN                                         ; get a 16 bit number into X
F494  8CC000                                  cmpx    #romini                                         ; is it out of SDC-ROM area?
F497  255F                                    blo     badParam                                                ; yes, bad format number, show error
F499  8CF000                                  cmpx    #$f000                                          ; is it higher than the beginning of 4th sector?
F49C  225A                                    bhi     badParam                                                ; yes, bad format number, show error
F49E  9FFA                                    stx     <dDest                                          ; save destination into variable
F4A0  9DA5                                    jsr     <redLChr                                                ; advance one char
F4A2  BD89AA                                  jsr     ckComma                                         ; CkComma
F4A5  BD8E83                                  jsr     get16bN                                         ; get a 16 bit number into X
F4A8  1F10                                    tfr     x,d                                                     ; pass value to D
F4AA  D3FA                                    addd    <dDest                                          ; add destination value
F4AC  254A                                    bcs     badParam                                                ; yes, sum exceeds flash capacity
F4AE  1083FF00                                cmpd    #$ff00                                          ; uses result the reserved 256 bytes?
F4B2  2244                                    bhi     badParam                                                ; yes, bad format number, show error
F4B4  1F10                                    tfr     x,d                                                     ; get length again
F4B6  D376                                    addd    <dSource                                                ; add source beginning
F4B8  10837FFF                                cmpd    #basini-1                                       ; exceeds RAM area?
F4BC  223A                                    bhi     badParam                                                ; yes, bad format number, show error
F4BE  9F50            savLen  stx     <FPA0                                                   ; save byte count
                      ; - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -                                                           
                                                                                                                      ; search slot where SDC is
F4C0  17F714                                  lbsr    tstSDC                                          ; is SDC already active?
F4C3  271C                                    beq     xwrt03                                          ; yes, save slot number (if no MPI this will be $00)
F4C5  B60114                                  lda     floSDC                                          ; no, get hdw control byte
F4C8  8508                                    bita    #%00001000                                      ; is a SDC in the system?
F4CA  1027F631                                lbeq    NotSDCPresent                           ; no, show message
                                                                                                                      ; yes, calculate occupied slot
F4CE  B60114                                  lda     floSDC                                          ; get hdw control byte
F4D1  84C0                                    anda    #%11000000                                      ; use bits 7-6 (slot where SDC is)
F4D3  44                                      lsra                                                                    ; move them to
F4D4  44                                      lsra                                                                    ; bits 5-4 (lower bits of high nibble)
F4D5  8AC0                                    ora     #%11000000                                      ; set bits 7-6. High nibble OK
F4D7  3402                                    pshs    a                                                               ; save high nibble
F4D9  44                                      lsra                                                                    ; low nibble   
F4DA  44                                      lsra                                                                    ; gets high nibble
F4DB  44                                      lsra                                                                    ; that will end  
F4DC  44                                      lsra                                                                    ; as zero
F4DD  AAE0                                    ora     ,s+                                                     ; add high nibble (now they are the same value)
F4DF  2003                                    bra     xwrt04                                          ; save slot number
F4E1  B6FF7F          xwrt03  lda     $ff7f                                                   ; get active slot number
F4E4  97F9            xwrt04  sta     <sdcSlot                                                ; save slot where SDC is
                                                                                                                      ; copy the flasher routine to $1da. Should not exceed $2d2
F4E6  8EF507          cpyFlsh ldx     #xflash                                         ; point to beginning of program to be copied
F4E9  CE01DA                                  ldu     #$1da                                                   ; point to destination
F4EC  EC81            cpyFl01 ldd     ,x++                                                    ; get a word
F4EE  EDC1                                    std     ,u++                                                    ; put at destination
F4F0  8CF60F                                  cmpx    #erasEnd+2                                      ; got to end of program?
F4F3  25F7                                    blo     cpyFl01                                         ; no, loopback
F4F5  7E01DA                                  jmp     $1da                                                    ; execute copied program in low RAM
                      ; - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -                                                           
F4F8  8EE4DD          badParam        ldx     #wrongPar-1                                     ; point to message
F4FB  BD90E5                                  jsr     outstr                                          ; show it
F4FE  9DA5                                    jsr     <redLChr                                                ; read last char
F500  2704                                    beq     ebadPrm                                         ; if null, exit
F502  9D9F            bP001           jsr     <getNChr                                                ; skip an unused byte
F504  26FC                                    bne     bP001                                                   ; if not null, loopback
F506  39              ebadPrm rts                                                                     ; return
                      ; - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -                                                           
                      ; This code manages the process of flashing a bank
                      ; - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -                                                           
F507  1A50            xflash  orcc  #$50                    ; mask interrupts
F509  96F9                     lda    <sdcSlot                                                ; get the slot where SDC is
                                                                                                                      ; this is to avoid starting the just flashed bank!
F50B  F6FF7F                   ldb    $ff7f                                                   ; get current active slot
F50E  D7F9                                    stb     <sdcSlot                                                ; save it to go back to it at the end
F510  B7FF7F                                  sta     $ff7f                                                   ; switch to the SDC slot
F513  B7FFD8                             sta   R1CLR                   ; normal speed for CoCo 3 (0,89MHz) - $ffd8 (resets $ffd7)
F516  B7FFDE                   sta   RAMROM                  ; activate RAM/ROM map mode = MAP0 - resets $ffdf
F519  D6EB                     ldb   <DCDRV                  ; get target bank number (parameter at $eb)
F51B  8D18                     bsr   FlashKill               ; erase target bank
F51D  308C5D                   leax  burner,pcr                   ; point X at burner subroutine
F520  9FEE                     stx   <DCBPT                  ; store subroutine address: $ee
F522  9E76                                    ldx     <dSource                                                ; point X at source data
F524  DEFA                                    ldu     <dDest                  ; point U at flash destination
F526  8D42                     bsr   callX                   ; write data to the flash
F528  D6F9                     ldb    <sdcSlot                                                ; get original active slot
F52A  F7FF7F                   stb    $ff7f                                                   ; set it again
F52D  0F71                     clr   <$71                    ; force a cold boot
F52F  1A50                     orcc  #$50                    ; mask interrupts
F531  6E9FFFFE                 jmp   [$fffe]                 ; execute the system reset routine
                      ; - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -                                                           
                      ;  Erase flash bank specified in ACCB
                      ; - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -                                                           
F535                  FlashKill 
F535  D786                                    stb   <TargBank               ; store bank number: $86
F537  9650                                    lda     <FPA0                                                   ; get byte count                                                                                                                                                        ; 2B - 20
F539  84F0                                    anda    #$f0                                                    ; use only high nibble                                                                                                                                  ; 20 - 20
F53B  3402                                    pshs    a                                                               ; push value (number of sectors of 4K)                                                                                          ; 20 - 20
F53D  DC50                                    ldd     <FPA0                                                   ; get full byte count                                                                                                                                   ; 2B00 - 2000
F53F  840F                                    anda    #$0f                                                    ; mask high nibble of high byte                                                                                                         ; 0B00 - 0000
F541  10830000                                cmpd    #$0000                                          ; is masked D equal 0?                                                                                                                                  ; NO   - YES
F545  2706                                    beq     flK01                                                   ; yes, don't correct values                                                                                                                     ; corr - jump
F547  8610                                    lda     #$10                                                    ; add one sector (x16)
F549  ABE4                                    adda    ,s                                                              ; to pushed value                                                                                                                                                       ; 30
F54B  A7E4                                    sta     ,s                                                              ; update it in the stack                                                                                                                                ; 30
F54D  3502            flK01           puls    a                                                               ; get number of sectors to kill (x16 'cause in high nibble)                                     ; 30 - 20
F54F  8140                                    cmpa    #$40                                                    ; number of sectors greater than 4?
F551  2302                                    bls     flK02                                                   ; no, go on
F553  8640                                    lda     #$40                                                    ; yes, limit to 4
F555  1F89            flK02           tfr     a,b                                                     ; copy to B                                                                                                                                                                     ; 30 - 20
F557  54                                      lsrb                                                                    ; move
F558  54                                      lsrb                                                                    ; 4 times to
F559  54                                      lsrb                                                                    ; the right to
F55A  54                                      lsrb                                                                    ; divide by 16 to get the actual number of sectors      to kill                                 ; 03 - 02
F55B  96FA                                    lda     <dDest                                          ; get destination ini high byte                                                                                                         ; D3 - C0       ($D300 - $C000)
F55D  84F0                                    anda    #$f0                                                    ; use high nibble                                                                                                                                                       ; D0 - C0
                                                                                                                      ; now D should contain: 
                                                                                                                      ;     A = High byte of address of 1st sector to kill ($C0-$D0-$E0-$F0)
                                                                                                                      ;     B = number of sectors to be killed
                                                                                                                      ; save parameter for later use
F55F  9752                     sta   <FPA0+2                 ; store MSB of default sector address: $52
F561  D703                     stb   <ErasCtr                ; store sector count: $03
F563  0F53                     clr   <FPA0+3                 ; clear LSB of sector address: $53
F565  308C65                   leax  eraser,pcr              ; point X at erasure routine
F568  9FEE                     stx   <DCBPT                                           ; save it: $ee
F56A  9686            callX    lda   <TargBank                        ; get current flash bank selection: $86
F56C  8407                     anda  #$07                    ; mask out undefined bits
F56E  3403                     pshs  a,cc                    ; save current bank and IRQ mask
F570  AD9F00EE                 jsr   [DCBPT]                 ; call utility routine: $ee -> eraser (routine)
F574  B6FF22                   lda   $ff22                   ; read PIA data reg to clear any CART interrupt
F577  3503                     puls  cc,a                    ; restore old IRQ mask and pop saved bank number
F579  7FFF4A                   clr   FLSHREG                 ; clear the Flash byte register: $ff4a
F57C  39                       rts                           ; return
                      ; - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -                                                           
                      ; Write Data to Flash
                      ; Entry:
                      ;   X = source data address
                      ;   U = destination flash address
                      ;   High word of FPA0 = byte count ($50)
                      ; - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -                                                           
F57D  DC50            burner   ldd   <FPA0                   ; D = byte count: $50
F57F  31CB                     leay  d,u                     ; Y = destination end address
F581  109F52                   sty   <FPA0+2                 ; save end address in FPA0: $52
F584  3184                     leay  ,x                      ; point Y at the source data
F586  30C4                     leax  ,u                      ; point X at the destination
F588  338C1F                   leau  burnPrfx,PCR            ; point U at the prefix data for Byte Program
F58B  9686                     lda   <TargBank               ; get target bank number: $86
F58D  B7FF4B                   sta   BANKREG                 ; activate target bank: ff4b
F590  8A80                     ora   #$80                    ; setup bank number combined with PGM Enable
F592  E6A0            burn010  ldb   ,y+                     ; get next data byte from source
F594  E184                     cmpb  ,x                      ; does flash already contain this value?
F596  2604                     bne   burn020                 ; branch if needs to change
F598  3001                     leax  1,x                     ; increment flash address
F59A  2008                     bra   burn030                 ; skip ahead to end of loop
F59C  3456            burn020  pshs  u,x,b,a                 ; save registers
F59E  8D17                     bsr   flshPrefx               ; send the unlock sequence
F5A0  3556                     puls  a,b,x,u                 ; restore target bank and destination address
F5A2  8D20                     bsr   flshWrCyc               ; write the byte to the Flash destination
F5A4  9C52            burn030  cmpx  <FPA0+2                 ; have we reached the final address: $52
F5A6  25EA                     blo   burn010                 ; loop if more
F5A8  2016                     bra   flshDone                ; restore original bank selection and return
                                                                                                                      ; - - -  Prefix Data for Byte-Program  - - -
F5AA  81AA            burnPrfx fdb   $81aa                   ; Bank = 1, Data = $AA...
F5AC  D555                     fdb   $d555                   ; ...Address = $D555 ($5555)
F5AE  8055                     fdb   $8055                   ; Bank 0, Data = $55...
F5B0  EAAA                     fdb   $eaaa                   ; ...Address = $EAAA ($2AAA)
F5B2  81A0                     fdb   $81a0                   ; Bank 1, Data = $A0...
F5B4  D555                     fdb   $d555                   ; ...Address = $D555 ($5555)
F5B6  00                       fcb   $0                                 ; end of sequence
                      ; - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -                                                           
                      ; Send a Programming Prefix Sequence to the Flash.
                      ; Enter with U pointing to Prefix data.
                      ; - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -                                                           
F5B7                  flshPrefx
F5B7  3716                                    pulu  a,b,x                   ; get next bank, data byte and address
F5B9  4D                       tsta                          ; if reached end-of-sequence..
F5BA  2710                     beq   flshRts                 ; ..then return
F5BC  8D06                     bsr   flshWrCyc               ; write byte to flash
F5BE  20F7                     bra   flshPrefx               ; loop for next byte of sequence
                      ; - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -                                                           
                                                                                                                      ; Restores old flash bank and returns upon completion of Flash operation
F5C0  A663            flshDone lda   3,s                     ; get orignal bank number from stack
F5C2  847F                     anda  #$7f                    ; make sure the PGM Enable bit is cleared
                                                                                                                      ;  *** Fall Thru ***
                      ; - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -                                                           
                      ; Perform a Write Cycle to the Flash.
                      ; Enter with Address in X, Bank+PGM Enable in ACCA and the data byte in ACCB.
                      ; - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -                                                           
F5C4                  flshWrCyc
F5C4  B7FF4B                                  sta   BANKREG                 ; select Flash bank / PGM Enable: $ff4b
F5C7  F7FF4A                   stb   FLSHREG                 ; store data byte in Flash register: $ff4a
F5CA  E580                     bitb  ,x+                     ; perform Flash Write by reading target address
F5CC  39              flshRts  rts                           ; return
                      ; - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -                                                           
                      ; Erase Flash Sector
                      ; - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -                                                           
F5CD  338C27          eraser   leau  erasPrfx,PCR            ; point U to the prefix data for Sector Erase
F5D0  8DE5                     bsr   flshPrefx               ; write the prefix bytes
F5D2  CC8030                   ldd   #$8030                  ; ACCA = PGM Enable;  ACCB = final prefix byte
F5D5  9A86                     ora   <TargBank               ; combine target bank number with PGM Enable: $86
F5D7  9E52                     ldx   <FPA0+2                 ; X = sector address: $52 ($C0xx)
F5D9  8DE9                     bsr   flshWrCyc               ; write final prefix byte at sector address
F5DB  8407                     anda  #$07                    ; turn off the..
F5DD  B7FF4B                   sta   BANKREG                 ; ..PGM Enable bit: $ff4b
F5E0  4F                       clra                          ; clear ACCD for use..
F5E1  5F                       clrb                          ; ..as timeout counter
F5E2  6D1F            eras025  tst   -1,x                    ; poll the sector address: $C0xx-1
F5E4  2B05                     bmi   eras030                 ; branch if bit 7 is set (erasure complete)
F5E6  830001                   subd  #1                      ; decrement timeout counter
F5E9  26F7                     bne   eras025                 ; loop if not yet timed out
F5EB  30890FFF        eras030  leax  $0fff,x                 ; increment X to point at the next sector
F5EF  9F52                     stx   <FPA0+2                 ; save sector address in FPA0: $52 ($D0xx)
F5F1  0A03                     dec   <ErasCtr                ; decrement erase sector counter: $03
F5F3  26D8                     bne   eraser                  ; loop if erasing another sector
F5F5  20C9                     bra   flshDone                ; restore original bank selection and return
                                                                                                                      ; - - -  Programming Prefix for Sector Erase  - - -
F5F7  81AA            erasPrfx fdb   $81aa                   ; Bank = 1, Data = $AA...
F5F9  D555                     fdb   $d555                   ; ...Address = $D555 ($5555)
F5FB  8055                     fdb   $8055                   ; Bank 0, Data = $55...
F5FD  EAAA                     fdb   $eaaa                   ; ...Address = $EAAA ($2AAA)
F5FF  8180                     fdb   $8180                   ; Bank 1, Data = $80...
F601  D555                     fdb   $d555                   ; ...Address = $D555 ($5555)
F603  81AA                     fdb   $81aa                   ; Bank = 1, Data = $AA...
F605  D555                     fdb   $d555                   ; ...Address = $D555 ($5555)
F607  8055                     fdb   $8055                   ; Bank 0, Data = $55...
F609  EAAA                     fdb   $eaaa                   ; ...Address = $EAAA ($2AAA)
F60B  0000                     fdb   $0000                   ; end of sequence
F60D  0000            erasEnd  fdb   $0000                                                    ; ...a record is 4 bytes long ...
                      ;--------------------------------------------------------------------------
                      ; process for command SCOPY MEM @number   ,&HC000,&H4000,&H3F00
                      ; or alternate        SCOPY MEM SLOTnumber,&HC000,&H4000,&H3F00
                      ;--------------------------------------------------------------------------
00F8                  oriNum  equ     $f8                                                     ; number of requested slot-bank (re-used)
00F9                  cpyWhat equ     $f9                                                     ; command that identifies BANK - SLOT (re-used)
0052                  oldSlot equ     $52                                                     ; FPA2 (re-used)
0053                  oldBank equ     $53                                                     ; FPA3 (re-used)
                      
F60F  9D9F            xscopy  jsr     <getNChr                                                ; get next char
F611  819B                                    cmpa    #$9b                                                    ; is the token for MEM?
F613  1026F155                                lbne    snError                                         ; no, show error message
F617  9D9F                                    jsr     <getNChr                                                ; get next byte
F619  97F9                                    sta     <cpyWhat                                                ; save received command
F61B  8140                                    cmpa    #64                                                     ; is it sign '@' for BANK?
F61D  2717                                    beq     xsbnk01                                         ; yes, go get bank number
F61F  81EF                                    cmpa    #$ef                                                    ; is the token for SLOT?
F621  1026F147                                lbne    snError                                         ; no, show error
                                                                                                                      ; processing slot command
F625  9D9F                                    jsr     <getNChr                                                ; get number into A
F627  8131                                    cmpa    #$31                                                    ; is greater or equal 1?
F629  2507                                    bcs     xscerr1                                         ; no, show error
F62B  8134                                    cmpa    #$34                                                    ; is less or equal 4?
F62D  2203                                    bhi   xscerr1                                           ; no, show error 
F62F  4A                                      deca                                                                    ; internally MUST work with 0-1-2-3
F630  2013                                    bra     xsParam                                         ; yes, goto next parameters
F632  C60D            xscerr1 ldb     #$0d                                                    ; MU error
F634  200C                                    bra     xscerr0                                         ; goto bas_system_error
                                                                                                                      ; processing bank command
F636  9D9F            xsbnk01 jsr     <getNChr                                                ; get number into A
F638  8130                                    cmpa    #$30                                                    ; is greater or equal 0?
F63A  2504                                    bcs     xscerr2                                         ; no, show error
F63C  8137                                    cmpa    #$37                                                    ; is less or equal 7?
F63E  2305                                    bls     xsParam                                         ; yes, goto next parameters
F640  C611            xscerr2         ldb     #$11                                                    ; SD error
F642  7E8344          xscerr0 jmp     syserr                                          ; bas_system_error
                                                                                                                      ; look for more parameters
F645  8030            xsParam suba    #$30                                                    ; convert ASCII to integer (0-3 or 0-7)
F647  97F8                                    sta     <oriNum                                                 ; save requested bank/slot number
F649  9D9F                                    jsr     <getNChr                                                ; advance one char
F64B  260F                                    bne     otherPar                                                ; if more chars, read them
                                                                                                                      ; default parameters for writting a ROMPACK
F64D  8EC000                                  ldx     #$c000                                          ; get origin of data
F650  9F76                                    stx     <dSource                                                ; save start into variable
F652  8E3000                                  ldx     #$3000                                          ; get destination in bank
F655  9FFA                                    stx     <dDest                                          ; save destination into variable
F657  8E3F00                                  ldx     #$3f00                                          ; get max. length (16k-256bytes)
F65A  2028                                    bra     storeLen                                                ; skip section                  
F65C  BD89AA          otherPar        jsr     ckComma                                         ; CkComma
F65F  BD8E83                                  jsr     get16bN                                         ; get a 16 bit number into X
F662  8CC000                                  cmpx    #romini                                         ; is it on the upper 32k?
F665  1025FE8F                                lblo    badParam                                                ; no, bad format number, show error
F669  9F76                                    stx     <dSource                                                ; save start into variable
F66B  9DA5                                    jsr     <redLChr                                                ; advance one char
F66D  BD89AA                                  jsr     ckComma                                         ; CkComma
F670  BD8E83                                  jsr     get16bN                                         ; get a 16 bit number into X
F673  8C4000                                  cmpx    #$4000                                          ; is it leaving space for a 16k ROM?
F676  1022FE7E                                lbhi    badParam                                                ; yes, bad format number, show error
F67A  9FFA                                    stx     <dDest                                          ; save destination into variable
F67C  9DA5                                    jsr     <redLChr                                                ; advance one char
F67E  BD89AA                                  jsr     ckComma                                         ; CkComma
F681  BD8E83                                  jsr     get16bN                                         ; get a 16 bit number into X
F684  1F10            storeLen        tfr     x,d                                                     ; pass value to D
F686  D3FA                                    addd    <dDest                                          ; add destination value to calculate end value
F688  10837FFF                                cmpd    #basini-1                                       ; does result exceed the RAM area?
F68C  1022FE68                                lbhi    badParam                                                ; yes, bad format number, show error
F690  DD50                                    std     <FPA0                                                   ; save last byte to be written
F692  1F10                                    tfr     x,d                                                     ; get byte count again
F694  D376                                    addd    <dSource                                                ; add first one to copy
F696  1025FE5E                                lbcs    badParam                                                ; if carry it goes past $FFFF, show message
                                                                                                                      ; stack control
F69A  DC50                                    ldd     <FPA0                                                   ; get last byte to write
F69C  C30032                                  addd    #50                                                     ; 50 security bytes (stack depth)
F69F  109321                                  cmpd    <$21                                                    ; is last calculated address lower than stack pointer?
F6A2  250D                                    blo     stkOK                                                   ; yes, so it's sure to continue
F6A4  DCFA                                    ldd     <dDest                                          ; get first destination byte in RAM
F6A6  109323                                  cmpd    <$23                                                    ; is higher than strings beginning area?
F6A9  2206                                    bhi     stkOK                                                   ; yes, so it's sure to continue
F6AB  8EE4F6                                  ldx     #stackPrb-1                                     ; point to message
F6AE  7E90E5                                  jmp     outstr                                          ; show it
                                                                                                                      ; HDW verification
F6B1  96F9            stkOK           lda     cpyWhat                                         ; get kind of origin to copy
F6B3  81EF                                    cmpa    #$ef                                                    ; is SLOT requested?
F6B5  2609                                    bne     seeBank                                         ; no, go to bank control
F6B7  17F4FF                                  lbsr    tstMPI                                          ; yes, is there an active MPI?
F6BA  270D                                    beq     xsProces                                                ; yes go on     
F6BC  17F43C                                  lbsr    NotMPIPresent                           ; no, show message
F6BF  39                                      rts                                                                     ; return
F6C0  B60114          seeBank lda     floSDC                                          ; get hdw control byte
F6C3  8508                                    bita    #%00001000                                      ; is there an SDC?
F6C5  1027F436                                lbeq    NotSDCPresent                           ; no, show error
                                                                                                                      ; section to copy the code to be executed into $1da
F6C9  8EF6DC          xsProces        ldx     #xsMovCod                                       ; point to origin of code
F6CC  CE01DA                                  ldu     #$1da                                                   ; point to destination
F6CF  EC81            xsMov01 ldd     ,x++                                                    ; get a word
F6D1  EDC1                                    std     ,u++                                                    ; put at destination
F6D3  8CF774                                  cmpx    #xsLoEnd+1                                      ; got to the end?
F6D6  25F7                                    blo     xsMov01                                         ; no, loopback
F6D8  BD01DA                                  jsr     $1da                                                    ; call code at $1da
F6DB  39                                      rts                                                                     ; return
                                                                                                                      ; code to be executed
F6DC  1A50            xsMovCod        orcc    #dimask                                         ; disable interrupts
F6DE  4F                                      clra                                                                    ; $00 is code for NO old value
F6DF  9752                                    sta     oldSlot                                         ; for old slot
F6E1  4A                                      deca                                                                    ; $ff is code for NO old value
F6E2  9753                                    sta     oldBank                                         ; for old bank
F6E4  96F9                                    lda     <cpyWhat                                                ; get kind of element to copy
F6E6  8140                                    cmpa    #64                                                     ; was command a copy BANK?
F6E8  271C                                    beq     xsCopBnk                                                ; yes, go there
                                                                                                                      ; change to the requested slot if not already there and save old slot
F6EA  B6FF7F                                  lda     $ff7f                                                   ; get active slot number
F6ED  9752                                    sta     <oldSlot                                                ; save it
F6EF  8403                                    anda    #3                                                              ; to have just the number
F6F1  91F8                                    cmpa    <oriNum                                         ; is the requested one?
F6F3  2740                                    beq     startCpy                                                ; yes, go start the copy
F6F5  D6F8                                    ldb     <oriNum                                         ; get requested slot number
F6F7  3404                                    pshs    b                                                               ; save it
F6F9  58                                      lslb                                                                    ; move to
F6FA  58                                      lslb                                                                    ; the high
F6FB  58                                      lslb                                                                    ; nibble
F6FC  58                                      lslb                                                                    ; so 
F6FD  EAE0                                    orb     ,s+                                                     ; now both are equal
F6FF  CACC                                    orb     #$cc                                                    ; convert to std slot number
F701  F7FF7F                                  stb     $ff7f                                                   ; change of slot
F704  202F                                    bra     startCpy                                                ; done, go start the copy
F706  B60114          xsCopBnk        lda     floSDC                                          ; get hdw control byte
F709  84C0                                    anda    #%11000000                                      ; use bits 7-6 (slot where SDC is)
F70B  44                                      lsra                                                                    ; move them to
F70C  44                                      lsra                                                                    ; bits 5-4 (lower bits of high nibble)
F70D  8AC0                                    ora     #%11000000                                      ; set bits 7-6. High nibble OK
F70F  3402                                    pshs    a                                                               ; save high nibble
F711  44                                      lsra                                                                    ; low nibble   
F712  44                                      lsra                                                                    ; gets high nibble
F713  44                                      lsra                                                                    ; that will end  
F714  44                                      lsra                                                                    ; as zero
F715  AAE0                                    ora     ,s+                                                     ; add high nibble (now they are the same value)
F717  B1FF7F                                  cmpa    $ff7f                                                   ; is this slot active right now?
F71A  2708                                    beq     cmpBank                                         ; yes, go see if bank differs
F71C  F6FF7F                                  ldb     $ff7f                                                   ; get actual slot
F71F  B7FF7F                                  sta     $ff7f                                                   ; change of slot
F722  D752                                    stb     <oldSlot                                                ; save old slot
F724  96F8            cmpBank lda     <oriNum                                         ; compare requested bank with actually selected one 
F726  B1FF4B                                  cmpa    $ff4b                                                   ; is active now?
F729  2708                                    beq   savBank                                           ; yes, save bank
F72B  F6FF4B                                  ldb     $ff4b                                                   ; no, get active one
F72E  B7FF4B                                  sta     $ff4b                                                   ; change to requested
F731  1F98                                    tfr     b,a                                                     ; pass B to A
F733  9753            savBank sta     <oldBank                                                ; save old bank number
F735  5F              startCpy        clrb                                                                    ; 0 = not in MAP1
F736  B6C000                                  lda     romini                                          ; get value in $C000
F739  43                                      coma                                                                    ; reverse bits
F73A  B7C000                                  sta     romini                                          ; update ram cell
F73D  B1C000                                  cmpa    romini                                          ; retains the new value?
F740  2608                                    bne     noMap1                                          ; no, we are in MAP0
F742  43                                      coma                                                                    ; reverse again
F743  B7C000                                  sta     romini                                          ; restore value
F746  5C                                      incb                                                                    ; 1 = we are in MAP1
F747  B7FFDE                                  sta     $ffde                                                   ; switch to MAP0 to copy the requested ROM
F74A  D7F8            noMap1  stb     <oriNum                                         ; re-use variable to save that value
F74C  9E76                                    ldx     <dSource                                                ; point to source ROM
F74E  DEFA                                    ldu     <dDest                                          ; point to destination RAM
F750  EC81            xsLoop01        ldd     ,x++                                                    ; get a word
F752  EDC1                                    std     ,u++                                                    ; put at destination
F754  119350                                  cmpu    <FPA0                                                   ; got to the end?
F757  23F7                                    bls     xsLoop01                                                ; no, loopback
                                                                                                                      ; restore old bank-slot numbers and Map Type
F759  9652                                    lda     <oldSlot                                                ; get old slot
F75B  2703                                    beq     getBank                                         ; if $00, do nothing
F75D  B7FF7F                                  sta     $ff7f                                                   ; change to old slot number
F760  9653            getBank lda     <oldBank                                                ; get old bank number
F762  2B03                                    bmi     exscopy                                         ; if $ff, do nothing
F764  B7FF4B                                  sta     $ff4b                                                   ; switch to old bank number
F767  0DF8            exscopy tst     <oriNum                                         ; were we initially in MAP1?
F769  2703                                    beq     noMapChg                                                ; no, skip next
F76B  B7FFDF                                  sta     $ffdf                                                   ; back to MAP1
F76E  1CAF            noMapChg        andcc   #eimask                                         ; enable interrupts
F770  7FFF48                                  clr     $ff48                                                   ; to unsure no led stays lit on FDC
F773  39              xsLoEnd rts                                                                     ; return
                      
                      ;--------------------------------------------------------------------------
                      ; routine to verify the status returned by the SDC to a command
                      ;         Bit 7 ($80) set on any failure.
                      ;         Bit 5 ($20) set if target file is already in use.
                      ;         Bit 4 ($10) set if target file not found.
                      ;         Bit 3 ($08) set on miscellaneous hardware errors.
                      ;         Bit 2 ($04) set if path name is invalid.                      
                      ;--------------------------------------------------------------------------
F774                  vfyStatus
F774  242D                                    bcc     evfy                                                    ; if no error return
F776  5D                       tstb                                                                   ; update flags upon result status 
F777  2724                     beq    timeOutError                            ; if zero, timeout
F779  C520                     bitb   #$20                                                    ; detected file in use?
F77B  2610                     bne    targetInUseError                        ; yes, message
F77D  C510                     bitb   #$10                                                    ; was file not found?
F77F  2610                     bne    targetNotFoundError             ; no, message
F781  C508                               bitb         #$08                                                    ; was a hdw error?
F783  2610                     bne    miscHardwareError                       ; yes, message
F785  C504                     bitb   #$04                                                    ; was invalid pathname?
F787  2610                     bne    pathNameInvalidError            ; yes, message
F789                  unknownError                                                                    ; this is for unknown errors. Should not get here!
F789  8EE55C                                  ldx     #unknown-1                                      ; point to message
F78C  10                       fcb    #$10                                                    ; skip next ldx 
F78D                  targetInUseError                                                                ; 
F78D  8EE510                   ldx    #targetInUse-1                          ; point to message
F790  10                       fcb    #$10                                                    ; skip next ldx 
F791                  targetNotFoundError                                                     ; 
F791  8EE56B                   ldx    #targetNotFound-1                       ; point to message 
F794  10                       fcb    #$10                                                    ; skip next ldx 
F795                  miscHardwareError                                                               ; 
F795  8EE546                                  ldx     #miscHardware-1                 ; point to message 
F798  10                       fcb    #$10                                                    ; skip next ldx 
F799                  pathNameInvalidError                                                    ; 
F799  8EE534                                  ldx     #pathNameInvalid-1              ; point to message 
F79C  10                       fcb    #$10                                                    ; skip next ldx 
F79D                  timeOutError                                                                    ; 
F79D  8EE507                                  ldx     #timeout-1                                      ; point to message 
F7A0  7E90E5          Exit            jmp   outstr                                            ; show message
F7A3  39              evfy            rts                                                                     ; return
                      
                      ;--------------------------------------------------------------------------
                      ; CoCo SDC Low-level interface routine
                      ;--------------------------------------------------------------------------
                      ; Hdw Addressing  - Dragon Scheme    -    CoCo Scheme
FF48                  CTRLATCH equ   $ff48                                            ; $ff40 - controller latch (write)
FF40                  CMDREG   equ   $ff40                                            ; $ff48 - command register (write)
FF40                  STATREG  equ   $ff40                                            ; $ff48 - status register (read)
FF41                  PREG1    equ   $ff41                                            ; $ff49 - param register 1
FF42                  PREG2    equ   $ff42                                            ; $ff4a - param register 2
FF43                  PREG3    equ   $ff43                                                    ; $ff4b - param register 3
FF42                  DATREGA  equ    PREG2                              ; first data register
FF43                  DATREGB  equ   PREG3                               ; second data register
                                                                                                                      ; Status Register Masks
0001                  BUSY     equ   %00000001                           ; set while a command is executing
0002                  READY    equ   %00000010                           ; set when ready for a data transfer
0080                  FAILsdc  equ   %10000000                           ; set on command failure
                                                                                                                      ; Mode and Command Values
000B                  CMDMODE  equ   $0b                                                      ; was $43 for CoCo - control latch value to enable command mode
0080                  CMDREAD  equ   $80                                 ; read logical sector
00A0                  CMDWRITE equ   $a0                                 ; write logical sector
00C0                  CMDEX    equ   $c0                                 ; extended command
00E0                  CMDEXD   equ   $e0                                 ; extended command with data block
                      ;--------------------------------------------------------------------------
                      ; CommSDC by Darren Atkinson
                      ;
                      ;    This is the core routine used for all
                      ;    transactions with the SDC controller.
                      ;
                      ; Entry:
                      ;    A = Command code
                      ;    B = LSN hi byte   or  First parameter byte
                      ;    X = LSN lo word   or  2nd and third parameter bytes
                      ;    U = Address of 256 byte I/O buffer ($ffff = none)
                      ;
                      ; Exit:
                      ;    Carry set on error.
                      ;    B = controller status code.
                      ;    A, X, Y and U are preserved.
                      ;--------------------------------------------------------------------------
F7A4  3473            CommSDC  pshs  u,y,x,a,cc                               ; preserve resgisters
F7A6  64E4                     lsr   ,s                                       ; shift carry flag out of saved cc
                                                                                                                      ; put controller in command mode
F7A8  108EFF42                 ldy   #DATREGA                                 ; setup y for hardware addressing (ff42 - ff4a for coco)
F7AC  860B                     lda   #CMDMODE                                 ; the magic number
F7AE  A726                     sta   +6,y                                     ; was -10,y for coco - send to control latch (ff48 - ff40 for coco)
                                                                                                                      ; put input parameters into the hardware registers.
                                                                                                                      ; it does no harm to put random data in the
                                                                                                                      ; registers for commands which dont use them.
F7B0  E73F                     stb   -1,y                                     ; high byte to param reg 1 (ff41 - ff49 for coco)
F7B2  AFA4                     stx   ,y                                       ; low word to param regs 2 and 3 (ff42-ff43 - ff4a-ff4b for coco)
                                                                                                                      ; wait for not busy.
F7B4  8D61                     bsr   waitForIt                           ; run polling loop
F7B6  2555                     bcs   cmdExit                             ; exit if error or timeout
                                                                                                                      ; send command to controller
F7B8  A661                     lda   1,s                                 ; get preserved command code from stack
F7BA  A73E                     sta   -2,y                                ; send to command register (ff40 - ff48 for coco)
                                                                                                                      ; determine if a data block needs to be sent.
                                                                                                                      ; any command which requires a data block will
                                                                                                                      ; have bit 5 set in the command code.
F7BC  8520                     bita  #$20                                     ; test the "send block" command bit
F7BE  272B                     beq   rxBlock                                  ; branch if no block to send
                                                                                                                              ; wait for ready to send
F7C0  8D55                     bsr   waitForIt                           ; run polling loop
F7C2  2549                     bcs   cmdExit                             ; exit if error or timeout
F7C4  30C4                     leax  ,u                                  ; move data address to x
                                                                                                                              ; send 256 bytes of data
F7C6  CC2008                   ldd   #32*256+8                                ; 32 chunks of 8 bytes
F7C9  EE84            txChunk  ldu   ,x                                       ; send one chunk...
F7CB  EFA4                     stu   ,y                                                       ; ff42-ff43 - ff4a-ff4b for coco
F7CD  EE02                     ldu   2,x
F7CF  EFA4                     stu   ,y
F7D1  EE04                     ldu   4,x
F7D3  EFA4                     stu   ,y
F7D5  EE06                     ldu   6,x
F7D7  EFA4                     stu   ,y
F7D9  3A                       abx                                            ; point x at next chunk
F7DA  4A                       deca                                           ; decrement chunk counter
F7DB  26EC                     bne   txChunk                                  ; loop until all 256 bytes sent
                                                                                                                      ; wait for command completion
F7DD  8605                     lda   #5                                       ; timeout retries
F7DF                  waitCmplt
F7DF  8D36                                    bsr   waitForIt                                 ; run polling loop
F7E1  C501                     bitb  #BUSY                                    ; test busy bit
F7E3  2728                     beq   cmdExit                                  ; exit if completed
F7E5  4A                       deca                                           ; decrement retry counter
F7E6  26F7                     bne   waitCmplt                                ; repeat until 0
F7E8  43                       coma                                           ; set carry for timeout error
F7E9  2022                     bra   cmdExit                                  ; exit
                                                                                                                      ; for commands which return a 256 byte response block the
                                                                                                                      ; controller will set the ready bit in the status register
                                                                                                                      ; when it has the data ready for transfer.   for commands
                                                                                                                      ; which do not return a response block the busy bit will
                                                                                                                      ; be cleared to indicate that the command has completed.
F7EB  8D26            rxBlock  bsr   longWait                                 ; run long status polling loop
F7ED  231E                     bls   cmdExit                                  ; exit if error, timeout or completed
F7EF  3041                     leax  1,u                                      ; test the provided buffer address
F7F1  271A                     beq   cmdExit                                  ; exit if "no buffer" ($ffff)
F7F3  30C4                     leax  ,u                                       ; move data address to x
                                                                                                                      ; read 256 bytes of data
F7F5  CC2008                   ldd   #32*256+8                                ; 32 chunks of 8 bytes
F7F8  EEA4            rxChunk  ldu   ,y                                       ; read one chunk... (ff42-ff43 - ff4a-ff4b for coco)
F7FA  EF84                     stu   ,x
F7FC  EEA4                     ldu   ,y
F7FE  EF02                     stu   2,x
F800  EEA4                     ldu   ,y
F802  EF04                     stu   4,x
F804  EEA4                     ldu   ,y
F806  EF06                     stu   6,x
F808  3A                       abx                                            ; update x for next chunk
F809  4A                       deca                                           ; decrement chunk counter
F80A  26EC                     bne   rxChunk                                  ; loop until all 256 bytes transferred
F80C  5F                       clrb                                           ; status code for success, clear carry
                                                                                                                      ; exit
F80D  69E4            cmdExit  rol   ,s                                       ; rotate carry into saved cc on stack
F80F  6F26                     clr   +6,y                                     ; was -10,y for coco - end command mode (ff48 - ff40 for coco)
F811  35F3                     puls  cc,a,x,y,u,pc                            ; restore irq masks, update carry and return
                      ;--------------------------------------------------------------------------
                      ; Wait for controller status to indicate either "Not Busy" or "Ready".
                      ; Will time out if neither condition satisfied within a suitable period.
                      ;
                      ; Exit:
                      ;    CC.C set on error or time out.
                      ;    CC.Z set on "Not Busy" status (if carry cleared).
                      ;    B = status
                      ;    X is clobbered.  A, Y and U are preserved.
                      ;--------------------------------------------------------------------------
F813  8D02            longWait bsr   waitForIt                                ; enter here for doubled timeout
F815  241B                     bcc   waitRet                                  ; return if cleared in 1st pass
F817                  waitForIt
F817  8E0000                                  ldx   #0                                        ; setup timeout counter
F81A  53              waitLp   comb                                           ; set carry for assumed FAIL
F81B  E63E                     ldb   -2,y                                     ; read status (FF40 - FF48 for CoCo)
F81D  2B13                     bmi   waitRet                                  ; return if FAILED
F81F  54                       lsrb                                           ; BUSY --> Carry
F820  240E                     bcc   waitDone                                 ; branch if not busy
F822  C501                     bitb  #READY/2                                 ; test READY (shifted)
F824  260B                     bne   waitRdy                                  ; branch if ready for transfer
F826  8D0A                     bsr   waitRet                                  ; consume some time
F828  C681                     ldb   #$81                                     ; status = timeout
F82A  3082                     leax  ,-x                                      ; decrement timeout counter
F82C  2704                     beq   waitRet                                  ; return if timed out
F82E  20EA                     bra   waitLp                                   ; try again
F830  5F              waitDone clrb                                           ; Not Busy: status = 0, set Z
F831  59              waitRdy  rolb                                           ; On Ready: clear C and Z
F832  39              waitRet  rts                                            ; return
                      
                      ; *******************************************************************************
                      ; EXPERimenTs area
                      ; to help Boot any 'bootable' disk in drive 1 if SPACE IS pressed
                      ; or start AUTOLOAD.DWL is SHIFT is NOT pressed
                      ; *******************************************************************************
F833  C601            expert  ldb     #1                                                              ; set drive number
F835  D7EB                                    stb     <$eb                                                    ; in system variable
F837  F7060A                                  stb     >$60a                                                   ; and in DOS area
F83A  B6FF02                                  lda   $ff02                                                     ; save PIA value
F83D  C67F                     ldb   #$7f                                                     ; value to test for Space & Shift
F83F  F7FF02                   stb   $ff02                                                    ; put it in the PIA
F842  F6FF00                   ldb   $ff00                                                    ; read PIA output
F845  B7FF02                   sta   $ff02                                                    ; restore old value in PIA
F848  53                       comb                                                                   ; invert read bits
F849  C420                     andb  #$20                                                     ; is bit5 set?
F84B  2605                                    bne   doBoot                                            ; yes, SPACE is pressed, BOOT
F84D  BDC16A          noBoot  jsr   $c16a                                                     ; attempt to read a disk sector
F850  200B                                    bra     doDWAuto                                                ; simulate it has failed
F852  10CE7F36        doBoot  lds     #$7f36                                          ; set new stackpointer
F856  8D20                     bsr    rebReset                                                ; rebuild Reset
F858  1CAF                                    andcc   #$af                                                    ; enable interrupts
F85A  7EC13D                                  jmp     $c13d                                                   ; call BOOT drive 1
F85D  7FFF48          doDWAuto clr    $ff48                                                   ; ensure motors off
F860  1CAF                                    andcc #$af                                                      ; enable interrupts 
F862  8EB4B2                   ldx   #$b4b2                                           ; get Basic Intro message 
F865  BD90E5                   jsr   outstr                                           ; show it
F868  8EC34C                   ldx   #$c34c                                           ; get DOS intro message
F86B  BD90E5                   jsr   outstr                                           ; show it
F86E  8D08                                    bsr     rebReset                                                ; rebuild Reset
F870  170133                                  lbsr    chkshift                                                ; check for shift to call AUTOLOAD.DWL
F873                  retFDwl 
F873  1CAF                                    andcc   #$af                                                    ; enable interrupts
F875  7E8371                                  jmp     okprmpt                                         ; goto Basic Iterpreter
                      ; *******************************************************************************       
F878  8EC6B4          rebReset        ldx   #$c6b4                                            ; get new reset routine
F87B  9F72                                    stx     <$72                                                    ; put in variables area
F87D  8655                                    lda     #$55                                                    ; flag value for warmreset
F87F  9771                                    sta     <$71                                                    ; update variable
F881  39                                      rts                                                                     ; return
                      ; *******************************************************************************
                      ; DWLOAD implementation for patched Dragon 32 ROM (upper 8KB) 2014 Tormod Volden
                      ; *******************************************************************************
                                                                                                                      ; dwload temporary RAM storage
0010                  noexec   equ   $10                                                      ; inhibit exec after loading
0011                  bootflag equ   $11                                                      ; zero if run from command line, $55 if autorun on boot
01CD                  sstack   equ   $1cd                                                     ; backup of original stack
01CF                  sector   equ   $1cf                                                     ; counter, used when loading file contents, just below namebuf/DECB header
01D1                  namebuf  equ   $1d1                                                     ; fixed by BASIC read filename function
01D1                  decbhdr  equ   namebuf                                          ; DECB headers get copied in here
01D9                  startbuf equ   namebuf+8                                        ; dw input data buffer, can reuse buffer after DECB header slot
02D9                  endbuf   equ   startbuf+256                             ; 
                      ;tmpstack equ   $3fc                                                    ; normally enough unused area below here (goes 30 bytes deep)
                                                                                                                      ; Entry point from command line (DWLOAD or EXEC)
                                                                                                                      ; for later to see that we came from DWLOAD command
F882  0F11            dwload   clr   <bootflag                                        ; clear our autoboot flag
F884  8EF990                   ldx   #dwtext-1                                        ; 
F887  BD90E5                   jsr   outstr                                           ; print string
F88A  1A50            Start           orcc  #IntMasks                                 ; disable FIRQ, IRQ
F88C  8EFF20                   ldx   #PIA1Base                        ; PIA1
F88F  C602                     ldb   #$02
F891  E784                     stb   ,x                               ; make RS-232 output marking
                                                                                                                      ; Spin for a while so that the RS-232 bit stays hi for a time
F893  8EA000          Reset           ldx   #$a000
F896  301F            Spin     leax  -1,x
F898  26FC                     bne    Spin
                                                                                                                      ; Request named object from DW server
F89A  9611                     lda   <bootflag                                        ; coming from autoboot?
F89C  263E                     bne   autoreq                                          ; then use default name
F89E  9DA5                     jsr   <redLChr                                         ; peek at next character
F8A0  804E                     suba  #'N'                                                     ; DLOADN ?
F8A2  9710                     sta   <noexec                                          ; 0 = noexec
F8A4  2602                     bne   getn
F8A6  9D9F                     jsr   <getNChr                                         ; read next char (the N)
F8A8  BDB7AA          getn     jsr   getfnam                                          ; read file name and length into $1d1 = namebuf
F8AB  8E01D0                   ldx   #namebuf-1                                       ; packet buffer start
F8AE  6F84                     clr   ,x                                                               ; zero MSB for Y
F8B0  10AE81                   ldy   ,x++                                                     ; length of file name (16 bit)
F8B3  2729                     beq   noname                                           ; no file name given?
F8B5  E61F                     ldb   -1,x                                                     ; length of name (8 bit)
F8B7  6F85                     clr   b,x                                                      ; zero terminate name string (for error)
F8B9  6C83                     inc   ,--x                                                     ; 1 = DriveWire OP_NAMEOBJ_MOUNT
F8BB  3122                     leay  2,y                                                      ; length of DW packet, name length + 2
F8BD  BDFAE2          reqobj  jsr   DWWrite
F8C0  8E0012                   ldx   #dnum                                                    ; Our drive number variable
F8C3  6F84                     clr   ,x
F8C5  3121                     leay  1,y                                                      ; read one byte back (Y is 0 after DWWrite)
F8C7  BDFA5D                   jsr   DWRead                                           ; get drive number
F8CA  2401                     bcc    reqOK                                                   ; if open has not failed go on
F8CC  39                                      rts                                                                     ; else return
F8CD  6D84            reqOK    tst   ,x
F8CF  2616                     bne   ReadDECB                                         ; successful mount
F8D1  0D11                     tst   <bootflag
F8D3  102600FC                 lbne  stealth                                          ; silent failure, BASIC OK prompt
F8D7  C619                     ldb   #$19                                                     ; MO ERROR
F8D9  7E8344                   jmp   syserr                                           ; system error
F8DC  9710            autoreq  sta   <noexec                          ; a is non null, autoexec
F8DE  8EF9C1          noname   ldx   #autoname
F8E1  108E000E                 ldy   #12+2                                                    ; length of DW packet
F8E5  20D6                     bra   reqobj
                                                                                                                      ; named object has been mounted, proceed to read in file
F8E7  8E0000          ReadDECB        ldx   #0000
F8EA  BF01CF                   stx   sector
F8ED  CE02D9                   ldu   #endbuf                                          ; start with empty buffer
F8F0  8CF8F9                                  cmpx    #nextseg
F8F3  2104                                    brn     nextseg
F8F5  10FF01CD                                sts     sstack
                                                                                                                      ; read DECB header
F8F9  108E0005        nextseg  ldy   #5
F8FD  8E01D1                   ldx   #decbhdr                                         ; copy DECB header here
F900  BDF9E3                   jsr   copybuf                                          ; moves x ahead
F903  A61B                     lda   -5,x                                             ; DECB segment header starts with zero
F905  2753                     beq   ml                                               ; normal data segment
F907  8155                     cmpa  #$55                                             ; Dragon DOS file?
F909  261A                     bne   noddd
F90B  3124                     leay  4,y                                              ; remaining header bytes (y is 0 here)
F90D  BDF9E3                   jsr   copybuf
F910  6A18                     dec   -8,x                                             ; Dragon DOS BASIC = 1
F912  2604                     bne   ddbin
F914  301F                     leax  -1,x                                             ; length at x-5
F916  2022                     bra   ldbas
F918  10AE1B          ddbin    ldy   -5,x                                             ; length
F91B  AE19                     ldx   -7,x                                             ; load address
F91D  BDF9E3                   jsr   copybuf                                          ; read whole file
F920  8E01D9                   ldx   #decbhdr+8                                       ; exec address ptr + 2
F923  204B                     bra   oldsp
F925  4C              noddd    inca
F926  102600A6                 lbne  ErrDWL                                           ; must be $FF otherwise
F92A  3121                     leay  1,y                                              ; Y is 0 after copybuf
F92C  10BC01CF                 cmpy  sector                                           ; only first sector can be BASIC header
F930  263A                     bne   endseg                                           ; otherwise end flag
F932  118301DE                 cmpu  #startbuf+5                                      ; must be first segment also
F936  2634                     bne   endseg
                                                                                                                      ; loading DECB BASIC file
                                                                                                                      ; bytes 4 and 5 are actually part of the program
F938  335E                     leau  -2,u                                             ; u was 5 bytes into read buffer here
                                                                                                                      ; at this point x is past 5-byte header, u is 3 bytes into first 256-bytes block
                                                                                                                      ; for Dragon DOS x is 8 bytes into header, u is 9 bytes into first 256-bytes block
F93A  10AE1C          ldbas    ldy   -4,x                                             ; read whole BASIC program
F93D  9E19                     ldx   <$19
F93F  BDF9E3                   jsr   copybuf
F942  BDF964                   jsr   prbang
                                                                                                                      ; set up BASIC pointers and finish
F945  9F1B                     stx   <$1b                                             ; end of BASIC program
F947  9F1D                     stx   <$1d
F949  9F1F                     stx   <$1f
F94B  0D10                     tst   <noexec
F94D  1027BDDC                                lbeq  $b72d                                             ; print OK, run basvect1, basvect2, readline
F951  BD841F                   jsr   basvect1                                         ; BasVect1 reset BASIC stack etc
F954  BD83ED                   jsr   basvect2                                         ; BasVect2 initialize BASIC
F957  7E849F                   jmp   runbasic                                         ; run_basic
F95A  10AE1C          ml       ldy   -4,x                                             ; DECB segment length
F95D  AE1E                     ldx   -2,x                                             ; DECB segment loading address
F95F  BDF9E3                   jsr   copybuf
F962  2095                     bra   nextseg
F964  BD9A89          prbang   jsr   backspc                                          ; print backspace to devnum
F967  8621                     lda   #'!'                                             ; print bang
F969  7EB54A                   jmp   outchr                                           ; print to devnum
F96C  EE1C            endseg  ldu   -4,x                                              ; new stack pointer specified?
F96E  2603                     bne   setsp
F970  FE01CD          oldsp    ldu   sstack                                           ; otherwise restore original SP
F973  1F34            setsp    tfr   u,s
F975  8DED                     bsr   prbang
F977  1CAF                     andcc #~IntMasks                                       ; enable interrupts
F979  AE1E                     ldx   -2,x                                             ; exec address
F97B  1F10                     tfr   x,d
F97D  4C                       inca
F97E  2708                     beq   retbas                                           ; return to basic if exec address $FFxx
F980  9F9D                     stx   <$9d                                             ; save BASIC EXEC address
F982  0D10                     tst   <noexec
F984  2702                     beq   retbas
F986  AD84                     jsr   ,x                                               ; and run loaded program
F988  39              retbas   rts
                                                                                                                      ; vector table for chainloaders, etc
F989  FA5D            dwrvec   fdb   DWRead
F98B  FAE2            dwwvec   fdb   DWWrite
F98D  FA0F            dwdvec   fdb   DoRead
F98F  0000                     fdb   0
F991  44574C4F4144    dwtext   fcc   /DWLOAD/
F997  0D00                     fcb   $0d,0
F999  EEE1            fromboot ldu   ,s++                                                     ; check return address (and drop it)
F99B  1183B469                 cmpu  #$b469                                           ; ROM code location (not a copy)
F99F  263F                     bne   gocmdl                                           ; return if run from a copied code segment
F9A1  8CB44F                   cmpx  #wrmStrt                                         ; coldstart sets this, warmstart doesn't
F9A4  263A                     bne   gocmdl                                           ; return if warmstart
                                                                                                                      ; check for SHIFT key
F9A6  B6FF02          chkshift lda   $ff02                                                    ; save PIA
F9A9  C67F                     ldb   #$7f
F9AB  F7FF02                   stb   $ff02
F9AE  F6FF00                   ldb   $ff00
F9B1  B7FF02                   sta   $ff02                                                    ; restore PIA
F9B4  53                       comb
F9B5  C440                     andb  #$40
                      ;;;                     bne   gocmdl                                            ; return if shift key pressed
F9B7  2625                                    bne   ste01                                                     ; exit but enabling interrupts
F9B9  5C                       incb                                                                   ; B was 0 before
F9BA  D711                     stb   <bootflag                                        ; our autoboot flag = 1
F9BC  BDF88A                   jsr   Start
                      ;;;         bra   gocmdl
F9BF  201D                                    bra     ste01                                                   ; exit but enabling interrupts
F9C1  01              autoname fcb   $01                                                      ; OP_NAMEOBJ_MOUNT
F9C2  0C                       fcb   12                                                               ; length of name string
F9C3  4155544F4C4F41442E44574C00          fcn   /AUTOLOAD.DWL/
F9D0  16BE78          ErrDWL   lbra  ioerror                                          ; BASIC IO ERROR
                      
F9D3  CCF873          stealth  ldd    #retFDwl                                                ; get return address from DWLOAD
F9D6  10A362                                  cmpd    2,s                                                     ; is this on stack as 2nd retAdrs?
F9D9  2603                                    bne     ste01                                                   ; no, skip next one
F9DB  3262                                    leas    2,s                                                     ; get rid of return address (from checkshift)
F9DD  39                                      rts                                                                     ; return
                      
F9DE  1CAF            ste01           andcc #~IntMasks                                        ; enable interrupts
F9E0  7E8371          gocmdl   jmp   okprmpt                                          ; BASIC OK prompt
                                                                                                                      ; copy y chars from read buffer to x, updates u
F9E3                  copybuf
F9E3  118302D9        copyl    cmpu  #endbuf
F9E7  261D                     bne   copym
                                                                                                                      ; fill up buffer via DW - resets buffer pointer u
F9E9  3430                     pshs  x,y
F9EB  BE01CF                   ldx   sector
F9EE  108E01D9                 ldy   #startbuf
F9F2  1F23                     tfr   y,u
F9F4  8D19                     bsr   DoRead
F9F6  25D8                     bcs   ErrDWL
F9F8  26D6                     bne   ErrDWL
F9FA  3001                     leax  1,x
F9FC  BF01CF                   stx   sector
F9FF  862E                     lda   #'.'                                                     ; print dot
FA01  BDB54A                   jsr   outchr
FA04  3530                     puls  x,y
FA06  A6C0            copym    lda   ,u+
FA08  A780                     sta   ,x+
FA0A  313F                     leay  -1,y
FA0C  26D5                     bne   copyl
FA0E  39                       rts
                                                                                                                      ; also used by DWRead/DWWrite
FF20                  PIA1Base equ   $ff20
0050                  IntMasks equ   $50
                                                                                                                      ; DW4 Opcodes
00D2                  OP_READEX       equ   'R'+$80                                   ; Read one sector
00F2                  OP_REREADEX     equ   'r'+$80                                   ; Re-read one sector
                                                                                                                              ; Error definitions
00F3                  E_CRC    equ   $f3                              ; Same as NitrOS-9 E$CRC
                                                                                                                              ; below code is taken from toolshed/dwdos
FA0F  9612            DoRead  lda   <dnum                                                     ; our drive number
FA11  5F                       clrb                                                                   ; LSN bits 23-16
FA12  3436                     pshs  d,x,y
FA14  86D2                     lda   #OP_READEX
FA16  3402            ReRead   pshs  a
FA18  30E4                     leax  ,s
FA1A  108E0005                                ldy   #$0005
FA1E  1700C1                                  lbsr  DWWrite
FA21  3502                                    puls  a
FA23  AE64                                    ldx   4,s                                                       ; get read buffer pointer
FA25  108E0100                                ldy   #256                                                      ; read 256 bytes
FA29  CC0085                                  ldd   #133*1                                            ; 1 second timeout
FA2C  8D2F                                    bsr   DWRead
FA2E  252B                     bcs   ReadEx
FA30  2629                     bne   ReadEx
                                                                                                                      ; Send 2 byte checksum
FA32  3420                                    pshs  y
FA34  30E4                                    leax  ,s
FA36  108E0002                                ldy   #2
FA3A  1700A5                                  lbsr  DWWrite
FA3D  108E0001                                ldy   #1
FA41  CC0085                                  ldd   #133*1
FA44  8D17                                    bsr   DWRead
FA46  3262                                    leas  2,s
FA48  2511                                    bcs   ReadEx
FA4A  260F                    bne   ReadEx
FA4C  A6E4                                    lda   ,s
FA4E  270B                                    beq   ReadEx
FA50  81F3                                    cmpa  #E_CRC
FA52  2606                                    bne   ReadErr
FA54  86F2                                    lda   #OP_REREADEX
FA56  6FE4                                    clr   ,s
FA58  20BC                                    bra   ReRead  
FA5A  53              ReadErr  comb
FA5B  35B6            ReadEx  puls  d,x,y,pc
                                                                                                                      ; DW4 Addresses
FF20                  BBOUT           equ     $ff20
FF22                  BBIN            equ     $ff22
                      ;***********************************************************************
                      ; DWRead
                      ;    Receive a response from the DriveWire server.
                      ;    Times out if serial port goes idle for more than 1.4 (0.7) seconds.
                      ;    Serial data format:  1-8-N-1
                      ;    4/12/2009 by Darren Atkinson
                      ; ----------------------------------------------------------------------
                      ;    Added a timeout control to the Becker part to avoid infinite loops
                      ;    11-Mars-2016 by Pere Serrat
                      ; ----------------------------------------------------------------------
                      ; Entry:
                      ;    X  = starting address where data is to be stored
                      ;    Y  = number of bytes expected
                      ; Exit:
                      ;    CC = carry set on framing error, Z set if all bytes received
                      ;    X  = starting address of data received
                      ;    Y  = checksum
                      ;    U is preserved.  All accumulators are clobbered
                      ;***********************************************************************
                                              IF becker > 0
                      DWRead   clra                                   ; clear Carry (no framing error)
                               deca                                   ; clear Z flag, A = timeout msb ($ff)
                               tfr    cc,b
                               pshs  u,x,dp,b,a                       ; preserve registers, push flags and timeout msb
                               leau  ,x                                                               ; U points to received data buffer
                               ldx   #$0000                                           ; checksum to zero
                               leas   -2,s                                                    ; create a word hole for a 16 bits counter
                      loop0    ldd    #$F000                                          ; reset counter to a fixed value
                                              std     ,s                                                              ; to count 16 times 256
                      loop1    dec    1,s                                                     ; decrement low byte counter
                                              bne     loop2                                                   ; if not zero, skip next byte control
                                              inc     ,s                                                              ; increment high byte
                                              beq     notReady                                                ; if zero, exit via notReady
                      loop2           ldb   $ff49                                                     ; pol if DW is ready
                               bitb  #$02                                                     ; is ready?
                               beq   loop1                                                    ; no, loop
                               ldb   $ff4a                                                    ; get data byte
                               stb   ,u+                                                      ; put at destination buffer
                               abx                                                                    ; accumulate in checksum word
                               leay  ,-y                                                      ; decrement byte counter
                                              bne     loop0                                                   ; not all received, loopback
                                              leas    2,s                                                     ; discard word counter off the stack
                               tfr   x,y                                                      ; pass checksum to Y
                               ldb   #0
                               lda   #3
                               leas  1,s                              ; remove timeout msb from stack
                      outLoop  inca                                   ; A = status to be returned in C and Z (bit2-Z; bit0-C), now A=4 (Z=1 C=0)
                               ora   ,s                               ; place status information into the..
                               sta   ,s                               ; ..C and Z bits of the preserved CC
                               leay  ,x                               ; get checksum into Y
                               puls  cc,dp,x,u,pc                     ; restore registers and return
                      
                      notReady leas   3,s                                                     ; to remove word counter and timeout
                                              clra                                                                    ; to get a #$01 in next opcode (to set Carry flag)
                                              bra     outLoop                                         ; update flags and exit
                      
                                              ELSE
FA5D  4F              DWRead   clra                                   ; clear Carry (no framing error)
FA5E  4A                       deca                                   ; clear Z flag, A = timeout msb ($ff)
FA5F  1FA9                     tfr   cc,b
FA61  345E                     pshs  u,x,dp,b,a                       ; preserve registers, push timeout msb
FA63  1F8B                     tfr   a,dp                             ; set direct page to $FFxx
FA65                           setdp $ff
FA65  3384                     leau  ,x                               ; U = storage ptr
FA67  8E0000                   ldx   #0                               ; initialize checksum
FA6A  8601                     lda   #$01                             ; A = serial in mask
FA6C  2032                     bra   rx0030                           ; go wait for start bit
                                                                                                                      ; Read a byte
FA6E  3341            rxByte   leau  1,u                              ; bump storage ptr
FA70  31A2                     leay  ,-y                              ; decrement request count
FA72  9622                     lda   <BBIN                            ; read bit 0
FA74  44                       lsra                                   ; move bit 0 into Carry
FA75  CCFF20                   ldd   #$ff20                           ; A = timeout msb, B = shift counter
FA78  A7E4                     sta   ,s                               ; reset timeout msb for next byte
FA7A  56                       rorb                                   ; rotate bit 0 into byte accumulator
FA7B  9622            rx0010   lda   <BBIN                            ; read bit (d1, d3, d5)
FA7D  44                       lsra
FA7E  56                       rorb
FA7F  A561                     bita  1,s                              ; 5 cycle delay
FA81  2508                     bcs   rx0020                           ; exit loop after reading bit 5
FA83  9622                     lda   <BBIN                            ; read bit (d2, d4)
FA85  44                       lsra
FA86  56                       rorb
FA87  33C4                     leau  ,u
FA89  20F0                     bra   rx0010
FA8B  9622            rx0020   lda   <BBIN                            ; read bit 6
FA8D  44                       lsra
FA8E  56                       rorb
FA8F  31A4                     leay  ,y                               ; test request count
FA91  2737                     beq   rx0050                           ; branch if final byte of request
FA93  9622                     lda   <BBIN                            ; read bit 7
FA95  44                       lsra
FA96  56                       rorb                                   ; byte is now complete
FA97  E75F                     stb   -1,u                             ; store received byte to memory
FA99  3A                       abx                                    ; update checksum
FA9A  9622                     lda   <BBIN                            ; read stop bit
FA9C  8401                     anda  #$01                             ; mask out other bits
FA9E  2737                     beq   rxExit                           ; exit if framing error
                                                                                                                      ; Wait for a start bit or timeout
FAA0  9522            rx0030   bita  <BBIN                            ; check for start bit
FAA2  27CA                     beq   rxByte                           ; branch if start bit detected
FAA4  9522                     bita  <BBIN                            ; again
FAA6  27C6                     beq   rxByte
FAA8  C6FF                     ldb   #$ff                             ; init timeout lsb
FAAA  9522            rx0040   bita  <BBIN
FAAC  27C0                     beq   rxByte
FAAE  C001                     subb  #1                               ; decrement timeout lsb
FAB0  9522                     bita  <BBIN
FAB2  27BA                     beq   rxByte
FAB4  24F4                     bcc   rx0040                           ; loop until timeout lsb rolls under
FAB6  9522                     bita  <BBIN
FAB8  27B4                     beq   rxByte
FABA  EBE4                     addb  ,s                               ; B = timeout msb - 1
FABC  9522                     bita  <BBIN
FABE  27AE                     beq   rxByte
FAC0  E7E4                     stb   ,s                               ; store decremented timeout msb
FAC2  9522                     bita  <BBIN
FAC4  27A8                     beq   rxByte
FAC6  25D8                     bcs   rx0030                           ; loop if timeout hasn't expired
FAC8  200D                     bra   rxExit                           ; exit due to timeout
FACA  9622            rx0050   lda   <BBIN                            ; read bit 7 of final byte
FACC  44                       lsra
FACD  56                       rorb                                   ; byte is now complete
FACE  E75F                     stb   -1,u                             ; store received byte to memory
FAD0  3A                       abx                                    ; calculate final checksum
FAD1  9622                     lda   <BBIN                            ; read stop bit
FAD3  8401                     anda  #$01                             ; mask out other bits
FAD5  8A02                     ora   #$02                             ; return SUCCESS if no framing error
                                                                                                                      ; Clean up, set status and return
FAD7  3261            rxExit   leas  1,s                              ; remove timeout msb from stack
FAD9  4C                       inca                                   ; A = status to be returned in C and Z
FADA  AAE4                     ora   ,s                               ; place status information into the..
FADC  A7E4                     sta   ,s                               ; ..C and Z bits of the preserved CC
FADE  3184                     leay  ,x                               ; return checksum in Y
FAE0  35D9                     puls  cc,dp,x,u,pc                     ; restore registers and return
FAE2                           setdp $00
                                              ENDIF
                      ;******************************************************
                      ; DWWrite
                      ;    Send a packet to the DriveWire server.
                      ;    Serial data format:  1-8-N-1
                      ;    4/12/2009 by Darren Atkinson
                      ; Entry:
                      ;    X  = starting address of data to send
                      ;    Y  = number of bytes to send
                      ; Exit:
                      ;    X  = address of last byte sent + 1
                      ;    Y  = 0
                      ;    All others preserved
                      ;******************************************************
                                              IF becker > 0
                      DWWrite  pshs  d,cc                             ; preserve registers
                      txByte  lda   ,x+
                               sta   $ff4a
                               leay  -1,y                     ; decrement byte counter
                               bne   txByte                   ; loop if more to send
                               puls  cc,d,pc                  ; restore registers and return
                      
                                              ELSE                                                       
FAE2  340F            DWWrite  pshs  dp,d,cc                  ; preserve registers
FAE4  CC04FF                   ldd   #$04ff                   ; A = loop counter, B = $ff
FAE7  1F9B                     tfr   b,dp                     ; set direct page to $FFxx
FAE9                           setdp $ff                                      
FAE9  D623                     ldb   <$ff23                   ; read PIA 1-B control register
FAEB  C4F7                     andb  #$f7                     ; clear sound enable bit
FAED  D723                     stb   <$ff23                   ; disable sound output
FAEF  8C                       fcb   $8c                      ; skip next instruction
FAF0  D720            txByte   stb   <BBOUT                   ; send stop bit
FAF2  E680                     ldb   ,x+                      ; get a byte to transmit
FAF4  12                       nop                                                    
FAF5  58                       lslb                           ; left rotate the byte two positions..
FAF6  59                       rolb                           ; ..placing a zero (start bit) in bit 1
FAF7  D720            tx0020   stb   <BBOUT                   ; send bit (start bit, d1, d3, d5)
FAF9  56                       rorb                           ; move next bit into position
FAFA  1E88                     exg   a,a                                      
FAFC  12                       nop                                                    
FAFD  D720                     stb   <BBOUT                   ; send bit (d0, d2, d4, d6)
FAFF  56                       rorb                           ; move next bit into position
FB00  33C4                     leau  ,u                                               
FB02  4A                       deca                           ; decrement loop counter
FB03  26F2                     bne   tx0020                   ; loop until 7th data bit has been sent
FB05  D720                     stb   <BBOUT                   ; send bit 7
FB07  CC0402                   ldd   #$0402                   ; A = loop counter, B = MARK value
FB0A  31A2                     leay  ,-y                      ; decrement byte counter
FB0C  26E2                     bne   txByte                   ; loop if more to send
FB0E  D720                     stb   <BBOUT                   ; leave bit banger output at MARK
FB10  358F                     puls  cc,d,dp,pc               ; restore registers and return
FB12                           setdp $00
                                              ENDIF
                      ; ============================================================================================================================
                                                                                                                      ; THIS PART WILL ONLY BE NEEDED IF WE MODIFY THE PATCHES-CODE
                                                                                                                      ; IN THIS CASE THE DOSPLUS HAS TO BE RE-PATCHED TO UPDATE LINKS/HOOKS
                                                                                                                      ; No problem as long as we just work from the init routine downwards
                                                                                                                      ; simply EXEC $reptch from $6xxx having DPlus at $4000 and resave the DOS
                                                                                                                      ; for easy of use just need to do: jmp $c005
                      ; ============================================================================================================================
                                                                                                                      ; 1st patch. To redirect ReadAbsoluteSector and WriteAbsoluteSector to DW4 new process
FB12  338DC5F0        reptch  leau    $c106,pcr                                       ; the DOS address code to be redirected here
FB16  CC7E12                                  ldd     #$7e12                                          ; opcodes for jmp and NOP
FB19  8EE008                                  ldx     #entry                                          ; new entry point for DOS sector READ-WRITE commands
FB1C  A7C0                                    sta     ,u+                                                     ; put jmp
FB1E  AFC1                                    stx     ,u++                                                    ; put the new dispatch routine
FB20  E7C4                                    stb     ,u                                                              ; put NOP
                                                                                                                      ; 2nd patch - for DSKINIT while writting to NOT directory sectors
FB22  338DC7FF                                leau    $c325,pcr                                       ; point to intercept
FB26  8EE0CA                                  ldx     #inidsk                                         ; new added code
FB29  A7C0                                    sta     ,u+                                                     ; put jmp
FB2B  AFC4                                    stx     ,u                                                              ; put the new code address
                                                                                                                      ; 3rd patch to fill the three buffers for DSKINIT ($800-$900-$a00) writting to directory sectors
FB2D  338DCEC2                                leau    $c9f3,pcr                                       ; point to intercept
FB31  8EE145                                  ldx     #fildir                                         ; new added code
FB34  A7C0                                    sta     ,u+                                                     ; put jmp
FB36  AFC4                                    stx     ,u                                                              ; put the new code address
                                                                                                                      ; 4th patch to correct track number when verifying tracks in DSKINIT to VDK files
FB38  338DD668                                leau    $d1a4,pcr                                       ; point to intercept
FB3C  8EE1EE                                  ldx     #modtrk                                         ; new added code
FB3F  A7C0                                    sta     ,u+                                                     ; put jmp
FB41  AFC4                                    stx     ,u                                                              ; put the new code address
                                                                                                                      ; 5th patch to update system table at $6a7-8-9-a for BACKUP
FB43  338DCCEF                                leau    $c836,pcr                                       ; point to intercept
FB47  8EE276                                  ldx     #patbck                                         ; new added code
FB4A  A7C0                                    sta     ,u+                                                     ; put jmp
FB4C  AFC4                                    stx     ,u                                                              ; put the new code address
                                                                                                                      ; 6th patch to prevent a rutine to test hdw for DW4 drives
FB4E  338DC61E                                leau    $c170,pcr                                       ; point to intercept
FB52  8EE29F                                  ldx     #noHdw                                          ; new added code
FB55  A7C0                                    sta     ,u+                                                     ; put jmp
FB57  AFC1                                    stx     ,u++                                                    ; put the new code address
FB59  E7C4                                    stb     ,u                                                              ; put NOP
                                                                                                                      ; 7th patch for DOS reset routine
FB5B  338DCB6A                                leau    $c6c9,pcr                                       ; point to intercept
FB5F  8EE2C1                                  ldx     #dosReset                                       ; new added code
FB62  A7C0                                    sta     ,u+                                                     ; put jmp
FB64  AFC1                                    stx     ,u++                                                    ; put the new code address
                      
FB66  39                                      rts                                                                     ; return to caller
                                              
                      ;********************************************************************************************************************************
                                              IF drgrom > 0
                                                      if (*%256 > 0)
                                                                      rzb      256-(*%256)    ; make it ROM compatible (length multiple of 256 bytes)
                                                      endif
                                              ENDIF
                      ;********************************************************************************************************************************
FB67                  absEnd  end
